/*
 * =-=-=-=-=-=-=-=-=-=-=-  NOTICE  =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * The following is the license that applies to this copy of the Prayer2
 * software.  For a license to use the Prayer2 software under conditions
 * other than those described here, please contact MindNever Software at
 * following addresses:
 *
 * MindNever Software <info@mindnever.org>
 * MindNever Software +381 63 550-161 or +381 64 220-7115
 * 
 * =-=-=-=-=-=-=-=-=-=-  COPYRIGHT  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Prayer2 is:
 *   Copyright (c) 1999-2001 by Tamara Cetinjski <tamara@mindnever.org>
 *   Copyright (c) 2001-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *   Copyright (c) 1999-2002 by MindNever Software Group <info@mindnever.org>
 *
 * Tamara Cetinjski is
 *   Copyright (c) 1999-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *
 *
 * =-=-=-=-=-=-=-=-=-=-  LICENSE  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Redistributions in any form must be accompanied by information on
 *    how to obtain complete source code for the Prayer2 software and any
 *    accompanying software that uses the Prayer2 software.  The source code
 *    must either be included in the distribution or be available for no
 *    more than the cost of distribution plus a nominal fee, and must be
 *    freely redistributable under reasonable conditions.  For an
 *    executable file, complete source code means the source code for all
 *    modules it contains.  It does not include source code for modules or
 *    files that typically accompany the major components of the operating
 *    system on which the executable file runs.
 *
 * =-=-=-=-=-=-=-=-=-=-  NO WARRANTY =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
 *
 * THIS SOFTWARE IS PROVIDED BY MINDNEVER SOFTWARE ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
 * NON-INFRINGEMENT, ARE DISCLAIMED.  IN NO EVENT SHALL MINDNEVER SOFTWARE
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, HAIR, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
#define USE_BUILTIN_MATH
#define INTUI_V36_NAMES_ONLY
#define __USE_SYSBASE

#include <exec/types.h>
#include <intuition/intuition.h>
#include <intuition/classes.h>
#include <intuition/imageclass.h>

#include <proto/intuition.h>
#include <proto/utility.h>
#include <proto/graphics.h>
#include <graphics/gfxbase.h>
#include <utility/hooks.h>
#include <proto/exec.h>
#include <exec/memory.h>
#include <proto/cybergraphics.h>
#include <cybergraphx/cybergraphics.h>



#include <proto/dos.h>

#include	<exec/ports.h>
#include	<exec/tasks.h>
#include	<exec/execbase.h>

#include	<string.h>
#include	<stdarg.h>
#include 	<zlib.h>

#include "PrayerImageClass.h"
#include "pif.h"

#include "opt68k.h"
#include "classhelp.h"

static struct c2pStruct
{
 struct BitMap *bmap;
 UWORD startX, startY, width, height;
 UBYTE *chunkybuffer;
} c2p = {0};


struct PIData {
		struct BitMap *BitMap;
		struct PIFHdr *Pif;
		struct Screen *DestScr;
		struct Image *Image;
		BOOL	Remap;
		WORD	Precision;
		BOOL	UseFastMem;
		UBYTE	PenAlloc[256];
	};

static ULONG	PI_OMSet(Class *cl, struct Image *im, struct opUpdate *msg);
static ULONG	PI_OMGet(Class *cl, struct Image *im, struct opGet *msg);
static ULONG	PI_OMNew(Class *cl, struct Image *im, struct opSet *msg);
static void		PI_OMDispose(Class *cl, struct Image *im, Msg msg);
static void		PI_IMDraw(Class *cl, struct Image *im, struct impDraw *msg);
static void		LayoutImage(struct PIData *pi);
static UBYTE *RemapImage(UBYTE *raw, struct PIData *pi);


#ifdef __MORPHOS__
static ULONG __saveds PI_Dispatcher_Native(void)
{
Class *cl = (Class *) REG_A0;
struct Image *im = (struct Image *) REG_A2;
Msg msg = (Msg) REG_A1;

#else
static ULONG __saveds PI_Dispatcher(Class *cl ASM("a0"), struct Image *im ASM("a2"), Msg msg ASM("a1"))
{
#endif

switch(msg->MethodID) {
		case	OM_SET:	//
		case	OM_UPDATE:	//
			return(PI_OMSet(cl, im, (APTR) msg));

		case	OM_GET:	//
			return(PI_OMGet(cl, im, (APTR) msg));

		case	OM_NEW:	//
			return(PI_OMNew(cl, im, (APTR) msg));

		case	OM_DISPOSE:	//
			PI_OMDispose(cl, im, msg);
			return(TRUE);
		case	IM_DRAW:
			PI_IMDraw(cl, im, (APTR) msg);
			return(TRUE);

		default:
			return(DoSuperMethodA(cl, (APTR) im, msg));
	}
}


#ifdef __MORPHOS__
struct EmulLibEntry PI_Dispatcher = {
		TRAP_LIB,
		0,
		(void (*)(void))PI_Dispatcher_Native
	};
#endif


static ULONG	PI_OMNew(Class *cl, struct Image *im, struct opSet *msg)
{
struct PIData *pi;

if(im = (APTR) DoSuperMethodA(cl, (APTR) im, (Msg) msg)) {
		if(pi = INST_DATA(cl, (APTR) im)) {
				PI_OMSet(cl, im, (APTR) msg);
				pi->Image = im;
				LayoutImage(pi);
			}
	}
return((ULONG) im);
}

static void PI_OMDispose(Class *cl, struct Image *im, Msg msg)
{
struct PIData *pi = INST_DATA(cl, im);
ULONG i, j;
if(pi->BitMap) {
		WaitBlit();
		FreeBitMap(pi->BitMap);
		if(pi->DestScr) {
				for(i = 0; i < 256; i++) {
						j = pi->PenAlloc[i];
						while(j--)
							ReleasePen(pi->DestScr->ViewPort.ColorMap, i);
					}
			}
	}
DoSuperMethodA(cl, (APTR) im, (Msg) msg);
}

static ULONG	PI_OMSet(Class *cl, struct Image *im, struct opUpdate *msg)
{
struct PIData *pi = INST_DATA(cl, im);
struct TagItem *ti, *tstate = msg->opu_AttrList;

while(ti = NextTagItem(&tstate))
	switch(ti->ti_Tag) {
			case	PIA_PIF: // initialize
				pi->Pif = (APTR) ti->ti_Data;
				break;
			case	PIA_DestScr:
				pi->DestScr = (APTR) ti->ti_Data;
				break;
			case	PIA_Remap:
				pi->Remap = ti->ti_Data;
				break;
			case	PIA_Precision:
				pi->Precision = ti->ti_Data;
				break;
			case	PIA_UseFastMem:
				pi->UseFastMem = ti->ti_Data;
				break;
			case	IA_Left:
				im->LeftEdge = ti->ti_Data;
				break;
			case	IA_Top:
				im->TopEdge = ti->ti_Data;
				break;
			case	IA_Width:
				im->Width = ti->ti_Data;
				break;
			case	IA_Height:
				im->Height = ti->ti_Data;
				break;
		}
return(TRUE);
}

static ULONG	PI_OMGet(Class *cl, struct Image *im, struct opGet *msg)
{
struct PIData *pi = INST_DATA(cl, im);

switch(msg->opg_AttrID) {
		case	IA_Left:
			*msg->opg_Storage = im->LeftEdge;
			return(TRUE);
		case	IA_Top:
			*msg->opg_Storage = im->TopEdge;
			return(TRUE);
		case	IA_Width:
			*msg->opg_Storage = im->Width;
			return(TRUE);
		case	IA_Height:
			*msg->opg_Storage = im->Height;
			return(TRUE);
		case	PIA_BitMap:
			*msg->opg_Storage = (ULONG) pi->BitMap;
			return(TRUE);
	}
return(TRUE);
}

static void		PI_IMDraw(Class *cl, struct Image *im, struct impDraw *msg)
{
struct PIData *pi = INST_DATA(cl, im);

if(pi->BitMap && msg->imp_RPort)
	BltBitMapRastPort(pi->BitMap, 0, 0, msg->imp_RPort,
										im->LeftEdge + msg->imp_Offset.X, 
										im->TopEdge + msg->imp_Offset.Y,
										im->Width,
										im->Height, 0x0C0);

}

static void		LayoutImage(struct PIData *pi)
{
struct PIFHdr *phdr;
unsigned long depth = 1, rawlen;
UBYTE *raw, *chunky;
struct RastPort rp;

if(phdr = pi->Pif) {
		if(IS_PIF(phdr->Id) && (PIF_VERSION(phdr->Id) <= 0)) {
				rawlen = (phdr->Width * phdr->Height) + (phdr->NumColors * sizeof(struct PrayerPalette));
				if(raw = AllocVec(rawlen, MEMF_ANY)) {
						if(!uncompress(raw, &rawlen, phdr->CompressedData, phdr->DataLength)) {
								if(pi->DestScr && pi->Remap)
									depth = GetBitMapAttr(pi->DestScr->RastPort.BitMap, BMA_DEPTH);

								pi->Image->Width = phdr->Width;
								pi->Image->Height = phdr->Height;
//								pi->Image->Depth = depth;
								pi->Image->PlanePick = (1 << depth) - 1;

								if(pi->BitMap = AllocBitMap(phdr->Width, phdr->Height, depth, BMF_MINPLANES, pi->DestScr ? pi->DestScr->RastPort.BitMap : NULL)) {
										if(pi->DestScr) {
												CopyMem(&pi->DestScr->RastPort, &rp, sizeof(rp));
												rp.Layer = NULL;
											} else {
												InitRastPort(&rp);
											}

										rp.BitMap = pi->BitMap;

										chunky = raw + (phdr->NumColors * sizeof(struct PrayerPalette));

										if(CyberGfxBase && (depth > 8)) {
												struct PrayerPalette *pal = (APTR) raw;
												ULONG *raw24, i, l;
												UBYTE p;
												
												l = phdr->Width * phdr->Height;
												
												if(raw24 = AllocVec(sizeof(ULONG) * l, MEMF_CLEAR)) {
														for(i = 0; i < l; i++) {
																p = chunky[i];
																raw24[i] = (pal[p].R << 24 | pal[p].G << 16 | pal[p].B << 8);
															}
														WritePixelArray(raw24,
																						0, 0,
																						phdr->Width * sizeof(ULONG),
																						&rp,
																						0, 0, phdr->Width, phdr->Height,
																						RECTFMT_RGBA);
														FreeVec(raw24);
													}
											} else {

												if(pi->Remap && pi->DestScr)
													RemapImage(raw, pi);

												if(GetBitMapAttr(pi->BitMap, BMA_FLAGS) & BMF_STANDARD) {
														c2p.bmap = pi->BitMap;
														c2p.width = pi->Image->Width;
														c2p.height = pi->Image->Height;
														c2p.chunkybuffer = chunky;
														ChunkyToPlanarAsm(&c2p);
													} else if(GfxBase->LibNode.lib_Version >= 40) {
														WriteChunkyPixels(&rp, 0, 0,
																							phdr->Width, phdr->Height, chunky, phdr->Width);
													}
											}
									}
							}
						FreeVec(raw);
					}
			}
	}
}

Class	*MakePrayerImageClass(void)
{
Class *PIClass;
if(PIClass = MakeClass(NULL, IMAGECLASS, NULL, sizeof(struct PIData), 0)) {
		PIClass->cl_Dispatcher.h_Entry = (LONG (*)()) &PI_Dispatcher;
	}
return(PIClass);
}

void	 FreePrayerImageClass(Class *ImageClass)
{
if(ImageClass)
	FreeClass(ImageClass);
}

static UBYTE *RemapImage(UBYTE *raw, struct PIData *pi)
{
UBYTE	pens[256];
ULONG	i;
struct	PrayerPalette *pal = (APTR) raw;

memset(&pi->PenAlloc, 0, sizeof(pi->PenAlloc));

for(i = 0; i < pi->Pif->NumColors; i++) {
		if(pi->DestScr) {

				pens[i] = ObtainBestPen(pi->DestScr->ViewPort.ColorMap, pal->R << 24, pal->G << 24, pal->B << 24,
														OBP_Precision, pi->Precision,
														TAG_DONE);
				pi->PenAlloc[pens[i]]++; // mark this pen as allocated.		
			}
		pal++;
	}

i = pi->Image->Width * pi->Image->Height;
raw = (APTR)pal;
while(i--) { *raw = pens[*raw];raw++;}
return((APTR)pal);
}
