/*
 * =-=-=-=-=-=-=-=-=-=-=-  NOTICE  =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * The following is the license that applies to this copy of the Prayer2
 * software.  For a license to use the Prayer2 software under conditions
 * other than those described here, please contact MindNever Software at
 * following addresses:
 *
 * MindNever Software <info@mindnever.org>
 * MindNever Software +381 63 550-161 or +381 64 220-7115
 * 
 * =-=-=-=-=-=-=-=-=-=-  COPYRIGHT  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Prayer2 is:
 *   Copyright (c) 1999-2001 by Tamara Cetinjski <tamara@mindnever.org>
 *   Copyright (c) 2001-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *   Copyright (c) 1999-2002 by MindNever Software Group <info@mindnever.org>
 *
 * Tamara Cetinjski is
 *   Copyright (c) 1999-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *
 *
 * =-=-=-=-=-=-=-=-=-=-  LICENSE  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Redistributions in any form must be accompanied by information on
 *    how to obtain complete source code for the Prayer2 software and any
 *    accompanying software that uses the Prayer2 software.  The source code
 *    must either be included in the distribution or be available for no
 *    more than the cost of distribution plus a nominal fee, and must be
 *    freely redistributable under reasonable conditions.  For an
 *    executable file, complete source code means the source code for all
 *    modules it contains.  It does not include source code for modules or
 *    files that typically accompany the major components of the operating
 *    system on which the executable file runs.
 *
 * =-=-=-=-=-=-=-=-=-=-  NO WARRANTY =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
 *
 * THIS SOFTWARE IS PROVIDED BY MINDNEVER SOFTWARE ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
 * NON-INFRINGEMENT, ARE DISCLAIMED.  IN NO EVENT SHALL MINDNEVER SOFTWARE
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, HAIR, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
/*
 * PrayerListClass.c
 *
 */

#define USE_BUILTIN_MATH
#define INTUI_V36_NAMES_ONLY
#define __USE_SYSBASE

#include <exec/types.h>
#include <exec/libraries.h>
#include <exec/execbase.h>
#include <intuition/intuition.h>
#include <intuition/intuitionbase.h>
#include <intuition/classes.h>
#include <intuition/gadgetclass.h>
#include <graphics/gfxbase.h>
#include <graphics/gfxmacros.h>
#include	<proto/dos.h>
#include <proto/intuition.h>
#include <proto/graphics.h>
#include <proto/layers.h>
#include <proto/utility.h>
#include	<proto/exec.h>
#include	<exec/ports.h>
#include	<stdarg.h>

#include	<stdlib.h>
#include	<string.h>
#include	<stdio.h>

#include "classhelp.h"

#include "drag.h"

#ifndef max
#define max(a,b) ((a) > (b) ? (a) : (b))
#endif

#include "PrayerListClass.h"

void debugf(char *ctl,...);

#define	REG(x) register __ ## x

enum {
		PEN_N,
		PEN_S,
		PEN_H
	};

struct PLData {
		struct List *Items;
		LONG	Top;
		LONG	Visible;
		LONG	Selected;
		LONG	Total;
		LONG	Highlighted;
		LONG	OldTop;
		LONG	OldSelected;
		LONG	OldHighlighted;
		WORD	ItemHeight;
		LONG	MaxPen;
		BOOL	NoDraw;
		// usefull pointers
		APTR	TopPtr;
		APTR	SelectedPtr;
		APTR	HighlightedPtr;
		APTR	OldSelectedPtr;
		APTR	OldHighlightedPtr;
		APTR	OldTopPtr;
		struct	TextFont *Font;
		struct	Rectangle	Bounds;
		ULONG	BgPen[3], FgPen[3];

		BOB *Bob;

		struct DragMark DragMark;

		int term;
	};


static void		PL_GMRender		(Class *cl, struct Gadget *g, struct gpRender *msg);
static ULONG	PL_GMGoActive	(Class *cl, struct Gadget *g, struct gpInput *msg);
static ULONG	PL_GMHandleInput(Class *cl, struct Gadget *g, struct gpInput *msg);
static void		PL_GMGoInactive	(Class *cl, struct Gadget *g, struct gpGoInactive *msg);
static void		PL_GMLayout		(Class *cl, struct Gadget *g, struct gpLayout *msg);
static ULONG	PL_OMSet		(Class *cl, struct Gadget *g, struct opUpdate *msg);
static ULONG	PL_OMGet		(Class *cl, struct Gadget *g, struct opGet *msg);
static ULONG	PL_OMNew		(Class *cl, struct Gadget *g, struct opSet *msg);
static void		PL_OMDispose	(Class *cl, struct Gadget *g, Msg msg);

static struct Node *GetItem(struct List *list, ULONG num);

static int GetColor(struct PLData *pl, struct Node *n);

static char *method_name(int id)
{
static char ab[64];

switch(id) {
		case	OM_NEW: return("OM_NEW");
		case	OM_DISPOSE: return("OM_DISPOSE");
		case	OM_SET: return("OM_SET");
		case	OM_GET: return("OM_GET");
		case	OM_ADDTAIL: return("OM_ADDTAIL");
		case	OM_REMOVE: return("OM_REMOVE");
		case	OM_NOTIFY: return("OM_NOTIFY");
		case	OM_UPDATE: return("OM_UPDATE");
		case	OM_ADDMEMBER: return("OM_ADDMEMBER");
		case	OM_REMMEMBER: return("OM_REMMEMBER");
		case	GM_HITTEST: return("GM_HITTEST");
		case	GM_RENDER: return("GM_RENDER");
		case	GM_GOACTIVE: return("GM_GOACTIVE");
		case	GM_HANDLEINPUT: return("GM_HANDLEINPUT");
		case	GM_GOINACTIVE: return("GM_GOINACTIVE");
		case	GM_HELPTEST: return("GM_HELPTEST");
		case	GM_LAYOUT: return("GM_LAYOUT");
		case	GM_DOMAIN: return("GM_DOMAIN");
		case	GM_KEYTEST: return("GM_KEYTEST");
		case	GM_KEYGOACTIVE: return("GM_KEYGOACTIVE");
		case	GM_KEYGOINACTIVE: return("GM_KEYGOINACTIVE");
	}

sprintf(ab, "%ld", id);
return(ab);
}

#ifdef __MORPHOS__
static ULONG __saveds PL_Dispatcher_Native(VOID)
{
Class *cl = REG_A0;
struct Gadget *g = REG_A2;
Msg msg = REG_A1;

#else
static ULONG __saveds PL_Dispatcher(Class *cl asm("a0"), struct Gadget *g asm("a2"), Msg msg asm("a1"))
{

#endif
ULONG rc;

switch(msg->MethodID) {
		case	GM_RENDER:
			PL_GMRender(cl, g, (APTR) msg);
			rc = TRUE;
			break;

		case	GM_GOACTIVE:
			rc = PL_GMGoActive(cl, g, (APTR) msg);
			break;

		case	GM_HANDLEINPUT:
			rc = PL_GMHandleInput(cl, g, (APTR) msg);
			break;

		case	GM_GOINACTIVE:
			PL_GMGoInactive(cl, g, (struct gpGoInactive *)msg);
			rc = TRUE;
			break;

		case	GM_LAYOUT:
			PL_GMLayout(cl, g, (APTR) msg);
			rc = TRUE;
			break;

		case	OM_SET:
			rc = PL_OMSet(cl, g, (APTR) msg);
			break;

		case	OM_UPDATE:
			rc = PL_OMSet(cl, g, (APTR) msg);
			break;

		case	OM_GET:
			rc = PL_OMGet(cl, g, (APTR) msg);
			break;

		case	OM_NEW:
			rc = PL_OMNew(cl, g, (APTR) msg);
			break;

		case	OM_DISPOSE:
			PL_OMDispose(cl, g, msg);
			rc = TRUE;
			break;

		default:
			rc = DoSuperMethodA(cl, (APTR) g, msg);
			break;
	}

return(rc);
}

#ifdef __MORPHOS__
struct EmulLibEntry PL_Dispatcher = {
		TRAP_LIB,
		0,
		(void (*)(void))PL_Dispatcher_Native
	};
#endif


static void DrawItem(struct PLData *pl, struct RastPort *rp, struct Node *node, struct Rectangle *rect)
{
int i, col;
struct TextExtent textent;

col = GetColor(pl, node);

SetABPenDrMd(rp, pl->FgPen[col], pl->BgPen[col], JAM2);

Move(rp, rect->MinX, rect->MinY + rp->Font->tf_Baseline);

i = TextFit(rp, node->ln_Name, strlen(node->ln_Name), &textent, NULL, 1, 
				rect->MaxX - rect->MinX + 1,
				rect->MaxY - rect->MinY + 1);

Text(rp, node->ln_Name, i);

SetAPen(rp, pl->BgPen[col]);
RectFill(rp, rect->MinX, rect->MinY + rp->Font->tf_YSize, rect->MaxX - 1, rect->MaxY - 2);
RectFill(rp, rp->cp_x, rect->MinY, rect->MaxX - 1, rect->MaxY - 2);
}

static int GetColor(struct PLData *pl, struct Node *node)
{
if(node == pl->SelectedPtr)
	return(PEN_S);
if(node == pl->HighlightedPtr)
	return(PEN_H);

return(PEN_N);
}


static void DrawListItem(struct RastPort *rp, struct PLData *pl, ULONG itemno, struct Node *node)
{
struct Rectangle rect;

if(node && (itemno >= pl->Top) && (itemno < pl->Visible + pl->Top)) {
		if(itemno == pl->Selected) {
				pl->OldSelected = pl->Selected;
				pl->OldSelectedPtr = pl->SelectedPtr;
			}	else if(itemno == pl->Highlighted) {
				pl->OldHighlighted = pl->Highlighted;
				pl->OldHighlightedPtr = pl->HighlightedPtr;
			}

		rect.MinY = pl->Bounds.MinY + ((itemno - pl->Top) * pl->ItemHeight);
		rect.MaxY = rect.MinY + pl->ItemHeight + 1;

		rect.MinX = pl->Bounds.MinX;
		rect.MaxX = pl->Bounds.MaxX;

		DrawItem(pl, rp, node, &rect);
	}
}

static void RedrawAll(struct RastPort *rp, struct PLData *pl)
{
int t = pl->Top, v = pl->Visible;
struct	Node *node = pl->TopPtr;

if(node) {
		while(node->ln_Succ && v--) {
				DrawListItem(rp, pl, t++, node);
				node = node->ln_Succ;
			}
	}
}

#define SMOOTH_STEP	4

static void SmoothScroll(struct RastPort *rp, UWORD SrcX, UWORD SrcY,
																							WORD XOffset, WORD YOffset,
																							UWORD	Width, UWORD Height)
{
#if 0
int i, ia, y = 0;

ia = YOffset / SMOOTH_STEP;

for(i = 0; i < SMOOTH_STEP; i++) {
		ClipBlit(rp, SrcX, SrcY + y, 
							rp, SrcX, SrcY + y + ia,
							Width, Height, 0x0c0);

		y += ia;
		if(((y > 0) && (y > YOffset)) || ((y < 0) && (y < YOffset)))
			y = YOffset;
	}
#else
ClipBlit(rp, SrcX, SrcY, rp, SrcX + XOffset, SrcY + YOffset, Width, Height, 0x0c0);
#endif
}

void DrawPrayerListItem(Class *c, struct Gadget *g, struct RastPort *rp, struct Rectangle *rect, struct Node *node)
{
struct PLData *pl = INST_DATA(c, g);

DrawItem(pl, rp, node, rect);
}
			
static void		PL_GMRender(Class *cl, struct Gadget *g, struct gpRender *msg)
{
struct PLData *pl = INST_DATA(cl, g);
struct RastPort *rp = msg->gpr_RPort;
APTR oldfont = NULL;

if(pl->NoDraw)
	return;

PL_GMLayout(cl, g, (struct gpLayout *)msg);

SetMaxPen(rp, pl->MaxPen);

if(pl->Items && pl->Visible) {
		if(rp->Font != pl->Font) {
				oldfont = rp->Font;
				SetFont(rp, pl->Font);
			}

		switch(msg->gpr_Redraw) {
				case	GREDRAW_UPDATE: {
					int d, t, sy, ly;
					struct Node *node;

					if(d = pl->Top - pl->OldTop) {
							t = abs(d);
							if(t > (pl->Visible - 4)) {
									RedrawAll(rp, pl);
								} else {
									sy = (d * pl->ItemHeight);
									ly = (pl->Visible - t) * pl->ItemHeight;

									SmoothScroll(rp, pl->Bounds.MinX, (d > 0) ? pl->Bounds.MinY + sy : pl->Bounds.MinY,
																	0, -sy,
																	pl->Bounds.MaxX - pl->Bounds.MinX, ly);

									if(d > 0) {
											t = pl->Top + (pl->Visible - d);
										} else {
											t = pl->Top;
											d = -d;
										}

									node = GetItem(pl->Items, t);

									DrawListItem(rp, pl, pl->OldSelected, pl->OldSelectedPtr);
									DrawListItem(rp, pl, pl->OldHighlighted, pl->OldHighlightedPtr);

									while(node->ln_Succ && d--) {
											DrawListItem(rp, pl, t++, node);
											node = node->ln_Succ;
										}
								}

							SetMaxPen(rp, -1);

							pl->OldTop = pl->Top;
							pl->OldTopPtr = pl->TopPtr;
						}

					} break;

				case	GREDRAW_TOGGLE:

					if(pl->Selected != pl->OldSelected) {
							DrawListItem(rp, pl, pl->OldSelected, pl->OldSelectedPtr);
							DrawListItem(rp, pl, pl->Selected, pl->SelectedPtr);

							if(pl->Bob) {
									SetFont(&pl->Bob->rp, pl->Font);
									DrawItem(pl, &pl->Bob->rp, pl->SelectedPtr, &pl->Bob->rect);
									InitMasks(&pl->Bob->vs);
								}
						}

					if(pl->Highlighted != pl->OldHighlighted) {
							DrawListItem(rp, pl, pl->OldHighlighted, pl->OldHighlightedPtr);
							DrawListItem(rp, pl, pl->Highlighted, pl->HighlightedPtr);
						}

					break;

				case	GREDRAW_REDRAW:

					RedrawAll(rp, pl);
					pl->OldTop = pl->Top;
					pl->OldTopPtr = pl->TopPtr;

					if(pl->Total < pl->Visible) {
							SetAPen(rp, pl->BgPen[PEN_N]);
							RectFill(rp, pl->Bounds.MinX, pl->Bounds.MinY + (pl->Total * pl->ItemHeight) + 1,
													pl->Bounds.MaxX, pl->Bounds.MaxY);
						}
					break;
			}

		if(oldfont)
			SetFont(rp, oldfont);
	} else if(msg->gpr_Redraw == GREDRAW_REDRAW) {
		SetAPen(rp, pl->BgPen[PEN_N]);
		RectFill(rp, pl->Bounds.MinX, pl->Bounds.MinY, pl->Bounds.MaxX, pl->Bounds.MaxY);
	}

	SetMaxPen(rp, -1);
}

static ULONG	PL_GMGoActive	(Class *cl, struct Gadget *g, struct gpInput *msg)
{
struct PLData *pl = INST_DATA(cl, g);

if(pl->Items == NULL)
	return(GMR_NOREUSE);

g->Flags |= GFLG_SELECTED;

if(msg->gpi_IEvent == NULL)
	return(GMR_NOREUSE);

return(PL_GMHandleInput(cl, g, msg));
}

static ULONG	PL_GMHandleInput(Class *cl, struct Gadget *g, struct gpInput *msg)
{
struct PLData		*pl = INST_DATA (cl, g);
struct InputEvent *ie = msg->gpi_IEvent;
ULONG result = GMR_MEACTIVE;
LONG dm;
struct Node *node;

pl->term = PLE_CANCEL;

switch(ie->ie_Class) {
		case	IECLASS_RAWMOUSE: {
				LONG selected, d;
				switch(ie->ie_Code) {
						case	SELECTDOWN: // GMR_MEACTIVE
							if((msg->gpi_Mouse.X < 0) ||
								(msg->gpi_Mouse.X >= g->Width) ||
								(msg->gpi_Mouse.Y < 0) ||
								(msg->gpi_Mouse.Y >= g->Height))
								return(GMR_REUSE);

							d = (msg->gpi_Mouse.Y / pl->ItemHeight);
							selected = pl->Top + d;
							if(selected >= pl->Total)
								break;

							UpdateAttrs((APTR) g, msg->gpi_GInfo, 0,
																PLA_Selected, selected,
																TAG_DONE);

							NotifyAttrs((APTR) g, msg->gpi_GInfo, 0,
																PLA_Event, PLE_DOWN,
																PLA_Bob, pl->Bob,
																TAG_DONE);
							break;

						case	SELECTUP: // GMR_REUSE

							if((msg->gpi_Mouse.X < 0) ||
								(msg->gpi_Mouse.X >= g->Width))
								return(GMR_REUSE);

							pl->term = PLE_UP;

							result = GMR_REUSE;
							break;

						case	MENUDOWN: // Eat this !

							break;

						case	MENUUP:
							result = GMR_NOREUSE;
							break;

						default:

							pl->DragMark.Scroll = 0;

							if((msg->gpi_Mouse.X >= 0) &&
								(msg->gpi_Mouse.X < g->Width)) {
									dm = ((msg->gpi_Mouse.Y + (pl->ItemHeight / 2)) / pl->ItemHeight);
									if(dm > pl->Total)
										dm = pl->Total;
									if(dm > pl->Visible) {
											dm = pl->Visible;
											pl->DragMark.Scroll = 1;
										} else if(dm < 0) {
											dm = 0;
											pl->DragMark.Scroll = -1;
										}
								} else {
									dm = -1;
								}

							pl->DragMark.Nr = dm;
							pl->DragMark.This = pl->SelectedPtr;

							if(dm >= 0) {
									pl->DragMark.Rect.MinY = 
										pl->DragMark.Rect.MaxY = g->TopEdge + dm * pl->ItemHeight - 1;
									
									for(node = pl->TopPtr; (dm > 0) && (node->ln_Succ);dm--) {
											node = node->ln_Succ;
										}

									pl->DragMark.InsAfter = node->ln_Pred;
									if(pl->DragMark.InsAfter == pl->DragMark.This) {
											pl->DragMark.InsAfter = pl->DragMark.This->ln_Pred;
										}
								} else {
									 pl->DragMark.InsAfter = NULL;
								}

							NotifyAttrs((APTR) g, msg->gpi_GInfo, 0, 
																PLA_Event, PLE_MOVE,
																PLA_DragMark, &pl->DragMark,
																PLA_Bob, pl->Bob,
																TAG_DONE);
					}
			} break;
	}

return(result);
}

static void		PL_GMGoInactive	(Class *cl, struct Gadget *g, struct gpGoInactive *msg)
{
struct PLData *pl = INST_DATA(cl, g);

if(pl->term != PLE_IGNORE) {
		NotifyAttrs((APTR) g, msg->gpgi_GInfo, 0, 
													PLA_Event, pl->term,
													PLA_Bob, pl->Bob,
													PLA_Selected, pl->Selected,
													PLA_SelectedPtr, pl->SelectedPtr,
													PLA_Top, pl->Top,
													TAG_DONE);
		pl->term = PLE_IGNORE;
	}
g->Flags &= ~GFLG_SELECTED;
}

static void		PL_GMLayout(Class *cl, struct Gadget *g, struct gpLayout *msg)
{
struct PLData *pl = INST_DATA(cl, g);
LONG visible;

pl->NoDraw = TRUE;

pl->Bounds.MinX = g->LeftEdge;
pl->Bounds.MinY = g->TopEdge;
pl->Bounds.MaxX = g->LeftEdge + g->Width - 1;
pl->Bounds.MaxY = g->TopEdge + g->Height - 1;

pl->DragMark.Rect.MinX = pl->Bounds.MinX + 1;
pl->DragMark.Rect.MaxX = pl->Bounds.MaxX - 1;

if(pl->Font == NULL)
	pl->Font = msg->gpl_GInfo->gi_DrInfo->dri_Font;

pl->ItemHeight = pl->Font->tf_YSize + 1;

visible = g->Height / pl->ItemHeight;

UpdateAttrs((APTR) g, msg->gpl_GInfo, 0,
								PLA_Visible, visible,
								TAG_DONE);

pl->NoDraw = FALSE;
}

#define PLF_DO_SUPER_METHOD	(1L << 0)
#define	PLF_REDRAW	(1L << 1)
#define	PLF_SCROLL	(1L << 2)
#define PLF_TOGGLE	(1L << 3)
#define	PLF_NOTIFY	(1L << 4)
#define	PLF_NOTIFYALL (PLF_SCROLL|PLF_TOGGLE|PLF_NOTIFY)

static ULONG attr_set(struct PLData *pl, struct TagItem *ti)
{
struct TagItem *tstate = ti;
ULONG action = 0L;
int i;

while(ti = NextTagItem(&tstate))
	switch(ti->ti_Tag) {
			case	PLA_Bob:
				pl->Bob = (APTR) ti->ti_Data;
				break;
			case	PLA_Top:
				if((pl->Top != ti->ti_Data) && pl->Items) {
						LONG newtop = ti->ti_Data, maxtop;
						maxtop = pl->Total - pl->Visible;
						if(maxtop < 0)
							maxtop = 0;
						if(newtop < 0)
							newtop = 0;
						if(newtop > maxtop)
							newtop = maxtop;

						pl->Top = newtop;
						pl->TopPtr = GetItem(pl->Items, newtop);
						action |= (PLF_SCROLL|PLF_NOTIFY);
					}
				break;
			case	PLA_Selected:
				if(pl->Items)	{
						LONG	newselected = ti->ti_Data;

						if(newselected >= pl->Total)
							newselected = pl->Total - 1;

						if(newselected != pl->Selected) {
								action |= (PLF_TOGGLE|PLF_NOTIFY);

								pl->Selected = newselected;

								if(newselected < 0)
									pl->SelectedPtr = NULL;
								else
									pl->SelectedPtr = GetItem(pl->Items, newselected); // !!!!!!!!!!!!!!
							}
					}
				break;
			case	PLA_Highlighted:
				if(pl->Items) {
						LONG	newhl = ti->ti_Data;

						if(newhl >= pl->Total)
							newhl = pl->Total - 1;
						if(newhl != pl->Highlighted) {
								action |= (PLF_TOGGLE|PLF_NOTIFY);
								pl->Highlighted = newhl;
								if(newhl < 0)
									pl->HighlightedPtr = NULL;
								else
									pl->HighlightedPtr = GetItem(pl->Items, newhl); // !!!!!!!!!!!!!!
							}
					}
				break;
			case	PLA_MakeVisible: {
					LONG itemno = ti->ti_Data;
					if(itemno < 0)
						itemno = 0;
					if(itemno >= pl->Total)
						itemno = pl->Total + 1;

					if(itemno < pl->Top) {
							pl->Top = itemno;
							pl->TopPtr = GetItem(pl->Items, itemno);
							action |= (PLF_SCROLL | PLF_NOTIFY);
						} else if(itemno >= pl->Top + pl->Visible) {
							pl->Top = itemno - pl->Visible + 1;
							pl->TopPtr = GetItem(pl->Items, pl->Top);
							action |= (PLF_SCROLL | PLF_NOTIFY);
						}
				} break;
			case	PLA_Items:
				if((ti->ti_Data == ~0) || (ti->ti_Data == 0L)) {
						pl->Items = NULL;
					} else {
						APTR sp;
						LONG i = 0, sn;
						struct Node *node;

						pl->Items = (APTR) ti->ti_Data;
						action |= (PLF_REDRAW | PLF_NOTIFYALL);

						sp = pl->SelectedPtr;
						sn = pl->Selected;

						pl->Selected = pl->Highlighted = pl->OldHighlighted = pl->OldSelected = -1;
						pl->OldHighlightedPtr = pl->SelectedPtr = pl->OldSelectedPtr = NULL;


						for(node = pl->Items->lh_Head; node->ln_Succ; node = node->ln_Succ) {
								if((sp == node) || ((sn == i) && (pl->Selected < 0))) {
										pl->SelectedPtr = node;
										pl->Selected = i;
									}

								if((APTR) node == pl->HighlightedPtr)
									pl->Highlighted = i;

								i++;
							}

						if(pl->Total = i) {
								if(pl->Top + pl->Visible >= i) {
										pl->Top = i - pl->Visible;
										if(pl->Top < 0)
											pl->Top = 0;
									}

								pl->TopPtr = GetItem(pl->Items, pl->Top);

								if((pl->SelectedPtr == NULL) && (pl->Selected >= 0)) {
										pl->Selected = i - 1;
										pl->SelectedPtr = node->ln_Pred;
									}
								if(pl->Highlighted < 0)
									pl->HighlightedPtr = NULL;
							} else {
								pl->Items = NULL;
								pl->Top = 0;
								pl->TopPtr = NULL;
								pl->Selected = pl->Highlighted = -1;
								pl->SelectedPtr = pl->HighlightedPtr = NULL;
							}
					}
				break;
			case	PLA_Visible:
				pl->Visible = ti->ti_Data;
				if(pl->Top + pl->Visible >= pl->Total) {
						pl->Top = pl->Total - pl->Visible;
						if(pl->Top < 0)
							pl->Top = 0;
						action |= PLF_SCROLL;
					}
				break;
			case	PLA_BgPen:
				pl->BgPen[PEN_N] = ti->ti_Data;
				action |= (PLF_REDRAW);
				break;
			case	PLA_FgPen:
				pl->FgPen[PEN_N] = ti->ti_Data;
				action |= (PLF_REDRAW);
				break;
			case	PLA_BgPenS:
				pl->BgPen[PEN_S] = ti->ti_Data;
				action |= (PLF_REDRAW);
				break;
			case	PLA_FgPenS:
				pl->FgPen[PEN_S] = ti->ti_Data;
				action |= (PLF_REDRAW);
				break;
			case	PLA_BgPenH:
				pl->BgPen[PEN_H] = ti->ti_Data;
				action |= (PLF_REDRAW);
				break;
			case	PLA_FgPenH:
				pl->FgPen[PEN_H] = ti->ti_Data;
				action |= (PLF_REDRAW);
				break;
			case	PLA_TextFont:
				pl->Font = (APTR) ti->ti_Data;
				action |= (PLF_REDRAW);
				break;
			default:
				action |= PLF_DO_SUPER_METHOD;
				break;
		}

for(pl->MaxPen = 0, i = 0; i < 3; i++) {
		if(pl->MaxPen < pl->FgPen[i])
			pl->MaxPen = pl->FgPen[i];
		if(pl->MaxPen < pl->BgPen[i])
			pl->MaxPen = pl->BgPen[i];
	}

pl->DragMark.FgPen = pl->FgPen[PEN_N];
pl->DragMark.BgPen = pl->BgPen[PEN_N];

return(action);
}

static ULONG	PL_OMSet(Class *cl, struct Gadget *g, struct opUpdate *msg)
{
struct PLData *pl = INST_DATA(cl, g);

ULONG	action = 0, result;

action = attr_set(pl, msg->opu_AttrList);

if(action & PLF_DO_SUPER_METHOD)
	result = (DoSuperMethodA(cl, (APTR) g, (Msg) msg));
else
	result = TRUE;

if((action & (PLF_REDRAW | PLF_SCROLL | PLF_TOGGLE)) &&
	msg->opu_GInfo && !pl->NoDraw) {
			struct RastPort *rp;
			if(rp = ObtainGIRPort(msg->opu_GInfo)) {
					if(action & PLF_REDRAW) {
							DoMethod((APTR) g, GM_RENDER, msg->opu_GInfo, rp, GREDRAW_REDRAW);
						} else {
							if(action & PLF_SCROLL)
								DoMethod((APTR) g, GM_RENDER, msg->opu_GInfo, rp, GREDRAW_UPDATE);
							if(action & PLF_TOGGLE)
								DoMethod((APTR) g, GM_RENDER, msg->opu_GInfo, rp, GREDRAW_TOGGLE);
						}
					ReleaseGIRPort(rp);
				}
		}

if(action & PLF_NOTIFY) {
		ULONG tags[20];
		int i = 0;

		if(action & PLF_REDRAW) {
				tags[i++] = PLA_Visible; tags[i++] = pl->Visible;
				tags[i++] = PLA_Total; tags[i++] = pl->Total;
			}

		if(action & PLF_SCROLL) {
				tags[i++] = PLA_Top; tags[i++] = pl->Top;
			}

		if(action & PLF_TOGGLE) {
				tags[i++] = PLA_SelectedPtr; tags[i++] = (ULONG) pl->SelectedPtr;
				tags[i++] = PLA_Selected; tags[i++] = pl->Selected;
				tags[i++] = PLA_Highlighted; tags[i++] = pl->Highlighted;

			}

		tags[i] = TAG_DONE;

		DoMethod((APTR) g, OM_NOTIFY, tags, msg->opu_GInfo,
				(msg->MethodID == OM_UPDATE) ? msg->opu_Flags : 0);
	}
return(result);
}

static ULONG	PL_OMGet		(Class *cl, struct Gadget *g, struct opGet *msg)
{
struct PLData *pl = INST_DATA(cl, g);

switch(msg->opg_AttrID) {
		case	PLA_ItemWidth:
			*msg->opg_Storage = (ULONG) (pl->Bounds.MaxX - pl->Bounds.MinX);
			return(TRUE);
		case	PLA_ItemHeight:
			*msg->opg_Storage = (ULONG) pl->ItemHeight;
			return(TRUE);

		case	PLA_Highlighted:
			*msg->opg_Storage = (ULONG) pl->Highlighted;
			return(TRUE);
		case	PLA_HighlightedPtr:
			*msg->opg_Storage = (ULONG) pl->HighlightedPtr;
			return(TRUE);
		case	PLA_Selected:
			*msg->opg_Storage = (ULONG) pl->Selected;
			return(TRUE);
		case	PLA_Top:
			*msg->opg_Storage = (ULONG) pl->Top;
			return(TRUE);
		case	PLA_Total:
			*msg->opg_Storage = (ULONG) pl->Total;
			return(TRUE);
		case	PLA_Items:
			*msg->opg_Storage = (ULONG) pl->Items;
			return(TRUE);
		case	PLA_Visible:
			*msg->opg_Storage = (ULONG) pl->Visible;
			return(TRUE);
		case	PLA_SelectedPtr:
			*msg->opg_Storage = (ULONG) pl->SelectedPtr;
			return(TRUE);
	}
return(DoSuperMethodA(cl, (APTR) g, (Msg) msg));
}

static ULONG	PL_OMNew		(Class *cl, struct Gadget *g, struct opSet *msg)
{
struct PLData *pl;

if(g = (APTR) DoSuperMethodA(cl, (APTR) g, (Msg) msg)) {
		if(g->Flags & GFLG_EXTENDED)
			((struct ExtGadget *)g)->MoreFlags |= GMORE_SCROLLRASTER;
		if(pl = INST_DATA(cl, (APTR) g)) {
				memset(pl, 0, sizeof(*pl));

				pl->OldHighlighted = pl->Highlighted = pl->OldSelected = pl->Selected = -1;
				pl->term = PLE_IGNORE;

				attr_set(pl, msg->ops_AttrList);
			}
	}
return((ULONG) g);
}

static void		PL_OMDispose	(Class *cl, struct Gadget *g, Msg msg)
{
DoSuperMethodA(cl, (APTR) g, (Msg) msg);
}


static struct Node *GetItem(struct List *list, ULONG num)
{
struct Node *node = NULL;
if(list) {
		for(node = list->lh_Head; node->ln_Succ && num--; node = node->ln_Succ);
	}
if(node && node->ln_Succ)
	return(node);

return(NULL);
}

Class *MakePrayerListClass(void)
{
Class *PLClass;
if(PLClass = MakeClass(NULL, GADGETCLASS, NULL, sizeof(struct PLData), 0))
	PLClass->cl_Dispatcher.h_Entry = (LONG (*)()) &PL_Dispatcher;

return(PLClass);
}

void FreePrayerListClass(Class *PLClass)
{
if(PLClass)
	FreeClass(PLClass);
}



//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////


void debugf(char *ctl,...)
{
struct	MsgPort	*mp;
struct	Message	*mm;
char	*prt;
va_list args;


if(mm = AllocVec(4096, 0)) {
		prt = (char *) (mm + 1);

		va_start(args, ctl);
		RawDoFmt(ctl, args, (void (*)()) "\x16\xC0\x4E\x75", prt);
		va_end(args);
		mm->mn_ReplyPort = (APTR)SysBase->ThisTask;
		Forbid();
		if(mp = FindPort("LOG_AUTOFREE"))
			PutMsg(mp, mm);
		Permit();
		if(mp == NULL)
			FreeVec(mm);
	}
}
