/*
 * =-=-=-=-=-=-=-=-=-=-=-  NOTICE  =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * The following is the license that applies to this copy of the Prayer2
 * software.  For a license to use the Prayer2 software under conditions
 * other than those described here, please contact MindNever Software at
 * following addresses:
 *
 * MindNever Software <info@mindnever.org>
 * MindNever Software +381 63 550-161 or +381 64 220-7115
 * 
 * =-=-=-=-=-=-=-=-=-=-  COPYRIGHT  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Prayer2 is:
 *   Copyright (c) 1999-2001 by Tamara Cetinjski <tamara@mindnever.org>
 *   Copyright (c) 2001-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *   Copyright (c) 1999-2002 by MindNever Software Group <info@mindnever.org>
 *
 * Tamara Cetinjski is
 *   Copyright (c) 1999-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *
 *
 * =-=-=-=-=-=-=-=-=-=-  LICENSE  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Redistributions in any form must be accompanied by information on
 *    how to obtain complete source code for the Prayer2 software and any
 *    accompanying software that uses the Prayer2 software.  The source code
 *    must either be included in the distribution or be available for no
 *    more than the cost of distribution plus a nominal fee, and must be
 *    freely redistributable under reasonable conditions.  For an
 *    executable file, complete source code means the source code for all
 *    modules it contains.  It does not include source code for modules or
 *    files that typically accompany the major components of the operating
 *    system on which the executable file runs.
 *
 * =-=-=-=-=-=-=-=-=-=-  NO WARRANTY =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
 *
 * THIS SOFTWARE IS PROVIDED BY MINDNEVER SOFTWARE ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
 * NON-INFRINGEMENT, ARE DISCLAIMED.  IN NO EVENT SHALL MINDNEVER SOFTWARE
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, HAIR, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
#define USE_BUILTIN_MATH
#define INTUI_V36_NAMES_ONLY
#define __USE_SYSBASE

#include <exec/types.h>
#include <intuition/intuition.h>
#include <intuition/classes.h>
#include <intuition/gadgetclass.h>
#include <intuition/imageclass.h>

#include <proto/intuition.h>
#include <proto/utility.h>
#include <proto/graphics.h>
#include <utility/hooks.h>
#include <proto/exec.h>


#include	<exec/ports.h>
#include	<exec/tasks.h>
#include	<exec/execbase.h>

#include	<string.h>
#include	<stdarg.h>

#include "PrayerPropClass.h"
#include "classhelp.h"
#include "opt68k.h"


struct PPData {
	LONG	Top;
	LONG	Total;
	LONG	Visible;
	LONG	Freedom;
	BOOL	Dragging;
	UWORD	DragX, DragY;
	UWORD	MaxOffs;
	WORD	MaxTop;
	struct Rectangle	Knob, OKnob;
	UWORD	PixelOffset;
	BOOL	NoDraw;
	BOOL	OState;
	UWORD	BgPen;
	APTR	GenImage;
	struct	BitMap *BitMap;
};

/* Local function prototypes */

static void		PP_GMRender		(Class *cl, struct Gadget *g, struct gpRender *msg);
static ULONG	PP_GMGoActive	(Class *cl, struct Gadget *g, struct gpInput *msg);
static ULONG	PP_GMHandleInput(Class *cl, struct Gadget *g, struct gpInput *msg);
static void		PP_GMGoInactive	(Class *cl, struct Gadget *g, struct gpGoInactive *msg);
static ULONG	PP_OMSet		(Class *cl, struct Gadget *g, struct opUpdate *msg);
static ULONG	PP_OMGet		(Class *cl, struct Gadget *g, struct opGet *msg);
static ULONG	PP_OMNew		(Class *cl, struct Gadget *g, struct opSet *msg);
static void		PP_OMDispose	(Class *cl, struct Gadget *g, Msg msg);

#ifdef __MORPHOS__
static ULONG __saveds PP_Dispatcher_Native(void)
{
Class *cl = (Class *) REG_A0;
struct Gadget *g = (struct Gadget *) REG_A2;
Msg msg = (Msg) REG_A1;

#else
static ULONG __saveds PP_Dispatcher(Class *cl ASM("a0"), struct Gadget *g ASM("a2"), Msg msg ASM("a1"))
{
#endif

switch(msg->MethodID) {
		case	GM_RENDER: //
			PP_GMRender(cl, g, (APTR) msg);
			return(TRUE);

		case	GM_GOACTIVE: //
			return(PP_GMGoActive(cl, g, (APTR) msg));

		case	GM_HANDLEINPUT: //
			return(PP_GMHandleInput(cl, g, (APTR) msg));

		case	GM_GOINACTIVE: //
			PP_GMGoInactive(cl, g, (struct gpGoInactive *)msg);
			return(TRUE);

		case	OM_SET:	//
		case	OM_UPDATE:	//
			return(PP_OMSet(cl, g, (APTR) msg));

		case	OM_GET:	//
			return(PP_OMGet(cl, g, (APTR) msg));

		case	OM_NEW:	//
			return(PP_OMNew(cl, g, (APTR) msg));

		case	OM_DISPOSE:	//
			PP_OMDispose(cl, g, msg);
			return(TRUE);

		default:
			return(DoSuperMethodA(cl, (APTR) g, msg));
	}
}

#ifdef __MORPHOS__
struct EmulLibEntry PP_Dispatcher = {
		TRAP_LIB,
		0,
		(void (*)(void))PP_Dispatcher_Native
	};
#endif

static void		PP_OMDispose(Class *cl, struct Gadget *g, Msg msg)
{
struct PPData *pp = INST_DATA(cl, g);
if(pp->GenImage)
	DisposeObject(pp->GenImage);
DoSuperMethodA(cl, (APTR) g, (Msg) msg);
}

static void		PP_GMGoInactive(Class *cl, struct Gadget *g, struct gpGoInactive *msg)
{
g->Flags &= ~GFLG_SELECTED;
}

static void PP_SetKnob(struct Gadget *g, struct PPData *pp)
{

if(pp->Freedom == FREEVERT) {
		pp->Knob.MinX = 0;
		pp->Knob.MaxX = g->Width - 1;
		pp->Knob.MinY = pp->PixelOffset;
		pp->Knob.MaxY = pp->PixelOffset + ((struct Image *)g->GadgetRender)->Height - 1;
	} else {
		pp->Freedom = FREEHORIZ;
		pp->Knob.MinY = 0;
		pp->Knob.MaxY = g->Height - 1;
		pp->Knob.MinX = pp->PixelOffset;
		pp->Knob.MaxX = pp->PixelOffset + ((struct Image *)g->GadgetRender)->Width - 1;
	}
}

static void PP_Layout(struct Gadget *g, struct PPData *pp)
{
// first, normalize everything.
LONG mx;
UWORD	imgw, imgh, i, m;

if(pp->Total < 0)
	pp->Total = 0;

if(pp->Visible < 1)
	pp->Visible = 1;


// FIXME: pp->MaxTop can't be negative

pp->MaxTop = (pp->Total - pp->Visible);

if(pp->MaxTop < 0)
	pp->MaxTop = 0;

if(pp->Top > pp->MaxTop)
	pp->Top = pp->MaxTop;

if(pp->Top < 0)
	pp->Top = 0;


if(pp->Freedom == FREEVERT)
	m = g->Height;
else
	m = g->Width;

if(pp->Total)
	i = (m * pp->Visible) / pp->Total;
else
	i = m;

if(i > m)
	i = m;

if(i < 9)
	i = 9;

if(pp->Freedom == FREEVERT) {
		imgw = g->Width;
		imgh = i;
	} else {
		imgh = g->Height;
		imgw = i;
	}

if(g->GadgetRender == NULL) {
		pp->GenImage = NewObject(NULL, "frameiclass",
														IA_Width, imgw,
														IA_Height, imgh,
														IA_EdgesOnly, TRUE,
														TAG_DONE);
		g->GadgetRender = pp->GenImage;
	} else if(g->GadgetRender == pp->GenImage)
		SetAttrs(g->GadgetRender, IA_Width, imgw, IA_Height, imgh, TAG_DONE);

if(pp->Freedom == FREEVERT)
	pp->MaxOffs = g->Height - ((struct Image *)g->GadgetRender)->Height;
else
	pp->MaxOffs = g->Width - ((struct Image *)g->GadgetRender)->Width;

if(pp->Top)
	mx = (pp->MaxOffs * pp->Top) / pp->MaxTop;
else
	mx = 0;

pp->PixelOffset = mx;
PP_SetKnob(g, pp);
}


static ULONG	PP_GMHandleInput(Class *cl, struct Gadget *g, struct gpInput *msg)
{
struct PPData *pp = INST_DATA(cl, g);
struct InputEvent *ie = msg->gpi_IEvent;
APTR rp;

if(ie->ie_Class == IECLASS_RAWMOUSE) {
		switch(ie->ie_Code) {
				case	SELECTDOWN: // initial click

					if((msg->gpi_Mouse.X >= pp->Knob.MinX) && 
						(msg->gpi_Mouse.X <= pp->Knob.MaxX) && 
						(msg->gpi_Mouse.Y >= pp->Knob.MinY) && 
						(msg->gpi_Mouse.Y <= pp->Knob.MaxY)) {

							pp->Dragging = TRUE;

							pp->DragX = msg->gpi_Mouse.X - pp->Knob.MinX;
							pp->DragY = msg->gpi_Mouse.Y - pp->Knob.MinY;

							if(rp = ObtainGIRPort(msg->gpi_GInfo)) {
									DoMethod((APTR) g, GM_RENDER, msg->gpi_GInfo, rp, GREDRAW_UPDATE);
									ReleaseGIRPort(rp);
								}
						} else {
							LONG	newtop = pp->Top;
							if(pp->Freedom == FREEHORIZ) {
									if(msg->gpi_Mouse.X < pp->Knob.MinX) {
											newtop = pp->Top - (pp->Visible);
										} else if(msg->gpi_Mouse.X > pp->Knob.MaxX) {
											newtop = pp->Top + (pp->Visible);
										}
								} else {
									if(msg->gpi_Mouse.Y < pp->Knob.MinY) {
											newtop = pp->Top - (pp->Visible);
										} else if(msg->gpi_Mouse.Y > pp->Knob.MaxY) {
											newtop = pp->Top + (pp->Visible);
										}
								}
							if(pp->Top != newtop)
								UpdateAttrs((APTR) g, msg->gpi_GInfo, 0, PGA_Top, newtop, TAG_DONE);

							msg->gpi_Termination = &pp->Top;
							return(GMR_NOREUSE | GMR_VERIFY);
						}
					break;
				case	SELECTUP:
					pp->Dragging = FALSE;

					if(rp = ObtainGIRPort(msg->gpi_GInfo)) {
							DoMethod((APTR) g, GM_RENDER, msg->gpi_GInfo, rp, GREDRAW_UPDATE);
							ReleaseGIRPort(rp);
						}

					msg->gpi_Termination = &pp->Top;
					return(GMR_NOREUSE | GMR_VERIFY);
					break;
				default:
					if(pp->Dragging) {
							LONG	offs = 0, newtop;

							if(pp->Freedom == FREEHORIZ)
								offs = msg->gpi_Mouse.X - pp->DragX;
							else
								offs = msg->gpi_Mouse.Y - pp->DragY;

							if(offs < 0)
								offs = 0;
							if(offs > pp->MaxOffs)
								offs = pp->MaxOffs;

							if(offs != pp->PixelOffset) {
									pp->PixelOffset = offs;
									PP_SetKnob(g, pp);
									if(rp = ObtainGIRPort(msg->gpi_GInfo)) {
											DoMethod((APTR) g, GM_RENDER, msg->gpi_GInfo, rp, GREDRAW_UPDATE);
											ReleaseGIRPort(rp);
										}

									newtop = ((pp->Total - pp->Visible) * offs) / pp->MaxOffs;
									if(newtop > pp->MaxTop)
										newtop = pp->MaxTop;

									pp->NoDraw = TRUE;

									UpdateAttrs((APTR) g, msg->gpi_GInfo, 0, PGA_Top, newtop, TAG_DONE);

									pp->NoDraw = FALSE;
								}
						}
					break;
			}
	}
return(GMR_MEACTIVE);
}

static ULONG	PP_GMGoActive(Class *cl, struct Gadget *g, struct gpInput *msg)
{
//struct PPData *pp = INST_DATA(cl, g);

g->Flags |= GFLG_SELECTED;

if(msg->gpi_IEvent == NULL)
	return(GMR_MEACTIVE);

return(PP_GMHandleInput(cl, g, msg));
}

static void		PP_GMRender(Class *cl, struct Gadget *g, struct gpRender *msg)
{
struct PPData *pp = INST_DATA(cl, g);
struct RastPort *rp = msg->gpr_RPort;
struct Image *img;

if(!memcmp(&pp->OKnob, &pp->Knob, sizeof(pp->Knob))
	&& (pp->OState == pp->Dragging)
	&& (msg->gpr_Redraw != GREDRAW_REDRAW))
	return;

img = g->GadgetRender;

if(pp->Dragging && g->SelectRender)
	img = g->SelectRender;

if(pp->BgPen == 0xffff)
	pp->BgPen = msg->gpr_GInfo->gi_DrInfo->dri_Pens[BACKGROUNDPEN];

SetAPen(rp, pp->BgPen);

if(msg->gpr_Redraw == GREDRAW_REDRAW) {
/*
		SetAPen(rp, msg->gpr_GInfo->gi_DrInfo->dri_Pens[SHINEPEN]);
		Move(rp, g->LeftEdge, g->TopEdge);
		Draw(rp, g->LeftEdge + g->Width, g->TopEdge);
		Draw(rp, g->LeftEdge + g->Width, g->TopEdge + g->Height);
		Draw(rp, g->LeftEdge, g->TopEdge + g->Height);
		Draw(rp, g->LeftEdge, g->TopEdge);
*/
		if(pp->BitMap) {
				BltBitMapRastPort(pp->BitMap, 0, 0, rp, g->LeftEdge, g->TopEdge, g->Width, g->Height, 0xc0);
			} else {
				RectFill(rp, g->LeftEdge, g->TopEdge, 
						g->LeftEdge + g->Width - 1, g->TopEdge + g->Height - 1);
			}
	} else {
		if(pp->BitMap) {
				BltBitMapRastPort(pp->BitMap, pp->OKnob.MinX, pp->OKnob.MinY, rp, 
														g->LeftEdge + pp->OKnob.MinX, g->TopEdge + pp->OKnob.MinY,
														pp->OKnob.MaxX - pp->OKnob.MinX + 1, 
														pp->OKnob.MaxY - pp->OKnob.MinY + 1, 0xc0);
			} else {
				RectFill(rp, g->LeftEdge + pp->OKnob.MinX, g->TopEdge + pp->OKnob.MinY, 
							g->LeftEdge + pp->OKnob.MaxX, g->TopEdge + pp->OKnob.MaxY);
			}
	}

DrawImageState(rp, img, g->LeftEdge + pp->Knob.MinX, g->TopEdge + pp->Knob.MinY, pp->Dragging ? IDS_SELECTED : IDS_NORMAL, msg->gpr_GInfo->gi_DrInfo);

memcpy(&pp->OKnob, &pp->Knob, sizeof(pp->OKnob));
pp->OState = pp->Dragging;
}

static ULONG	PP_OMGet(Class *cl, struct Gadget *g, struct opGet *msg)
{
struct PPData *pp = INST_DATA(cl, g);

switch(msg->opg_AttrID) {
		case	PGA_Top:
			*msg->opg_Storage = pp->Top;
			return(TRUE);
		case	PGA_Total:
			*msg->opg_Storage = pp->Total;
			return(TRUE);
		case	PGA_Visible:
			*msg->opg_Storage = pp->Visible;
			return(TRUE);
		case	PGA_Freedom:
			*msg->opg_Storage = pp->Freedom;
			return(TRUE);
	}
return(DoSuperMethodA(cl, (APTR) g, (Msg) msg));
}


static ULONG	PP_OMSet(Class *cl, struct Gadget *g, struct opUpdate *msg)
{
struct PPData *pp = INST_DATA(cl, g);
struct TagItem *ti, *tstate = msg->opu_AttrList;
APTR rp;
BOOL update = FALSE;

while(ti = NextTagItem(&tstate))
	switch(ti->ti_Tag) {
			case	GA_ID:
				g->GadgetID = ti->ti_Data;
				break;
			case	PGA_Top:
				if(pp->Top != ti->ti_Data) {
						pp->Top = ti->ti_Data;
						update = TRUE;
					}
				break;
			case	PGA_Total:
				if(pp->Total != ti->ti_Data) {
						pp->Total = ti->ti_Data;
						update = TRUE;
					}
				break;
			case	PGA_Visible:
				if(pp->Visible != ti->ti_Data) {
						pp->Visible = ti->ti_Data;
						update = TRUE;
					}
				break;
			case	PPA_BitMap:
				pp->BitMap = (APTR) ti->ti_Data;
				break;
			case	IA_BGPen:
				pp->BgPen = ti->ti_Data;
				break;
			default:
				DoSuperMethodA(cl, (APTR) g, (Msg) msg);
		}

if(update) {
		PP_Layout(g, pp);

		if(!pp->NoDraw) {
				if(rp = ObtainGIRPort(msg->opu_GInfo)) {
						DoMethod((APTR) g, GM_RENDER, msg->opu_GInfo, rp, GREDRAW_UPDATE);
						ReleaseGIRPort(rp);
					}
			}

		NotifyAttrs((APTR) g, msg->opu_GInfo, 
			(msg->MethodID == OM_UPDATE) ? msg->opu_Flags : 0,
			PGA_Top, pp->Top,
			PGA_Visible, pp->Visible,
			PGA_Total, pp->Total,
			TAG_DONE);
	}
return(TRUE);
}


static ULONG	PP_OMNew(Class *cl, struct Gadget *g, struct opSet *msg)
{
struct PPData *pp;
struct TagItem *ti, *tstate = msg->ops_AttrList;

if(g = (APTR) DoSuperMethodA(cl, (APTR) g, (Msg) msg)) {
		if(pp = INST_DATA(cl, (APTR) g)) {

				while(ti = NextTagItem(&tstate))
					switch(ti->ti_Tag) {
							case	PGA_Freedom:
								pp->Freedom = ti->ti_Data;
								break;
							case	PGA_Top:
								pp->Top = ti->ti_Data;
								break;
							case	PGA_Total:
								pp->Total = ti->ti_Data;
								break;
							case	PGA_Visible:
								pp->Visible = ti->ti_Data;
								break;
							case	IA_BGPen:
								pp->BgPen = ti->ti_Data;
								break;
							case	PPA_BitMap:
								pp->BitMap = (APTR) ti->ti_Data;
								break;
						}

				PP_Layout(g, pp);
			}
	}
return((ULONG) g);
}



Class *MakePrayerPropClass(void)
{
Class *PPClass;
if(PPClass = MakeClass(NULL, GADGETCLASS, NULL, sizeof(struct PPData), 0))
	PPClass->cl_Dispatcher.h_Entry = (LONG (*)()) &PP_Dispatcher;

return(PPClass);
}

void FreePrayerPropClass(Class *PPClass)
{
if(PPClass)
	FreeClass(PPClass);
}
