/*
 * =-=-=-=-=-=-=-=-=-=-=-  NOTICE  =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * The following is the license that applies to this copy of the Prayer2
 * software.  For a license to use the Prayer2 software under conditions
 * other than those described here, please contact MindNever Software at
 * following addresses:
 *
 * MindNever Software <info@mindnever.org>
 * MindNever Software +381 63 550-161 or +381 64 220-7115
 * 
 * =-=-=-=-=-=-=-=-=-=-  COPYRIGHT  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Prayer2 is:
 *   Copyright (c) 1999-2001 by Tamara Cetinjski <tamara@mindnever.org>
 *   Copyright (c) 2001-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *   Copyright (c) 1999-2002 by MindNever Software Group <info@mindnever.org>
 *
 * Tamara Cetinjski is
 *   Copyright (c) 1999-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *
 *
 * =-=-=-=-=-=-=-=-=-=-  LICENSE  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Redistributions in any form must be accompanied by information on
 *    how to obtain complete source code for the Prayer2 software and any
 *    accompanying software that uses the Prayer2 software.  The source code
 *    must either be included in the distribution or be available for no
 *    more than the cost of distribution plus a nominal fee, and must be
 *    freely redistributable under reasonable conditions.  For an
 *    executable file, complete source code means the source code for all
 *    modules it contains.  It does not include source code for modules or
 *    files that typically accompany the major components of the operating
 *    system on which the executable file runs.
 *
 * =-=-=-=-=-=-=-=-=-=-  NO WARRANTY =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
 *
 * THIS SOFTWARE IS PROVIDED BY MINDNEVER SOFTWARE ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
 * NON-INFRINGEMENT, ARE DISCLAIMED.  IN NO EVENT SHALL MINDNEVER SOFTWARE
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, HAIR, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
#define __USE_SYSBASE 1
#include <devices/timer.h>
#include <exec/semaphores.h>

#include <exec/memory.h>
#include <exec/execbase.h>
#include <proto/exec.h>
#include <dos/dos.h>
#include <stdlib.h>

struct timerequest *qtreq;

static struct MsgPort *qmp = NULL;
static struct timerequest *qtr = NULL;
static BOOL ftimer_started = FALSE;
static BOOL ftimer_open = FALSE;

static void alrm_init(void);
static void alrm_cleanup(void);
static void alrm_abort(void);

BOOL alarm(ULONG secs, ULONG micro)
{

if(qtr == NULL)
	alrm_init();
if(qtr)
	{
qtreq = qtr;
		alrm_abort();
		if(secs || micro)
			{
				qtr->tr_time.tv_secs = secs;
				qtr->tr_time.tv_micro = micro;

				qtr->tr_node.io_Command = TR_ADDREQUEST;
				SetSignal(0L, SIGBREAKF_CTRL_D);
				BeginIO((APTR) qtr);
				ftimer_started = TRUE;
			}
	}
return(ftimer_started);
}

static void alrm_init(void)
{
#ifdef NOTYET
NewList(&alrm_list);
#endif

if(qtr = AllocMem(sizeof(*qtr), MEMF_PUBLIC|MEMF_CLEAR))
	{
		if(qmp = AllocMem(sizeof(*qmp), MEMF_PUBLIC|MEMF_CLEAR))
			{
				qmp->mp_Flags = PA_SIGNAL;
				qmp->mp_SigBit = SIGBREAKB_CTRL_D;
				qmp->mp_SigTask = SysBase->ThisTask;
				NewList(&qmp->mp_MsgList);

				qtr->tr_node.io_Message.mn_ReplyPort = qmp;

				if(!OpenDevice(TIMERNAME, UNIT_VBLANK, (APTR) qtr, 0L))
					{
						ftimer_open = TRUE;
						atexit(alrm_cleanup);
					}
			}
	}
if(ftimer_open == FALSE)
	alrm_cleanup();
}

static void alrm_cleanup(void)
{
if(qtr)
	{
		if(qmp)
			{
				if(ftimer_open)
					{
						alrm_abort();

						CloseDevice((APTR) qtr);
						ftimer_open = FALSE;
					}
				FreeMem(qmp, sizeof(*qmp));
				qmp = NULL;
			}
		FreeMem(qtr, sizeof(*qtr));
		qtr = NULL;
	}
}

static void alrm_abort(void)
{
if(ftimer_started)
	{
		if(!CheckIO((APTR) qtr))
			AbortIO((APTR) qtr);

		WaitIO((APTR) qtr);

		ftimer_started = FALSE;
	}
}
