/*
 * =-=-=-=-=-=-=-=-=-=-=-  NOTICE  =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * The following is the license that applies to this copy of the Prayer2
 * software.  For a license to use the Prayer2 software under conditions
 * other than those described here, please contact MindNever Software at
 * following addresses:
 *
 * MindNever Software <info@mindnever.org>
 * MindNever Software +381 63 550-161 or +381 64 220-7115
 * 
 * =-=-=-=-=-=-=-=-=-=-  COPYRIGHT  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Prayer2 is:
 *   Copyright (c) 1999-2001 by Tamara Cetinjski <tamara@mindnever.org>
 *   Copyright (c) 2001-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *   Copyright (c) 1999-2002 by MindNever Software Group <info@mindnever.org>
 *
 * Tamara Cetinjski is
 *   Copyright (c) 1999-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *
 *
 * =-=-=-=-=-=-=-=-=-=-  LICENSE  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Redistributions in any form must be accompanied by information on
 *    how to obtain complete source code for the Prayer2 software and any
 *    accompanying software that uses the Prayer2 software.  The source code
 *    must either be included in the distribution or be available for no
 *    more than the cost of distribution plus a nominal fee, and must be
 *    freely redistributable under reasonable conditions.  For an
 *    executable file, complete source code means the source code for all
 *    modules it contains.  It does not include source code for modules or
 *    files that typically accompany the major components of the operating
 *    system on which the executable file runs.
 *
 * =-=-=-=-=-=-=-=-=-=-  NO WARRANTY =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
 *
 * THIS SOFTWARE IS PROVIDED BY MINDNEVER SOFTWARE ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
 * NON-INFRINGEMENT, ARE DISCLAIMED.  IN NO EVENT SHALL MINDNEVER SOFTWARE
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, HAIR, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
/*
 * AMPlifier-alike plugin interface. So sue me !
 */

#include <exec/semaphores.h>
#include <exec/memory.h>
#include <exec/libraries.h>

#include <libraries/reqtools.h>
#include <libraries/amplifierplugin.h>
#include <clib/amplifierplugin_protos.h>

#include <proto/utility.h>
#include <proto/exec.h>
#include <proto/reqtools.h>
#include <proto/dos.h>

#include <string.h>

#include "prayer2_audio.h"
#include "prayer2_audiodriver.h"

#include "amplifierplugin.h"


struct MsgPort *ampport = NULL;

#if defined(mc68000)

extern make_spect(register WORD *src asm("a0"), register UWORD *dst asm("a1"), register UWORD size asm("d0"));

static LONG _LibOpen(register struct Library *libbase asm("a6"));
static ULONG _LibClose(register struct Library *libbase asm("a6"));
static ULONG _LibExpunge(register struct Library *libbase asm("a6"));
static void LIB_PI_RemModule(register struct PluginCtrl *pctrl asm("a0"));
static struct PluginCtrl * LIB_PI_AddModuleA(register struct TagItem *taglist asm("a0"));

struct PluginMessage {
	struct Message msg;
	ULONG          PluginMask;
	struct Process *PluginTask;
	UWORD          **SpecRawL;
	UWORD          **SpecRawR;
	BOOL           Accepted;

	/* All data beyond this point is new for v1.2.                         */
	/* AmigaAMP v2.5 and up will detect this new data and act accordingly. */
	/* Older versions of AmigaAMP will simply ignore it.                   */

	ULONG            InfoMask;
	struct TrackInfo **tinfo;
	ULONG            reserved;
};

static ULONG vectors[] = {
		(ULONG) _LibOpen,
		(ULONG) _LibClose,
		(ULONG) _LibExpunge,
		(ULONG) 0L,
		(ULONG) LIB_PI_AddModuleA,
		(ULONG) LIB_PI_RemModule,
		(ULONG) ~0,
	};

static struct Library *ampbase = NULL;


static BOOL shutting_down = FALSE;
static struct List vis_list;
static struct SignalSemaphore vis_lock;
static ULONG aamp_cnt = 0;

enum {
		VISTYPE_AMPLIFIER,
		VISTYPE_AMIGAAMP,
	};

static UWORD	SpecRawR[512];
static UWORD	SpecRawL[512];

struct vis_node {
		struct vis_node *next;
		struct vis_node *prev;
		UWORD	Type;
		APTR quittask;
		ULONG quitmask;
		struct Hook *renderhook;
		char	*name;
		struct PluginCtrl ctrl;
		APTR	owner;
	};


void ampemu_init(void)
{
NewList(&vis_list);
InitSemaphore(&vis_lock);

memset(SpecRawR, 0, sizeof(SpecRawR));
memset(SpecRawL, 0, sizeof(SpecRawL));

if(ampport = CreateMsgPort()) {
		ampport->mp_Node.ln_Name = "AmigaAMP plugin port";
		ampport->mp_Node.ln_Pri = 5;
		AddPort(ampport);
		if(ampbase = MakeLibrary(&vectors[0], NULL, NULL, sizeof(struct Library), NULL)) {
				ampbase->lib_Node.ln_Type = NT_LIBRARY;
				ampbase->lib_Node.ln_Pri = 0;
				ampbase->lib_Node.ln_Name = "amplifierplugin.library";
				ampbase->lib_Flags = LIBF_SUMUSED;
				ampbase->lib_Version = 1;
				ampbase->lib_Revision = 0;
				ampbase->lib_IdString = "Prayer2.0 amplifierplugin.library interface";
				ampbase->lib_OpenCnt = 0;
				AddLibrary(ampbase);
			}
	}
}

void ampemu_cleanup(void)
{
struct vis_node *node;
APTR msg;

if(ampbase) {
		Forbid();
		Remove((APTR) ampbase);
		Permit();
		for(node = (APTR) vis_list.lh_Head; node->next; node = node->next) {
				if(node->quittask) {
						switch(node->Type) {
								case	VISTYPE_AMPLIFIER:
									Signal(node->quittask, node->quitmask);
									break;
								case	VISTYPE_AMIGAAMP:
									Signal(node->quittask, SIGBREAKF_CTRL_C);
									break;
							}
					}
			}
		Delay(10);

		while(ampbase->lib_OpenCnt > 0) {
//				rtEZRequestA("Please quit all visualization plugins !", "Ok", NULL, NULL, NULL);
Delay(50);
			}
		FreeMem((char *) ampbase - ampbase->lib_NegSize, ampbase->lib_NegSize + ampbase->lib_PosSize);		
	}

if(ampport) {
		Forbid();
		RemPort(ampport);
		Permit();
		shutting_down = TRUE;
		while(msg = GetMsg(ampport)) {
				ampemu_handlemsg(msg);
				ReplyMsg(msg);
			}
		DeleteMsgPort(ampport);
	}
}

void ampemu_update(void)
{
struct vis_node *node;
struct WaveBuffer *wbuf;
struct RenderData rd;
ULONG offs;
typedef long (* RENDERFUNC)(register struct Hook *hk asm("a0"), register struct PluginCtrl *pc asm("a2"), register struct RenderData *rd asm("a1"));
RENDERFUNC RenderFunc;

if(ampbase) {
		if(PShared.status == PSTATUS_PRAYING) {
				if(wbuf = AudioStatus.CurrentBuffer) {
						offs = AudioStatus.SampleOffs;
						rd.Samplerate = PShared.dec.frequency;
						rd.Channels = PShared.dec.channels;
						rd.waveform[0] = wbuf->Ch1 + offs;
						rd.waveform[1] = wbuf->Ch2 + offs;

//						rd.spectrum[0] = rd.spectrum[1] = spect_static;
						rd.spectrum[0] = rd.waveform[0]; // :-))
						rd.spectrum[1] = rd.waveform[1]; // :-))

						if(aamp_cnt) {
								// grrr, copy some WORDs .
								make_spect(rd.spectrum[0], SpecRawR, sizeof(SpecRawR));
								make_spect(rd.spectrum[1], SpecRawL, sizeof(SpecRawL));
							}

						ObtainSemaphore(&vis_lock);
						for(node = (APTR) vis_list.lh_Head; node->next; node = node->next) {
								switch(node->Type) {
										case	VISTYPE_AMPLIFIER:
											RenderFunc = (RENDERFUNC)node->renderhook->h_Entry;
											RenderFunc(node->renderhook, &node->ctrl, &rd);
											break;
										case	VISTYPE_AMIGAAMP:
											if(node->quittask)
												Signal(node->quittask, node->quitmask);
											break;
									}
							}
						ReleaseSemaphore(&vis_lock);
					}
			}
	}
}

static ULONG _LibExpunge(register struct Library *libbase asm("a6"))
{
return(NULL);
}

static LONG _LibOpen(register struct Library *libbase asm("a6"))
{
libbase->lib_OpenCnt++;
return((LONG) libbase);
}

static ULONG _LibClose(register struct Library *libbase asm("a6"))
{
libbase->lib_OpenCnt--;
return(0L);
}

static __saveds struct PluginCtrl * LIB_PI_AddModuleA(register struct TagItem *tag asm("a0"))
{
struct TagItem *tstate = tag;
struct vis_node *node;

if(node = AllocMem(sizeof(*node), MEMF_CLEAR)) {
		node->Type = VISTYPE_AMPLIFIER;
		while(tag = NextTagItem(&tstate))
			switch(tag->ti_Tag) {
					case	PIA_Name:
						node->name = (char *) tag->ti_Data;
						break;
					case	PIA_QuitTask:
						node->quittask = (APTR) tag->ti_Data;
						break;
					case	PIA_QuitMask:
						node->quitmask = (ULONG) tag->ti_Data;
						break;
					case	PIA_UserData:
						node->ctrl.UserData = (APTR) tag->ti_Data;
						break;
					case	PIA_RenderHook:
						node->renderhook = (struct Hook *) tag->ti_Data;
						break;
				}

		ObtainSemaphore(&vis_lock);		
		AddTail(&vis_list, (APTR) node);
		ReleaseSemaphore(&vis_lock);
		return(&node->ctrl);
	}
return(NULL);
}

static __saveds void LIB_PI_RemModule(register struct PluginCtrl *pctrl asm("a0"))
{
struct vis_node *node;

if(pctrl) {
		node = (APTR) ((char *) pctrl - (sizeof(*node) - sizeof(*pctrl)));
		ObtainSemaphore(&vis_lock);
		Remove((APTR) node);
		ReleaseSemaphore(&vis_lock);
		FreeMem(node, sizeof(*node));
	}
}


void ampemu_handlemsg(struct PluginMessage *pmsg)
{
struct vis_node *node;
if(pmsg->PluginTask) { // register
		if(shutting_down) {
				pmsg->Accepted = FALSE;
			} else {
				if(node = AllocMem(sizeof(*node), MEMF_CLEAR)) {
						node->Type = VISTYPE_AMIGAAMP;
						node->quitmask = pmsg->PluginMask;
						node->quittask = pmsg->PluginTask;
						node->owner = pmsg->msg.mn_ReplyPort->mp_SigTask;
						pmsg->Accepted = TRUE;
						*pmsg->SpecRawL = SpecRawL;
						*pmsg->SpecRawR = SpecRawR;
						ObtainSemaphore(&vis_lock);		
						AddTail(&vis_list, (APTR) node);
						aamp_cnt++;
						ReleaseSemaphore(&vis_lock);
					}
			}
	} else {
		for(node = (APTR) vis_list.lh_Head; node->next; node = node->next) {
				if(node->Type == VISTYPE_AMIGAAMP) {
						if(node->owner == pmsg->msg.mn_ReplyPort->mp_SigTask) {
								ObtainSemaphore(&vis_lock);		
								Remove((APTR) node);
								aamp_cnt--;
								ReleaseSemaphore(&vis_lock);
								FreeMem(node, sizeof(*node));
								break;
							}
					}
			}
	}
}

#else /* defined(mc68000) */

void ampemu_init(void)
{
}

void ampemu_cleanup(void)
{
}

void ampemu_update(void)
{
}


void ampemu_handlemsg(struct PluginMessage *pmsg)
{
}


#endif /* defined(mc68000) */
