/*
 * =-=-=-=-=-=-=-=-=-=-=-  NOTICE  =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * The following is the license that applies to this copy of the Prayer2
 * software.  For a license to use the Prayer2 software under conditions
 * other than those described here, please contact MindNever Software at
 * following addresses:
 *
 * MindNever Software <info@mindnever.org>
 * MindNever Software +381 63 550-161 or +381 64 220-7115
 * 
 * =-=-=-=-=-=-=-=-=-=-  COPYRIGHT  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Prayer2 is:
 *   Copyright (c) 1999-2001 by Tamara Cetinjski <tamara@mindnever.org>
 *   Copyright (c) 2001-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *   Copyright (c) 1999-2002 by MindNever Software Group <info@mindnever.org>
 *
 * Tamara Cetinjski is
 *   Copyright (c) 1999-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *
 *
 * =-=-=-=-=-=-=-=-=-=-  LICENSE  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Redistributions in any form must be accompanied by information on
 *    how to obtain complete source code for the Prayer2 software and any
 *    accompanying software that uses the Prayer2 software.  The source code
 *    must either be included in the distribution or be available for no
 *    more than the cost of distribution plus a nominal fee, and must be
 *    freely redistributable under reasonable conditions.  For an
 *    executable file, complete source code means the source code for all
 *    modules it contains.  It does not include source code for modules or
 *    files that typically accompany the major components of the operating
 *    system on which the executable file runs.
 *
 * =-=-=-=-=-=-=-=-=-=-  NO WARRANTY =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
 *
 * THIS SOFTWARE IS PROVIDED BY MINDNEVER SOFTWARE ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
 * NON-INFRINGEMENT, ARE DISCLAIMED.  IN NO EVENT SHALL MINDNEVER SOFTWARE
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, HAIR, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
#include <devices/ahi.h>
#include <exec/exec.h>
#include <proto/ahi.h>
#include <proto/dos.h>
#include <proto/exec.h>
#include <dos/dostags.h>
#include <string.h>


#ifndef min
#define min(a,b) ((a) < (b) ? (a) : (b))
#endif

#include	"prayer2_audiodriver.h"
#include "opt68k.h"

static void queueAHI(int init);

static struct AHIAudioCtrl *actrl = NULL;

struct AHISampleInfo Sample[2] = {
		{ AHIST_S16S, NULL, NULL },
		{ AHIST_S16S, NULL, NULL },
	};

struct Library *AHIBase = NULL;

static ULONG blksize = 0;
static UWORD active = 0;
static BOOL stereo = FALSE;
static struct WaveBuffer *inuse[2] = {NULL, NULL};
static struct List QueueList;
static ULONG QueueSeq = 0;


#ifdef __MORPHOS__

static ULONG __saveds SoundFunc_Native(void)
{
struct Hook *hook = REG_A0;
struct AHIAudioCtrl *actrl = REG_A2;
struct AHISoundMessage *smsg = REG_A1;

if(inuse[active])
	ma_DMAStart(inuse[active]);

if(inuse[active ^ 1]) {
		ma_FreeWBuffer(inuse[active ^ 1]);
		inuse[active ^ 1] = NULL;
	}

queueAHI(FALSE); // get from queue, and fill first free buffer
return(1);
}

struct EmulLibEntry SoundFunc = {
		TRAP_LIBNR,
		0,
		SoundFunc_Native
	};

#else
static ULONG __saveds SoundFunc(register struct Hook *hook ASM("a0"),
    register struct AHIAudioCtrl *actrl ASM("a2"),
    register struct AHISoundMessage *smsg ASM("a1"))
{

if(inuse[active])
	ma_DMAStart(inuse[active]);

if(inuse[active ^ 1]) {
		ma_FreeWBuffer(inuse[active ^ 1]);
		inuse[active ^ 1] = NULL;
	}

queueAHI(FALSE); // get from queue, and fill first free buffer
return(1);
}
#endif


struct Hook SoundHook =
{
  0,0,
  &SoundFunc,
  NULL,
  NULL,
};




static void queueAHI(int init)
{
int l, f_imm = 0;
struct WaveBuffer *wbuf;

if(QueueSeq < 3)
	return;

// first, find free sample buffer

if(inuse[0] == NULL)
	active = 0;
else if(inuse[1] == NULL)
	active = 1;
else return;

// fill buffer here, and ...

if(wbuf = (APTR) RemHead(&QueueList)) {
		if((inuse[0] == NULL) && (inuse[1] == NULL) && init)
			f_imm = AHISF_IMM;

		inuse[active] = wbuf;

		// mix, mix, mix...

		l = min(blksize, wbuf->Length); // number of samples

		if(stereo) {
				ahi_join_stereo(wbuf->Ch1, wbuf->Ch2, Sample[active].ahisi_Address, l);
				// join, but not join...
			} else {
				CopyMem(wbuf->Ch1, Sample[active].ahisi_Address, l * sizeof(WORD));
			}

		if(l != Sample[active].ahisi_Length) {
				AHI_UnloadSound(active, actrl);
				Sample[active].ahisi_Length = l;
				AHI_LoadSound(active, AHIST_DYNAMICSAMPLE, &Sample[active], actrl);
			}

		AHI_SetSound(0, active, 0L, 0L, actrl, f_imm);
	}
}


/////////// PUBLIC FUNCTIONS //////////////////////////

BOOL ahi_ad_AudioInit(struct AudioSettings *as)
{
ULONG pys, mxf, bfl, typ;

// FIXME: initialize ahi.device here.

if(AHIBase == NULL)
	return(0);

NewList(&QueueList);

QueueSeq = 0;

inuse[0] = inuse[1] = NULL;

actrl = AHI_AllocAudio(AHIA_AudioID, as->AhiMode,
											AHIA_MixFreq, as->MixFreq,
											AHIA_Channels, 1,
											AHIA_Sounds, 2,
											AHIA_SoundFunc, &SoundHook,
											TAG_DONE);
if(actrl == NULL)
	return(0);

AHI_GetAudioAttrs(AHI_INVALID_ID, actrl,
									AHIDB_MaxPlaySamples, &pys,
									TAG_DONE);
AHI_ControlAudio(actrl, AHIC_MixFreq_Query, &mxf,
									TAG_DONE);

bfl = pys * as->Freq / mxf;

if(bfl < as->BlkSiz)
	bfl = as->BlkSiz;

blksize = bfl;

if(stereo = as->Stereo) {
		bfl *= 2;
		typ = AHIST_S16S;
	} else {
		typ = AHIST_M16S;
	}

Sample[0].ahisi_Type = Sample[1].ahisi_Type = typ;
Sample[0].ahisi_Length = Sample[1].ahisi_Length = blksize;
Sample[0].ahisi_Address = AllocVec(bfl * sizeof(WORD), MEMF_PUBLIC|MEMF_CLEAR); // 
Sample[1].ahisi_Address = AllocVec(bfl * sizeof(WORD), MEMF_PUBLIC|MEMF_CLEAR);

if(Sample[0].ahisi_Address && Sample[1].ahisi_Address) {

		if(!AHI_LoadSound(0, AHIST_DYNAMICSAMPLE, &Sample[0], actrl)
			&& !AHI_LoadSound(1, AHIST_DYNAMICSAMPLE, &Sample[1], actrl)) {

				AHI_ControlAudio(actrl, AHIC_Play, TRUE, TAG_DONE);

				return(1);
			}
	}

ahi_ad_AudioCleanup();

return(0);
}

void ahi_ad_AudioCleanup(void)
{
if(actrl) {

		AHI_ControlAudio(actrl, AHIC_Play, FALSE, TAG_DONE);

		AHI_UnloadSound(0, actrl);
		AHI_UnloadSound(1, actrl);

		FreeVec(Sample[0].ahisi_Address); Sample[0].ahisi_Address = NULL;
		FreeVec(Sample[1].ahisi_Address); Sample[1].ahisi_Address = NULL;

		AHI_FreeAudio(actrl);
		actrl = NULL;
	}

if(inuse[0])
	ma_FreeWBuffer(inuse[0]);
if(inuse[1])
	ma_FreeWBuffer(inuse[1]);

inuse[0] = inuse[1] = NULL;
}

void ahi_ad_AudioFlush(void)
{
APTR wbuf;

AHI_SetSound(0, AHI_NOSOUND, 0, 0, actrl, AHISF_IMM);

while(wbuf = RemHead(&QueueList)) {
		ma_FreeWBuffer(wbuf);
	}

if(inuse[0])
	ma_FreeWBuffer(inuse[0]);
if(inuse[1])
	ma_FreeWBuffer(inuse[1]);

inuse[0] = inuse[1] = NULL;
QueueSeq = 0;
}

void ahi_ad_AudioQueue(struct WaveBuffer *wbuf)
{

Disable(); // Dammit!
AddTail(&QueueList, (APTR) wbuf);
QueueSeq++;
Enable();

queueAHI(TRUE);
}

void ahi_ad_AudioUpdate(struct AudioSettings *as)
{
// as->Freq;
if(actrl) {
		AHI_SetFreq(0, as->Freq, actrl, AHISF_IMM);
		AHI_SetVol(0, (0x10000L * as->Volume) / 100, 0x8000L, actrl, AHISF_IMM);
		if(as->Freq == 0)
			QueueSeq = 0;
	}
}

