/*
 * =-=-=-=-=-=-=-=-=-=-=-  NOTICE  =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * The following is the license that applies to this copy of the Prayer2
 * software.  For a license to use the Prayer2 software under conditions
 * other than those described here, please contact MindNever Software at
 * following addresses:
 *
 * MindNever Software <info@mindnever.org>
 * MindNever Software +381 63 550-161 or +381 64 220-7115
 * 
 * =-=-=-=-=-=-=-=-=-=-  COPYRIGHT  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Prayer2 is:
 *   Copyright (c) 1999-2001 by Tamara Cetinjski <tamara@mindnever.org>
 *   Copyright (c) 2001-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *   Copyright (c) 1999-2002 by MindNever Software Group <info@mindnever.org>
 *
 * Tamara Cetinjski is
 *   Copyright (c) 1999-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *
 *
 * =-=-=-=-=-=-=-=-=-=-  LICENSE  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Redistributions in any form must be accompanied by information on
 *    how to obtain complete source code for the Prayer2 software and any
 *    accompanying software that uses the Prayer2 software.  The source code
 *    must either be included in the distribution or be available for no
 *    more than the cost of distribution plus a nominal fee, and must be
 *    freely redistributable under reasonable conditions.  For an
 *    executable file, complete source code means the source code for all
 *    modules it contains.  It does not include source code for modules or
 *    files that typically accompany the major components of the operating
 *    system on which the executable file runs.
 *
 * =-=-=-=-=-=-=-=-=-=-  NO WARRANTY =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
 *
 * THIS SOFTWARE IS PROVIDED BY MINDNEVER SOFTWARE ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
 * NON-INFRINGEMENT, ARE DISCLAIMED.  IN NO EVENT SHALL MINDNEVER SOFTWARE
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, HAIR, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
#define	__USE_SYSBASE
#include	<devices/audio.h>
#include	<exec/ports.h>
#include	<exec/interrupts.h>
#include	<exec/memory.h>
#include	<exec/execbase.h>
#include	<hardware/custom.h>
#include	<hardware/cia.h>
#include	<hardware/dmabits.h>
#include	<hardware/intbits.h>
#include	<proto/dos.h>
#include	<proto/exec.h>

#include	<dos/dostags.h>
//#include	<dos.h>

#include	"prayer2_audiodriver.h"

#include "opt68k.h"

#define	CHAN_LEFT_MSB		(1 << 0)
#define	CHAN_RIGHT_MSB	(1 << 1)
#define	CHAN_RIGHT_LSB	(1 << 2)
#define	CHAN_LEFT_LSB		(1 << 3)

#define	CHAN_LEFT		(CHAN_LEFT_LSB|CHAN_LEFT_MSB)
#define	CHAN_RIGHT	(CHAN_RIGHT_LSB|CHAN_RIGHT_MSB)

typedef struct AudioStuff Audio;

struct AudioStuff {
		struct IOAudio io_r1;
		struct IOAudio io_r2;
		struct IOAudio io_l1;
		struct IOAudio io_l2;

		struct MsgPort		mp_b;
		struct Interrupt	ir_b;

		LONG	inuse;
		ULONG	data_len;
		struct WaveBuffer *wbuf;
		Audio *other;
	};

extern struct Custom custom;
extern struct CIA ciaa, ciab;

static Audio	*io1 = NULL, *io2 = NULL;
static struct IOAudio *io_c = NULL;
static struct	MsgPort *mp_c = NULL;
static ULONG	flags = 0;
static ULONG	MaxBlkSiz = 0;
static ULONG	QueueSeq = 0;
static UWORD boost = 100;
static volatile struct Process *audio_proc = NULL;
static BYTE child_sig = -1;
static UWORD Frequency = 44100;
static UWORD Volume = 100;
static struct Task *me = NULL;
static struct SignalSemaphore QueueLock;

#define ADF_FILTER			(1L << 2)
#define	ADF_STEREO			(1L << 3)
#define ADF_OPEN				(1L << 4)

struct List	queue;

#define ERR_NULL	error:
#define	EXC_NULL	{goto error;}
#define CHK_NULL(x) if(x == NULL) goto error;


static BOOL openAudio(void);
static BOOL initIO(Audio *io);
static void freeIO(Audio *io);
static void waitIO(Audio *io);
static BOOL checkIO(Audio *io);
static void closeAudio(void);
static void beginIO(Audio *io);
static void queueIO(Audio *io);
static void abortIO(Audio *io);

#ifdef __MORPHOS__


static void  __saveds AudioInt_BeginDMA_Native(void)
{
Audio *io = (Audio *)REG_A1;

ma_DMAStart(io->wbuf);
}

struct EmulLibEntry AudioInt_BeginDMA = {
		TRAP_LIBNR,
		0,
		AudioInt_BeginDMA_Native
	};

#else /* __MORPHOS__ */

static void  __saveds AudioInt_BeginDMA(register Audio *io ASM("a1"))
{
ma_DMAStart(io->wbuf);
}

#endif



static void queueIO(Audio *io)
{
struct WaveBuffer *wbuf;

if(io->wbuf) {
		ma_FreeWBuffer(io->wbuf);
		io->wbuf = NULL;
	}

if(QueueSeq < 3)
	return;

ObtainSemaphore(&QueueLock);
wbuf = (APTR) RemHead(&queue);
ReleaseSemaphore(&QueueLock);

if(wbuf) {
		c16to14(wbuf->Ch1, io->io_r1.ioa_Data, 
																		io->io_r2.ioa_Data,
																		wbuf->Length,
																		boost);
		if(flags & ADF_STEREO) {
				c16to14(wbuf->Ch2, io->io_l1.ioa_Data, 
																		io->io_l2.ioa_Data,
																		wbuf->Length,
																		boost);
			}
		io->data_len = wbuf->Length;
		io->wbuf = wbuf;

		beginIO(io);

	}
}

static void beginIO(Audio *io)
{
BOOL resync = FALSE;

io->io_r1.ioa_Length = 
	io->io_r2.ioa_Length = 
	io->io_l1.ioa_Length = 
	io->io_l2.ioa_Length = io->data_len;

if((io->other->inuse < 1) || (custom.dmaconr & DMAF_AUDIO) != DMAF_AUDIO) {
		abortIO(io->other);
		io_c->ioa_Request.io_Unit = (APTR)(CHAN_LEFT | CHAN_RIGHT);
		io_c->ioa_Request.io_Command = CMD_STOP;
		DoIO((APTR) io_c);
		resync = TRUE;
	}

if(flags & ADF_FILTER) {
//		__emit(0x08B9); // bclr
//		__emit(0x0001);
//		__emit(0x00BF);
//		__emit(0xE001);

		ciaa.ciapra &= ~CIAF_LED;	
	} else {
//		__emit(0x08F9); // bset
//		__emit(0x0001);
//		__emit(0x00BF);
//		__emit(0xE001);
		
		ciaa.ciapra |= CIAF_LED;
	}

io->inuse = 4;

BeginIO((APTR) &io->io_r1);
BeginIO((APTR) &io->io_l1);
BeginIO((APTR) &io->io_r2);
BeginIO((APTR) &io->io_l2);

if(resync) {
		io_c->ioa_Request.io_Command = CMD_START;
		DoIO((APTR) io_c);
	}
}

static BOOL openAudio(void)
{
UBYTE ch = 0x0f;

mp_c = CreateMsgPort(); CHK_NULL(mp_c);
io_c = (APTR) CreateIORequest(mp_c, sizeof(*io_c)); CHK_NULL(io_c);

io_c->ioa_Request.io_Message.mn_Node.ln_Pri = 127;
io_c->ioa_Request.io_Command = ADCMD_ALLOCATE;
io_c->ioa_Request.io_Flags = ADIOF_NOWAIT;
io_c->ioa_Data = &ch;
io_c->ioa_Length = 1;

if(OpenDevice("audio.device", 0, (APTR) io_c, 0))
	EXC_NULL;

io_c->ioa_Request.io_Unit = (APTR) (CHAN_LEFT|CHAN_RIGHT);

io_c->ioa_Request.io_Command = CMD_STOP;
DoIO((APTR) io_c);


io1 = AllocMem(sizeof(*io1), MEMF_CLEAR|MEMF_PUBLIC); CHK_NULL(io1);
io2 = AllocMem(sizeof(*io2), MEMF_CLEAR|MEMF_PUBLIC); CHK_NULL(io2);

if(!initIO(io1))
	EXC_NULL;
if(!initIO(io2))
	EXC_NULL;

io1->other = io2;
io2->other = io1;

flags |= ADF_OPEN;

ERR_NULL;

if(flags & ADF_OPEN)
	return(TRUE);

closeAudio();

return(FALSE);
}

static BOOL initIO(Audio *io)
{
// Copy device data

io_c->ioa_Request.io_Flags = 0;
CopyMem(io_c, &io->io_r1, sizeof(io->io_r1));
io->io_r1.ioa_Request.io_Command = CMD_WRITE;
io->io_r1.ioa_Cycles = 1;

// Copy all over
CopyMem(&io->io_r1, &io->io_l1, sizeof(io->io_l1));
CopyMem(&io->io_r1, &io->io_r2, sizeof(io->io_r2));
CopyMem(&io->io_r1, &io->io_l2, sizeof(io->io_l2));

// Setup BeginPort
io->ir_b.is_Code = (void (*)()) &AudioInt_BeginDMA;
io->ir_b.is_Data = io;
io->mp_b.mp_Node.ln_Type = NT_MSGPORT;
io->mp_b.mp_Flags = PA_SOFTINT;
io->mp_b.mp_SoftInt = (APTR) &io->ir_b;

NewList(&io->mp_b.mp_MsgList);
io->io_r1.ioa_Request.io_Flags |= ADIOF_WRITEMESSAGE;
io->io_r1.ioa_WriteMsg.mn_ReplyPort = &io->mp_b;

// Setup channels
io->io_r1.ioa_Request.io_Unit = (APTR) CHAN_RIGHT_LSB;
io->io_r2.ioa_Request.io_Unit = (APTR) CHAN_RIGHT_MSB;
io->io_l1.ioa_Request.io_Unit = (APTR) CHAN_LEFT_LSB;
io->io_l2.ioa_Request.io_Unit = (APTR) CHAN_LEFT_MSB;

io->io_r1.ioa_Data = AllocMem(MaxBlkSiz, MEMF_CHIP|MEMF_CLEAR); CHK_NULL(io->io_r1.ioa_Data);
io->io_r2.ioa_Data = AllocMem(MaxBlkSiz, MEMF_CHIP|MEMF_CLEAR); CHK_NULL(io->io_r2.ioa_Data);

if(flags & ADF_STEREO) {
		io->io_l1.ioa_Data = AllocMem(MaxBlkSiz, MEMF_CHIP|MEMF_CLEAR); CHK_NULL(io->io_l1.ioa_Data);
		io->io_l2.ioa_Data = AllocMem(MaxBlkSiz, MEMF_CHIP|MEMF_CLEAR); CHK_NULL(io->io_l2.ioa_Data);
	} else {
		io->io_l1.ioa_Data = io->io_r1.ioa_Data;
		io->io_l2.ioa_Data = io->io_r2.ioa_Data;
	}

return(TRUE);

ERR_NULL;
return(FALSE);
}

static void freeIO(Audio *io)
{
if(io) {
		waitIO(io);

		if(io->io_r1.ioa_Data)
			FreeMem(io->io_r1.ioa_Data, MaxBlkSiz);
		if(io->io_r2.ioa_Data)
			FreeMem(io->io_r2.ioa_Data, MaxBlkSiz);
		if(flags & ADF_STEREO) {
				if(io->io_l1.ioa_Data)
					FreeMem(io->io_l1.ioa_Data, MaxBlkSiz);
				if(io->io_l2.ioa_Data)
					FreeMem(io->io_l2.ioa_Data, MaxBlkSiz);
			}
		FreeMem(io, sizeof(*io));
	}
}

static void abortIO(Audio *io)
{
if(io->inuse) {
		AbortIO((APTR) &io->io_r1);
		AbortIO((APTR) &io->io_l1);
		AbortIO((APTR) &io->io_r2);
		AbortIO((APTR) &io->io_l2);
	}
}

static BOOL checkIO(Audio *io)
{
if((io->inuse == 0) || CheckIO((APTR) &io->io_r1))
	return(TRUE);
else
	return(FALSE);
}

static void waitIO(Audio *io)
{
if(io->inuse) {
		WaitIO((APTR) &io->io_r1);
		WaitIO((APTR) &io->io_l1);
		WaitIO((APTR) &io->io_r2);
		WaitIO((APTR) &io->io_l2);
		io->inuse = 0;
	}
}

static void closeAudio(void)
{
if(io1)
	freeIO(io1);
if(io2)
	freeIO(io2);

io1 = io2 = NULL;

if(mp_c) {
		if(io_c) {
				if((LONG) io_c->ioa_Request.io_Device > 0L)
					CloseDevice((APTR) io_c);
				DeleteIORequest((APTR) io_c);
				io_c = NULL;
			}
		DeleteMsgPort(mp_c);
		mp_c = NULL;
	}

flags = 0L;
}

static void runUpdate(void)
{
UBYTE vol_lsb, vol_msb;
static BOOL zero_freq = FALSE;

if(Frequency) {
		io_c->ioa_Period = SysBase->ex_EClockFrequency * 5 / Frequency;
		if(zero_freq) {
				io_c->ioa_Request.io_Command = CMD_START;
				DoIO((APTR) io_c);
				zero_freq = FALSE;
			}
	} else {
		io_c->ioa_Request.io_Command = CMD_STOP;
		DoIO((APTR) io_c);
		zero_freq = TRUE;
	}

io_c->ioa_Request.io_Command = ADCMD_PERVOL;

if(Volume > 100) {
		io_c->ioa_Volume = 64;
		boost = Volume;
	} else {
		io_c->ioa_Volume = (Volume * 64) / 100;
		boost = 100;
	}

vol_msb = io_c->ioa_Volume;

io_c->ioa_Request.io_Unit = (APTR)(CHAN_LEFT_MSB | CHAN_RIGHT_MSB);

DoIO((APTR) io_c);

if(Volume > 50)
	io_c->ioa_Volume = 1;
else
	io_c->ioa_Volume = 0;

vol_lsb = io_c->ioa_Volume;

io_c->ioa_Request.io_Unit = (APTR)(CHAN_LEFT_LSB | CHAN_RIGHT_LSB);
DoIO((APTR) io_c);

io_c->ioa_Request.io_Unit = (APTR)(CHAN_LEFT | CHAN_RIGHT);
}

static void runFlush(void)
{
APTR wbuf;

abortIO(io1);
abortIO(io2);
waitIO(io1);
waitIO(io2);

while(wbuf = RemTail(&queue))
	ma_FreeWBuffer(wbuf);

if(io1->wbuf)
	ma_FreeWBuffer(io1->wbuf);
if(io2->wbuf)
	ma_FreeWBuffer(io2->wbuf);

QueueSeq = 0;
}

static __saveds void audio_task(void)
{
BOOL need_break = FALSE;
ULONG signals, sigmask;

if(openAudio()) {
		Signal(me, 1L << child_sig);

		sigmask = SIGBREAKF_CTRL_C |
							SIGBREAKF_CTRL_D |
							SIGBREAKF_CTRL_E |
							SIGBREAKF_CTRL_F |
							(1L << mp_c->mp_SigBit);

		runUpdate();

		while(1) {
				signals = Wait(sigmask);

				if(signals & SIGBREAKF_CTRL_C)
					need_break = TRUE;
				if(signals & SIGBREAKF_CTRL_D)
					runFlush();
				if(signals & SIGBREAKF_CTRL_F)
					runUpdate();

				if(checkIO(io1)) {
						waitIO(io1);
						queueIO(io1);
					}
				if(checkIO(io2)) {
						waitIO(io2);
						queueIO(io2);
					}

				if(need_break) {
						if(IsListEmpty(&queue))
							break;
					}
			}

		waitIO(io1);
		waitIO(io2);

		runFlush();
		
		closeAudio();
	}

Forbid();
audio_proc = NULL;
Signal(me, 1L << child_sig);
}


static int signal_audio(ULONG signals)
{
int rc = FALSE;
Forbid();
if(audio_proc) {
		Signal((APTR) audio_proc, signals);
		rc = TRUE;
	}
Permit();
return(rc);
}


/////////// PUBLIC FUNCTIONS //////////////////////////

BOOL paula_14bit_ad_AudioInit(struct AudioSettings *as)
{
BOOL rc = FALSE;

if(audio_proc)
	return(TRUE);

child_sig = AllocSignal(-1);

MaxBlkSiz = as->BlkSiz * sizeof(WORD);
if(as->Stereo)
	flags |= ADF_STEREO;

me = FindTask(0);
InitSemaphore(&QueueLock);
NewList(&queue);

paula_14bit_ad_AudioUpdate(as);

SetSignal(0L, 1L << child_sig);

if(audio_proc = CreateNewProcTags(NP_Entry, audio_task,
															NP_Name, "14bit.task",
															NP_Output, Output(),
															NP_CloseOutput, FALSE,
															NP_Priority, 50,
															TAG_DONE)) {
		Wait(1L << child_sig);
		if(audio_proc)
			rc = TRUE;
	}
if(!rc) {
		FreeSignal(child_sig);
		child_sig = -1;
	}
return(rc);
}

void paula_14bit_ad_AudioCleanup(void)
{
signal_audio(SIGBREAKF_CTRL_C);

while(audio_proc) {
		Wait(1L << child_sig);
	}

FreeSignal(child_sig);
child_sig = -1;
}

void paula_14bit_ad_AudioFlush(void)
{
signal_audio(SIGBREAKF_CTRL_D);
}

void paula_14bit_ad_AudioQueue(struct WaveBuffer *wbuf)
{
ObtainSemaphore(&QueueLock);
AddTail(&queue, (APTR) wbuf);
QueueSeq++;
ReleaseSemaphore(&QueueLock);

signal_audio(SIGBREAKF_CTRL_E);
}

void paula_14bit_ad_AudioUpdate(struct AudioSettings *as)
{
Frequency = as->Freq;
Volume = as->Volume;

signal_audio(SIGBREAKF_CTRL_F);
}
