/*
 * =-=-=-=-=-=-=-=-=-=-=-  NOTICE  =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * The following is the license that applies to this copy of the Prayer2
 * software.  For a license to use the Prayer2 software under conditions
 * other than those described here, please contact MindNever Software at
 * following addresses:
 *
 * MindNever Software <info@mindnever.org>
 * MindNever Software +381 63 550-161 or +381 64 220-7115
 * 
 * =-=-=-=-=-=-=-=-=-=-  COPYRIGHT  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Prayer2 is:
 *   Copyright (c) 1999-2001 by Tamara Cetinjski <tamara@mindnever.org>
 *   Copyright (c) 2001-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *   Copyright (c) 1999-2002 by MindNever Software Group <info@mindnever.org>
 *
 * Tamara Cetinjski is
 *   Copyright (c) 1999-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *
 *
 * =-=-=-=-=-=-=-=-=-=-  LICENSE  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Redistributions in any form must be accompanied by information on
 *    how to obtain complete source code for the Prayer2 software and any
 *    accompanying software that uses the Prayer2 software.  The source code
 *    must either be included in the distribution or be available for no
 *    more than the cost of distribution plus a nominal fee, and must be
 *    freely redistributable under reasonable conditions.  For an
 *    executable file, complete source code means the source code for all
 *    modules it contains.  It does not include source code for modules or
 *    files that typically accompany the major components of the operating
 *    system on which the executable file runs.
 *
 * =-=-=-=-=-=-=-=-=-=-  NO WARRANTY =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
 *
 * THIS SOFTWARE IS PROVIDED BY MINDNEVER SOFTWARE ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
 * NON-INFRINGEMENT, ARE DISCLAIMED.  IN NO EVENT SHALL MINDNEVER SOFTWARE
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, HAIR, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
#include <exec/exec.h>

#include <proto/exec.h>
#include <proto/graphics.h>
#include <proto/intuition.h>

#include <proto/dos.h>

#include "drag.h"

struct _drag {
		struct RastPort rp;
		struct GelsInfo gInfo;
		struct VSprite gelHead, gelTail;
		WORD gi_nextline[8];
		WORD *gi_lastcolor[8];
		struct collTable gi_collhandler;
	};


//#define AllocRaster(w, h) AllocMem(w * h / 8, MEMF_CHIP|MEMF_CLEAR)
//#define FreeRaster(p, w, h) FreeMem(p, w * h / 8)

BOB *CreateBob(DRAG *dr, int width, int height, int depth)
{
BOB *im;
int i, rasdepth, bpr;
UBYTE *pl;

if(depth < 1)
	return(NULL);

if(depth > 8)
	depth = 8;

if(dr)
	rasdepth = GetBitMapAttr(dr->rp.BitMap, BMA_DEPTH);
else
	rasdepth = depth;

if(rasdepth > 8)
	rasdepth = 8;

if(im = AllocMem(sizeof(*im), MEMF_CLEAR)) {

		im->dr = dr;

		im->ras_w = (width + 15) & ~0x0f;
		im->ras_h = height * (depth + rasdepth + 1) + 1;

		InitRastPort(&im->rp);
		InitBitMap(&im->bm, depth, im->ras_w, height);

		// Attach BitMap to this RastPort
		im->rp.BitMap = &im->bm;



		bpr = im->ras_w >> 3; // BytesPerRow

		if(pl = AllocRaster(im->ras_w, im->ras_h)) { // 

				BltClear(pl, im->ras_w * im->ras_h / 8, 1);

				im->rect.MinX = 0;
				im->rect.MinY = 0;
				im->rect.MaxX = width;
				im->rect.MaxY = height;

				im->im.Width = width;
				im->im.Height = height;
				im->im.Depth = depth;
				im->im.ImageData = (UWORD *) pl;
				im->im.PlanePick = (1 << depth) - 1;

				for(i = 0; i < depth; i++) {
						im->bm.Planes[i] = pl;
						pl += (bpr * height);
					}

				im->bob.BobVSprite = &im->vs;
				im->bob.Flags = 0;
				im->bob.ImageShadow = (UWORD *) pl; pl += (bpr * height);
				im->bob.SaveBuffer = (UWORD *)pl; pl += (bpr * height * rasdepth);

				im->vs.VSBob = &im->bob;
				im->vs.Flags = SAVEBACK | OVERLAY;
				im->vs.Width = bpr / 2;
				im->vs.Height = height;
				im->vs.Depth = depth;
				im->vs.ImageData = im->im.ImageData;
				im->vs.CollMask = im->bob.ImageShadow;
				im->vs.BorderLine = (UWORD *)pl;
				im->vs.PlanePick = im->im.PlanePick;
				im->vs.PlaneOnOff = im->im.PlaneOnOff;

				InitMasks(&im->vs);

			} else {
				FreeMem(im, sizeof(*im));
				im = NULL;
			}
	}

return(im);
}

void DeleteBob(BOB *im)
{
if(im) {
		HideBob(im);
		FreeRaster((PLANEPTR) im->im.ImageData, im->ras_w, im->ras_h);
		FreeMem(im, sizeof(*im));
	}
}


DRAG *CreateDrag(struct RastPort *rp)
{
DRAG *dr;

if(dr = AllocMem(sizeof(*dr), MEMF_ANY|MEMF_CLEAR)) {
//		CopyMem(rp, &dr->rp, sizeof(dr->rp));

		InitRastPort(&dr->rp);

//		dr->rp.Layer = rp->Layer;
		dr->rp.BitMap = rp->BitMap;

		dr->gInfo.sprRsrvd = 0;
		dr->gInfo.leftmost = dr->gInfo.topmost = 1;
		dr->gInfo.rightmost = (rp->BitMap->BytesPerRow << 3) - 1;
		dr->gInfo.bottommost = rp->BitMap->Rows - 1;
		dr->gInfo.nextLine = &dr->gi_nextline[0];
		dr->gInfo.lastColor = &dr->gi_lastcolor[0];
		dr->gInfo.collHandler = &dr->gi_collhandler;
		
		dr->rp.GelsInfo = &dr->gInfo;

		InitGels(&dr->gelHead, &dr->gelTail, &dr->gInfo);
	}
return(dr);
}

void DeleteDrag(DRAG *dr)
{
if(dr) {

		FreeMem(dr, sizeof(*dr));
	}
}

void ShowBob(BOB *b, int x, int y)
{
b->vs.X = x;
b->vs.Y = y;

if(b->rp_attach == NULL) {
		b->rp_attach = &b->dr->rp;
		AddBob(&b->bob, b->rp_attach);
		b->old_x = -1;
		b->old_y = -1;
	}


if((x != b->old_x) || (y != b->old_y)) {
		SortGList(b->rp_attach);
		DrawGList(b->rp_attach, NULL);
		b->old_x = x;
		b->old_y = y;
	}
}

void HideBob(BOB *b)
{
if(b->rp_attach) {
		RemIBob(&b->bob, b->rp_attach, NULL);
		b->rp_attach = NULL;
		b->old_x = -1;
		b->old_y = -1;
	}
}
