/*
 * =-=-=-=-=-=-=-=-=-=-=-  NOTICE  =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * The following is the license that applies to this copy of the Prayer2
 * software.  For a license to use the Prayer2 software under conditions
 * other than those described here, please contact MindNever Software at
 * following addresses:
 *
 * MindNever Software <info@mindnever.org>
 * MindNever Software +381 63 550-161 or +381 64 220-7115
 * 
 * =-=-=-=-=-=-=-=-=-=-  COPYRIGHT  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Prayer2 is:
 *   Copyright (c) 1999-2001 by Tamara Cetinjski <tamara@mindnever.org>
 *   Copyright (c) 2001-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *   Copyright (c) 1999-2002 by MindNever Software Group <info@mindnever.org>
 *
 * Tamara Cetinjski is
 *   Copyright (c) 1999-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *
 *
 * =-=-=-=-=-=-=-=-=-=-  LICENSE  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Redistributions in any form must be accompanied by information on
 *    how to obtain complete source code for the Prayer2 software and any
 *    accompanying software that uses the Prayer2 software.  The source code
 *    must either be included in the distribution or be available for no
 *    more than the cost of distribution plus a nominal fee, and must be
 *    freely redistributable under reasonable conditions.  For an
 *    executable file, complete source code means the source code for all
 *    modules it contains.  It does not include source code for modules or
 *    files that typically accompany the major components of the operating
 *    system on which the executable file runs.
 *
 * =-=-=-=-=-=-=-=-=-=-  NO WARRANTY =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
 *
 * THIS SOFTWARE IS PROVIDED BY MINDNEVER SOFTWARE ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
 * NON-INFRINGEMENT, ARE DISCLAIMED.  IN NO EVENT SHALL MINDNEVER SOFTWARE
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, HAIR, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
#define	__USE_SYSBASE
#include	<libraries/reqtools.h>

#include	<exec/memory.h>
#include	<workbench/startup.h>
#include	<libraries/pm.h>
#include	<exec/execbase.h>
#include	<string.h>
#include	<stdlib.h>
#include	<stdarg.h>
//#include	<dos.h>
#include	<devices/timer.h>
#include	<proto/rexxsyslib.h>
#include	<rexx/storage.h>
#include	<dopus/appicon.h>
#include	<inline/dopus.h>
#include	<clib/alib_protos.h>
//#include	<pragmas/dopus_pragmas.h>
#include	<intuition/intuitionbase.h>
#include	<intuition/gadgetclass.h>
//#include	<devices/newmouse.h>
#include <graphics/sprite.h>
#include <graphics/gels.h>
#include <graphics/gfx.h>


#include	<proto/dos.h>
#include	<proto/exec.h>
#include	<proto/intuition.h>
#include	<proto/reqtools.h>
#include	<proto/timer.h>
#include	<proto/layers.h>
#include	<proto/graphics.h>
#include	<proto/utility.h>
#include	<proto/pm.h>
#include	<proto/commodities.h>
#include	<proto/wb.h>                    
#include	<proto/icon.h>


#include	"prayer2_gui.h"
#include	"prayer2_audio.h"
#include	"prayer2_prefs.h"
#include	"prayer2_audiodriver.h"
#include	"robe.h"
#include	"amplifierplugin.h"

#include	"key.h"
#include	<zlib.h>

#include "hipscope.h"

#include	"version.h"

#include "PrayerListClass.h"

#include "opt68k.h"
#include "sprintf.h"


extern struct WBStartup *_WBenchMsg;
#define WBenchMsg _WBenchMsg


struct Prayer_Shared PShared;

char *Version_String = "\0$VER: Prayer2.0 " VERSION_STRING " " __DATE__ "  Copyright 1999-2001 by Tamara C.";

static struct rtFileRequester	*qFileReq = NULL;
static struct rtFileRequester	*qPlayListReq = NULL;
static struct List PlayList;
static APTR FastMemoryPool = NULL;
static struct Node *CurrentTrack = NULL;
static struct Node *SelectedPtr = NULL;

static APTR PlayHistory[256] = {NULL};
static UBYTE ph_Cursor = 0;

static struct Library *DOpusBase = NULL;
struct Library *CyberGfxBase = NULL, *CGXSystemBase = NULL;
struct Library *ReqToolsBase = NULL;
struct Library *WorkbenchBase = NULL;

ULONG	NumTracks = 0;

static struct List	plugin_list;
static struct SignalSemaphore plugin_lock;

static char Tag_Artist[256];
static char IconText[256];
static char CapturePath[512];

#define	PRAYER_PORT	"PRAYER2"
#define	SPEC_PTR ((APTR) 1)

enum {
	PMODE_RND,
	PMODE_SIN,
	PMODE_PGM,
	PMODE_SLP,
	PMODE_RTM,
	PMODE__MAX
	};

enum {
		PM_PREFS = 1,
		PM_ADD,
		PM_PREV,
		PM_NEXT,
		PM_STOP,
		PM_PAUSE,
		PM_PLAY,
		PM_ABOUT,
		PM_QUIT,
	};

static char *pmodes[] = {"RND", "SIN", "PGM", "SLP", "RTM"};

struct FreezeData {
		UWORD	PlayMode;
		UBYTE	Volume;
		UBYTE	DayAlarm_Enabled;
		char	FilePath[512];
		char	PlayListPath[512];
		char	PlayListName[110];
		WORD	SleepSeconds; // number of seconds
		UBYTE	DayAlarm_Hour;
		UBYTE	DayAlarm_Minute;
	} FreezeData = {2, 100, 0};

static struct PrayerCommodity {
	struct	NewBroker	NewBroker;
	struct	MsgPort	*CxPort;
	ULONG	CxMPFlag;
	CxObj	*Broker;
	CxObj *Hotkeys[PRHK_MAX];
	IX		ix[PRHK_MAX];
} *Comm;



static void hipemu_init(void);
static void hipemu_cleanup(void);
static void hipemu_update(void);

APTR PAlloc(ULONG size);
void PFree(APTR mem);
static struct Node *GetRND(void);
static void PatternRemove(struct Window *win);
static long LoadPlayList(struct IntuiMessage *imsg);
static void DisplayFileName(char *file);
static void DisplayAbout(struct IntuiMessage *imsg);
static void SortPlayList(void);
static void SavePlayList(void);
static void FreePlayList(void);
static void InitPlugins(void);
static void FreePluginsAll(void);
static void ReloadPlugins(void);
static APTR SelectPlayListFunc(struct IntuiMessage *imsg);
static void playlist_click(APTR tags, ULONG sec, ULONG msec);
static void EventLoop(struct Node *as);
static void smallChange(void);
static struct Node *AddNode(char *zpath);
static void PlayCurrent(void);
static void SelectNext(void);
static void SelectPrev(void);
static struct Node *AddEntry(char *dir, char *name, BOOL subdir);
static void AddFiles(void);
static BOOL DisplayTime(ULONG t1, ULONG t2, UWORD flags);

static void SetVolume(UBYTE vol);
static void SetupMpega(void);
static void SetupHotkeys(struct PrayerCommodity *c, struct PrayerPrefs *p);
static void FreePrayerCommodity(struct PrayerCommodity *c);
static struct PrayerCommodity *InitPrayerCommodity(void);
static void PlayPrev(void);
static void PlayNext(void);
static struct PopupMenu *MakePopup(void);
static struct Node *RemNode(struct Node *node);
static ULONG GetItemNumber(struct Node *node);
static struct Node *AddEntryLock(BPTR parent, char *zname);
static char *GetLockName(BPTR lock);
static void ShowPopupText(struct IntuiMessage *imsg);
static void DisplayAlarmTime(void);
static void CalculateLevel(struct WaveBuffer *wbuf, ULONG sampleoffs);
static BOOL ButtonFunctions(struct IntuiMessage *imsg, UWORD func, ULONG arg);

static BOOL HandleHotkey(int f);
static BOOL HandleRexxMsg(struct RexxMsg *msg);
static void AddPrayerIcon(void);
static void RemovePrayerIcon(void);
static ULONG SendRexxMsg(char *zport, char *ctl, ...);

static void mPrayerPause(void);
static void mPrayerResume(void);

static void SetPrayerIcon(char *ztitle);

char *PStrDup(char *str);
static void UpdateTimers(void);
static void SetSleepSeconds(WORD sec);
static void GenCapturePath(void);


#define	DTF_TIME	(1 << 0)
#define	DTF_POS		(1 << 1)

static BYTE	prefs_sig;
static BOOL paused = FALSE;
static struct PrayerPrefs *Prefs;
static struct PrayerRobe	*Robe;
static BOOL fhide = FALSE, frefresh = TRUE, fshow = TRUE;
static UWORD	dtf = DTF_TIME|DTF_POS;
static int	ignore_sigbreakf_e = -1;
static struct PrayerWindow	*closethis = NULL;


static struct MsgPort *AppPort = NULL;
static struct DiskObject *DiskIcon = NULL;
static struct AppIcon *AppIcon = NULL;


struct Library *PopupMenuBase = NULL;


static char *stpblk(char *str)
{
while(isspace(*str)) str++;
return(str);
}

static int open_libs(void)
{
DOpusBase = OpenLibrary("dopus5.library", 0L);
CyberGfxBase = OpenLibrary("cybergraphics.library", 0L);
CGXSystemBase = OpenLibrary("cgxsystem.library", 41L);

ReqToolsBase = OpenLibrary("reqtools.library", 0);
if(!ReqToolsBase)
	return(-1);

PopupMenuBase = OpenLibrary("popupmenu.library", 0);
if(!PopupMenuBase)
	return(-1);

WorkbenchBase = OpenLibrary("workbench.library", 0);
if(!WorkbenchBase)
	return(-1);

}

static void close_libs(void)
{
if(DOpusBase)
	CloseLibrary(DOpusBase);
if(CyberGfxBase)
	CloseLibrary(CyberGfxBase);
if(CGXSystemBase)
	CloseLibrary(CGXSystemBase);
if(PopupMenuBase)
	CloseLibrary(PopupMenuBase);
if(ReqToolsBase)
	CloseLibrary(ReqToolsBase);
if(WorkbenchBase)
	CloseLibrary(WorkbenchBase);
}


main(int argc, char **argv)
{
LONG	i;
struct	MsgPort	*mp = NULL;
struct	AppMessage *appm = NULL;
struct Process *me = (APTR)FindTask(0);
struct Node *node = NULL, *t;
char *ic;

NewList(&PlayList);

mp = FindPort(PRAYER_PORT);

if(((argc != 1) || (WBenchMsg && (WBenchMsg->sm_NumArgs > 1))) &&
		(mp)) {
		if(argc > 0)
			i = argc - 1;
		else
			i = 0;
		if(appm = AllocVec(sizeof(*appm) + (i * sizeof(struct WBArg)), MEMF_CLEAR)) {
				appm->am_Type = AMTYPE_APPICON;
				appm->am_Version = AM_VERSION;
				appm->am_Message.mn_ReplyPort = CreateMsgPort();

				if(argc) {
						appm->am_NumArgs = i;
						appm->am_ArgList = (APTR)(appm + 1);
						for(i = 0; i < appm->am_NumArgs; i++) {
								appm->am_ArgList[i].wa_Lock = me->pr_CurrentDir;
								appm->am_ArgList[i].wa_Name = argv[i + 1];
							}
					} else {
						appm->am_NumArgs = WBenchMsg->sm_NumArgs - 1;
						appm->am_ArgList = WBenchMsg->sm_ArgList + 1;
					}
			}
	}

if(appm) {
		Forbid();
		if(mp = FindPort(PRAYER_PORT))
			PutMsg(mp, (APTR) appm);
		Permit();
	}

if(appm) {
		if(mp) {
				mp = appm->am_Message.mn_ReplyPort;
				while(!GetMsg(mp)) {WaitPort(mp);}
			}
		DeleteMsgPort(appm->am_Message.mn_ReplyPort);
		FreeVec(appm);
	}
if(mp)
	return(0);

FastMemoryPool = CreatePool(MEMF_ANY|MEMF_CLEAR, 10240, 1024);
if(FastMemoryPool == NULL)
	return(20);

if(open_libs() < 0) {
		close_libs();
		return(20);
	}

GetVar("ENVARC:Prayer2.freeze", (APTR)&FreezeData, sizeof(FreezeData), GVF_BINARY_VAR|GVF_DONT_NULL_TERM);

if(qFileReq = rtAllocRequestA(RT_FILEREQ, NULL))
	rtChangeReqAttr(qFileReq, RTFI_Dir, (ULONG) FreezeData.FilePath, TAG_DONE);

if(qPlayListReq = rtAllocRequestA(RT_FILEREQ, NULL))
	rtChangeReqAttr(qPlayListReq, RTFI_Dir, (ULONG) FreezeData.PlayListPath, TAG_DONE);

prefs_sig = AllocSignal(-1);


InitPrayerPrefs(prefs_sig);

if(Prefs = GetPrayerPrefs()) {
		SetupMpega();
		PShared.WaveDump = Prefs->CapturePlugin;
		GenCapturePath();
		PShared.capture_name = CapturePath;

		if(Comm = InitPrayerCommodity())
			SetupHotkeys(Comm, Prefs);

		LoadPlayList(NULL);

		if(argc > 0) {
				ic = argv[0];
				for(i = 1; i < argc; i++) {
						if((t = AddNode(argv[i])) && (node == NULL))
							node = t;
					}
			} else {
				ic = WBenchMsg->sm_ArgList[0].wa_Name;
				for(i = 1; i < WBenchMsg->sm_NumArgs; i++) {
				if(WBenchMsg->sm_ArgList[i].wa_Lock)
						if((t = AddEntryLock(WBenchMsg->sm_ArgList[i].wa_Lock, WBenchMsg->sm_ArgList[i].wa_Name))
							&& (node == NULL)) {
								node = t;
							}
					}
			}

		DiskIcon = GetDiskObjectNew(ic);

		if(AppPort = CreateMsgPort()) {
				AppPort->mp_Node.ln_Pri = 0;
				AppPort->mp_Node.ln_Name = PRAYER_PORT;
				AddPort(AppPort);
			}
		
		InitPlugins();

		//hipemu_init();
		ampemu_init();

		if(Prefs->Flags & PRF_ALWAYS_ADD_ICON)
			AddPrayerIcon();

		EventLoop(node);

		ampemu_cleanup();
		//hipemu_cleanup();

		if(AppPort) {
				RemPort(AppPort);
				RemovePrayerIcon();
				DeleteMsgPort(AppPort);
			}

		if(DiskIcon)
			FreeDiskObject(DiskIcon);

		Prayer_Stop();

		ExitPrayerPrefs();
		FreePluginsAll();

		FreePrayerCommodity(Comm);

		FreePrayerPrefs(Prefs);
	}


if(qFileReq) {
		strncpy(FreezeData.FilePath, qFileReq->Dir, sizeof(FreezeData.FilePath) - 1);
		rtFreeRequest(qFileReq);
	}

if(qPlayListReq) {
		strncpy(FreezeData.PlayListPath, qPlayListReq->Dir, sizeof(FreezeData.PlayListPath) - 1);
		rtFreeRequest(qPlayListReq);
	}


DeletePool(FastMemoryPool);

close_libs();

FreeSignal(prefs_sig);
SetVar("ENVARC:Prayer2.freeze", (APTR)&FreezeData, sizeof(FreezeData), GVF_BINARY_VAR|GVF_DONT_NULL_TERM|GVF_GLOBAL_ONLY);
return(0);
}

static void SetupMpega(void)
{
PShared.req.freqdiv1 = Prefs->Layer3FreqDiv;
PShared.req.freqdiv2 = Prefs->Layer12FreqDiv;

PShared.req.channels1 = Prefs->Layer3ForceMono ? 1 : 2;
PShared.req.channels2 = Prefs->Layer12ForceMono ? 1 : 2;

PShared.req.quality1 = Prefs->Layer3Qual;
PShared.req.quality2 = Prefs->Layer12Qual;

PShared.RefreshRate = Prefs->RefreshRate;
PShared.AudioDriverID = Prefs->AudioDriverId;
PShared.MixFreq = Prefs->MixFreq;
PShared.AudioID = Prefs->AudioID;

PShared.WaveDump = Prefs->CapturePlugin;
}


static void e_HandleAmpEmu(BOOL *notdone)
{
APTR msg;
while(msg = GetMsg(ampport)) {
		ampemu_handlemsg(msg);
		ReplyMsg(msg);
	}
}

static void e_HandleAppMsg(BOOL *notdone)
{
struct AppMessage *appmsg;
struct AppSnapshotMsg *msg;
struct Node *node = NULL, *t;
int i;

while(appmsg = (APTR) GetMsg(AppPort)) {
		switch(appmsg->am_Type) {
				case	MTYPE_APPSNAPSHOT:
						msg = (APTR) appmsg;
						if(msg->flags & APPSNAPF_CLOSE)
							*notdone = FALSE;
						if(msg->flags & APPSNAPF_INFO)
							OpenPrayerWindow(Robe, 1);
						if(msg->flags & APPSNAPF_MENU) {
								switch(msg->id) {
										case	0:
											EditPrayerPrefs(Prefs->Screen);
											break;
										case	1:
											AddFiles();
											break;
										case	2:
											PlayPrev();
											break;
										case	3:
											PlayNext();
											break;
										case	4:
											PlayCurrent();
											break;
										case	5:
											Prayer_Stop();
											break;
										case	6:
											if(paused)
												mPrayerResume();
											else
												mPrayerPause();
											break;
									}
							}
					break;
				case	AMTYPE_APPWINDOW:
				case	AMTYPE_APPICON:
					if(appmsg->am_NumArgs == 0)
						fshow = TRUE;
					SetListerList(Robe, PRT_LISTER_PLAYLIST, (struct List *)~0);
					for(i = 0; i < appmsg->am_NumArgs; i++) {
							if(appmsg->am_ArgList[i].wa_Lock) {
									t = AddEntryLock(appmsg->am_ArgList[i].wa_Lock, appmsg->am_ArgList[i].wa_Name);
									if(node == NULL)
										node = t;
								}
						}
					SetListerList(Robe, PRT_LISTER_PLAYLIST, &PlayList);
					break;
			}
		ReplyMsg((APTR) appmsg);
	}

if(node) {
		CurrentTrack = (APTR)node;
		Prayer_Stop();
		PlayCurrent();
	}
}

static void e_HandleCxMsg(BOOL *notdone)
{
CxMsg	*cxmsg;
int cmd;

while(cxmsg = (APTR) GetMsg(Comm->CxPort)) {
		cmd = CxMsgID(cxmsg);
		switch(CxMsgType(cxmsg)) {
				case	CXM_COMMAND:
					switch(cmd) {
							case	CXCMD_KILL:
								*notdone = FALSE;
								break;
							case	CXCMD_APPEAR:
								fshow = TRUE;
								break;
							case	CXCMD_DISAPPEAR:
								fhide = TRUE;
								break;
							case	CXCMD_ENABLE:
								if(paused)
									mPrayerResume();
								break;
							case	CXCMD_DISABLE:
								if(!paused)
									mPrayerPause();
								break;
							}
					break;
				case	CXM_IEVENT:
					if(!HandleHotkey(cmd))
						*notdone = FALSE;
					break;
			}
		ReplyMsg((APTR) cxmsg);
	}
}

static void e_HandlePrefs(BOOL *notdone)
{
struct PrayerPrefs *NewPrefs;
struct PrayerRobe	*newRobe;
int i;

if(NewPrefs = GetPrayerPrefs()) {
		if(Stricmp(Prefs->Robe, NewPrefs->Robe)) {
				if(newRobe = InitPrayerRobe(AppPort, NewPrefs)) {
						FreePrayerRobe(Robe);
						Robe = newRobe;
					}
				frefresh = TRUE;
			} else if(((Prefs->Screen && NewPrefs->Screen) && Stricmp(Prefs->Screen, NewPrefs->Screen))
				|| (Prefs->Screen && !NewPrefs->Screen)
				|| (!Prefs->Screen && NewPrefs->Screen)) {
				SetPrayerScreen(Robe, NewPrefs->Screen);
				frefresh = TRUE;
			}

		FreePrayerPrefs(Prefs);
		Prefs = NewPrefs;

		SetupMpega();
		ReloadPlugins();

		SetupHotkeys(Comm, Prefs);

		if(Prefs->Flags & PRF_ALWAYS_ADD_ICON) {
				AddPrayerIcon();
			} else {
				for(i = 0; i < Robe->NumWindows; i++) {
						if(Robe->Windows[i]) {
								RemovePrayerIcon();
								break;
							}
					}
			}
	}
}

static void e_HandleDragging(BOB *bob)
{
BOOL dragging = TRUE, dlck = FALSE, ok = FALSE;
struct	IntuiMessage	*imsg;
struct TagItem *ti;
ULONG signals, sigmask, ticks = 0, ot = ~0;
int bob_x, bob_y;
APTR lcked = NULL;
LONG oldNr = -2, newNr = -2, scrl = 0, idle;
struct Node *This = NULL, *InsAfter = NULL;
struct DragMark *dm = NULL;
struct Rectangle old_rect = {0}, new_rect = {0};
struct RastPort *rp;
UWORD lp;

sigmask = (1L << Robe->UserPort->mp_SigBit) | SIGBREAKF_CTRL_D;

alarm(0, 100000);

idle = 0;

while(dragging) {

		if(IsListEmpty(&Robe->UserPort->mp_MsgList))
			signals = Wait(sigmask);

		if(signals & SIGBREAKF_CTRL_D) {
				alarm(0, 100000);
				idle++;
				if(idle > 10) {
						if(ticks == ot) {// avoiding deadlock
								dragging = FALSE;
								dlck = TRUE;
							} else {
								ot = ticks;
							}
					}
			}

		while(imsg = (APTR) GetMsg(Robe->UserPort)) {
				if(imsg->Class == IDCMP_IDCMPUPDATE) {
						if(ti = FindTagItem(PLA_Event, imsg->IAddress)) {
								switch(ti->ti_Data) {
										case PLE_MOVE: // move drag structure
											if(ti = FindTagItem(PLA_DragMark, imsg->IAddress)) {
													dm = (APTR) ti->ti_Data;
													memcpy(&new_rect, &dm->Rect, sizeof(old_rect));
													newNr = dm->Nr;
													This = dm->This;
													InsAfter = dm->InsAfter;
													scrl = dm->Scroll;
													rp = imsg->IDCMPWindow->RPort;
												}

											bob_x = imsg->IDCMPWindow->WScreen->MouseX;
											bob_y = imsg->IDCMPWindow->WScreen->MouseY;

											if(!lcked) {
													LockLayers(&imsg->IDCMPWindow->WScreen->LayerInfo);
													lcked = &imsg->IDCMPWindow->WScreen->LayerInfo;
												}
											break;
										case	PLE_UP:
											ok = TRUE;
										default: // PLE_UP, PLE_CANCEL
											dragging = FALSE;
											break;
									}
							}
					} else if(imsg->Class == IDCMP_INTUITICKS) {
						ticks++;
					}
				ReplyMsg((APTR) imsg);
			}

		lp = rp->LinePtrn;

		if(scrl) {

				rp->LinePtrn = 0xffff;
				rp->linpatcnt = 0;
				
				HideBob(bob);
				if(oldNr >= 0) {

						SetAPen(rp, dm->BgPen);
						Move(rp, old_rect.MinX, old_rect.MinY);
						Draw(rp, old_rect.MaxX, old_rect.MaxY);
						
						oldNr = -1;
					}
				
				if(lcked)
					UnlockLayers(lcked);

				HandleHotkey(scrl < 0 ? PRHK_LISTER_UP : PRHK_LISTER_DOWN);

				if(lcked)
					LockLayers(lcked);
							
			} else {

				if(oldNr != newNr) {
						rp->LinePtrn = 0xffff;
						rp->linpatcnt = 0;

						if(oldNr >= 0) {

								SetAPen(rp, dm->BgPen);
								Move(rp, old_rect.MinX, old_rect.MinY);

								HideBob(bob);
								Draw(rp, old_rect.MaxX, old_rect.MaxY);
							}

						rp->LinePtrn = 0x3333;
						rp->linpatcnt = 0;

						if(newNr >= 0) {

								SetAPen(rp, dm->FgPen);
								Move(rp, new_rect.MinX, new_rect.MinY);

								HideBob(bob);
								Draw(rp, new_rect.MaxX, new_rect.MaxY);
							}

						ShowBob(bob, bob_x, bob_y);

						oldNr = newNr;
						memcpy(&old_rect, &new_rect, sizeof(old_rect));
					}

				if(dm)
					ShowBob(bob, bob_x, bob_y);
			}

		rp->LinePtrn = lp;
	}

HideBob(bob);
if(oldNr >= 0) {
		lp = rp->LinePtrn;
		rp->LinePtrn = 0xffff;
		rp->linpatcnt = 0;

		SetAPen(rp, dm->BgPen);
		Move(rp, old_rect.MinX, old_rect.MinY);
		Draw(rp, old_rect.MaxX, old_rect.MaxY);

		rp->LinePtrn = lp;
	}

if(lcked)
	UnlockLayers(lcked);

//if(dlck)
//	Printf("Intuition Deadlock Avoided !\n");

if(ok && This && InsAfter) {

		SetListerList(Robe, PRT_LISTER_PLAYLIST, (APTR) ~0);
		Remove(This);

		This->ln_Pred = InsAfter;
		This->ln_Succ = InsAfter->ln_Succ;
		This->ln_Succ->ln_Pred = This;

		InsAfter->ln_Succ = This;

		SetListerList(Robe, PRT_LISTER_PLAYLIST, &PlayList);
	}
}

static void e_HandleIMsg(BOOL *notdone)
{
struct Gadget *g = NULL;
struct	IntuiMessage	*imsg;
struct PopupMenu *p = NULL;
struct InputEvent repeat_ie = {0};
struct TagItem *ti;
int hk, gp;
ULONG	arg;
BOB *bob = NULL;

enum {
	BOB_Nop = 0,
	BOB_Hide = (1 << 0),
	BOB_Move = (1 << 1)
	} bob_op = BOB_Nop;


static struct	Gadget	*rmb = NULL, *rmbb = NULL;
static ULONG	ticks_o = 0;


while(imsg = (APTR) GetMsg(Robe->UserPort)) {
		g = (struct Gadget *)imsg->IAddress;

		switch(imsg->Class) {
				case	IDCMP_RAWKEY:
					repeat_ie.ie_Class = IECLASS_RAWKEY;
					repeat_ie.ie_Code = imsg->Code;
					repeat_ie.ie_Qualifier = imsg->Qualifier;

					for(hk = 0; hk < PRHK_MAX; hk++) {
							if((Comm->ix[hk].ix_Class != IECLASS_NULL) && MatchIX(&repeat_ie, &Comm->ix[hk])) {
									if(!HandleHotkey(hk))
										*notdone = FALSE;
									break;
								}
						}
					break;
				case	IDCMP_IDCMPUPDATE:
					if((ti = FindTagItem(PLA_SelectedPtr, imsg->IAddress))
						&& (ti->ti_Data))
						SelectedPtr = (APTR) ti->ti_Data;
					
					if(ti = FindTagItem(PLA_Bob, imsg->IAddress))
						bob = (APTR) ti->ti_Data;

					if(ti = FindTagItem(PLA_Event, imsg->IAddress)) {
							switch(ti->ti_Data) {
									case PLE_DOWN:
										playlist_click(imsg->IAddress, imsg->Seconds, imsg->Micros);
										break;
									case PLE_MOVE: // move drag structure
										bob_op |= BOB_Move;
										// imsg->MouseX, imsg->MouseY
										break;
									default: // PLE_UP, PLE_CANCEL
										bob_op |= BOB_Hide;
										break;
								}
						}
					break;
				case	IDCMP_CLOSEWINDOW:
					closethis = (APTR)imsg->IDCMPWindow->UserData;
					break;
				case	IDCMP_MOUSEBUTTONS:
					if(imsg->Code == MENUUP) {
							if(rmb) {
									gp = RemoveGadget(imsg->IDCMPWindow, rmb);
									rmb->Flags &= ~GFLG_SELECTED;
									AddGadget(imsg->IDCMPWindow, rmb, gp);
									RefreshGList(rmb, imsg->IDCMPWindow, NULL, 1);
								}
							rmb = rmbb = NULL;
						}

					if((g = GetHitGadget(imsg)) && (g->GadgetType != GTYP_WDRAGGING)) {
							switch(imsg->Code) {
									case	MENUDOWN: {
											if(((struct PrayerButton *)g->UserData)->Function_RMB != 0xffff) {
													rmb = rmbb = g;
													gp = RemoveGadget(imsg->IDCMPWindow, g);
													g->Flags |= GFLG_SELECTED;
													AddGadget(imsg->IDCMPWindow, g, gp);
													RefreshGList(g, imsg->IDCMPWindow, NULL, 1);
												}
										}	break;
									case	MENUUP: {
											struct PrayerButton *pbutton = (APTR)g->UserData;
											if(pbutton)
													*notdone = ButtonFunctions(imsg, pbutton->Function_RMB, ((struct PrayerButton *)g->UserData)->Argument_RMB);
										}	break;
								}
						} else if((imsg->Code == MENUUP) &&
										(imsg->MouseX > 0) &&
										(imsg->MouseY > 0) &&
										(imsg->MouseX < imsg->IDCMPWindow->Width) &&
										(imsg->MouseY < imsg->IDCMPWindow->Height)) {
								if(p = MakePopup()) {
									hk = PM_OpenPopupMenu(imsg->IDCMPWindow, PM_Menu, (ULONG) p, PM_Code, MENUDOWN, TAG_DONE);
									switch(hk) {
											case	PM_PREFS:
												EditPrayerPrefs(Prefs->Screen);
												break;
											case	PM_QUIT:
												*notdone = FALSE;
												break;
											case	PM_ABOUT:
												OpenPrayerWindow(Robe, 1);
												break;
											case	PM_ADD:
												AddFiles();
												break;
											case	PM_PREV:
												PlayPrev();
												break;
											case	PM_NEXT:
												PlayNext();
												break;
											case	PM_PLAY:
												PlayCurrent();
												break;
											case	PM_STOP:
												Prayer_Stop();
												break;
											case	PM_PAUSE:
												if(paused)
													mPrayerResume();
												else
													mPrayerPause();
												break;
											default:
												if((hk & 0xFFFF0000) == 0xF00D0000) {
														struct PrayerWindow *x;
														if(x = OpenPrayerWindow(Robe, hk + 1)) {
																frefresh = TRUE;
															}
													}
												break;
										}
									PM_FreePopupMenu(p);
								} else EditPrayerPrefs(Prefs->Screen);
						}
					break;
				case	IDCMP_ACTIVEWINDOW:
				case	IDCMP_INACTIVEWINDOW:
					SetWindowState((APTR) (imsg->IDCMPWindow->UserData));
					break;
				case	IDCMP_GADGETUP:
					if(g->GadgetID & 0x0100)
						arg = ((struct PrayerSlider *)g->UserData)->Argument;
					else if(g->GadgetID & 0x0200)
						arg = ((struct PrayerLister *)g->UserData)->Argument;
					else
						arg = ((struct PrayerButton *)g->UserData)->Argument;

					*notdone = ButtonFunctions(imsg, g->GadgetID, arg);

					break;

				case	IDCMP_MOUSEMOVE:
					if(abs(ticks_o - imsg->Micros) > 300000) {
							g = GetHitGadget(imsg);
							HighLightPrayerButton((APTR)imsg->IDCMPWindow->UserData, g ? g->UserData : NULL);
							ticks_o = imsg->Micros;
						}

					if(imsg->Qualifier & IEQUALIFIER_RBUTTON) {
							g = GetHitGadget(imsg);

							if((g != rmbb) && rmb) {
									gp = RemoveGadget(imsg->IDCMPWindow, rmb);
									rmb->Flags &= ~GFLG_SELECTED;
									AddGadget(imsg->IDCMPWindow, rmb, gp);
									RefreshGList(rmb, imsg->IDCMPWindow, NULL, 1);
									rmb = NULL;
								} else if(g && (g == rmbb)) {
									gp = RemoveGadget(imsg->IDCMPWindow, g);
									g->Flags |= GFLG_SELECTED;
									AddGadget(imsg->IDCMPWindow, g, gp);
									RefreshGList(g, imsg->IDCMPWindow, NULL, 1);
									rmb = g;
								}
						}

					if(imsg->Qualifier & IEQUALIFIER_LEFTBUTTON) {
							if(g = GetActiveGadget((struct PrayerWindow *)(imsg->IDCMPWindow->UserData))) {
									switch(g->GadgetID) {
											case	(PRT_SLIDER_VOLUME | 0x0100):
												SetVolume(GetPrayerSlider(g));
												break;
											case	(PRT_SLIDER_POSITION | 0x0100):
												gp = PShared.dec.duration / 1000;
												DisplayTime(gp, GetPrayerSlider(g), DTF_TIME);
												dtf &= ~DTF_TIME;
												break;
											case	(PRT_SLIDER_SLEEP | 0x0100):
												SetSleepSeconds(GetPrayerSlider(g));
												break;
										}
								}
						}
					break;
			}
		ReplyMsg((APTR) imsg);
	}

if(bob) {
		if(!(bob_op & BOB_Hide) && (bob_op & BOB_Move))
			e_HandleDragging(bob);

		bob_op = BOB_Nop;
	}
}

static void e_HandleSigE(BOOL *notdone)
{
struct WaveBuffer *wbuf;
if(ignore_sigbreakf_e < 0) {
		wbuf = AudioStatus.CurrentBuffer;
		DisplayTime(PShared.dec.duration / 1000, wbuf ? (wbuf->Position / 1000) : 0, dtf);
		CalculateLevel(wbuf, AudioStatus.SampleOffs);
	} else {
		ignore_sigbreakf_e--;
	}
//hipemu_update();
ampemu_update();
}

static void e_HandleSigF(BOOL *notdone)
{
int st = PShared.status;
static int err_cnt = 0;

if(st == PSTATUS_ERROR_F)
	PShared.status = PSTATUS_EXITED;

smallChange();
switch(st) {
		case	PSTATUS_ERROR_F:
			err_cnt++;
			if(err_cnt > 5)
				break;
		case	PSTATUS_SONGEND:
			SelectNext();
			PlayCurrent();
			err_cnt = 0;
			break;
		case	PSTATUS_PRAYING:
			err_cnt = 0;
			break;
	}
/*
if(err_cnt > 5) {
		PShared.status = PSTATUS_EXITED;
		rtEZRequest("Unrecoverable file error on %s!", "Ok", NULL, NULL, NULL, PShared.req.filename);
	}
*/
}

static void e_HandleChanges(BOOL *notdone)
{
int cnt;

if(fshow) {
		ShowPrayerWindows(Robe);

		if(!(Prefs->Flags & PRF_ALWAYS_ADD_ICON))
			RemovePrayerIcon();

		fshow = FALSE;
		frefresh = TRUE;
	}

if(fhide) {
		HidePrayerWindows(Robe);

		fhide = FALSE;
		frefresh = FALSE;
	}

if(frefresh) {
		SetVolume(FreezeData.Volume);
		UpdatePrayerButton(Robe, PRT_BUTTON_ALARM, FreezeData.DayAlarm_Enabled);
		UpdatePrayerButton(Robe, PRT_BUTTON_PAUSE, paused);
		Signal(SysBase->ThisTask, SIGBREAKF_CTRL_F|SIGBREAKF_CTRL_E);
		SetListerList(Robe, PRT_LISTER_PLAYLIST, &PlayList);
		SetListerHighlight(Robe, PRT_LISTER_PLAYLIST, GetItemNumber((APTR)CurrentTrack));

		SetSleepSeconds(FreezeData.SleepSeconds);
		frefresh = FALSE;
	}

if(closethis) {
		ClosePrayerWindow(Robe, closethis);
		closethis = NULL;
	}

for(cnt = 0; cnt < Robe->NumWindows; cnt++) {
		if(Robe->Windows[cnt])
			break;
	}

if(cnt == Robe->NumWindows) // All windows closed ?!
	AddPrayerIcon();
}

static void EventLoop(struct Node *node)
{
BOOL notdone = TRUE;
ULONG signals, sigmask;

if(Robe = InitPrayerRobe(AppPort, Prefs)) {

		OpenPrayerWindow(Robe, 1);

		if(node) {
				CurrentTrack = node;
				PlayCurrent();
			}

		if(FreezeData.PlayMode >= PMODE__MAX)
			FreezeData.PlayMode = PMODE__MAX - 1;

		UpdatePrayerDisplay(Robe, PRT_DISPLAY_PLAYMODE, pmodes[FreezeData.PlayMode]);

		sigmask = SIGBREAKF_CTRL_F|SIGBREAKF_CTRL_E|SIGBREAKF_CTRL_C|SIGBREAKF_CTRL_D;

		if(Comm)
			sigmask |= Comm->CxMPFlag;

		if(prefs_sig >= 0)
			sigmask |= 1L << prefs_sig;

		if(AppPort)
			sigmask |= 1L << AppPort->mp_SigBit;

		if(ampport)
			sigmask |= 1L << ampport->mp_SigBit;

		alarm(1,0);

		while(notdone) {

				signals = Wait(sigmask | (1L << Robe->UserPort->mp_SigBit));

				if(signals & SIGBREAKF_CTRL_C)
					notdone = FALSE;

				if(signals & (1L << ampport->mp_SigBit))
					e_HandleAmpEmu(&notdone);

				if(signals & (1L << AppPort->mp_SigBit))
					e_HandleAppMsg(&notdone);

				if(signals & (1L << Comm->CxPort->mp_SigBit))
					e_HandleCxMsg(&notdone);

				if((prefs_sig >= 0) && (signals & (1L << prefs_sig)))
					e_HandlePrefs(&notdone);

				if(signals & (1L << Robe->UserPort->mp_SigBit))
					e_HandleIMsg(&notdone);

				if(signals & SIGBREAKF_CTRL_E)
					e_HandleSigE(&notdone);

				if(signals & SIGBREAKF_CTRL_F)
					e_HandleSigF(&notdone);

				if(signals & SIGBREAKF_CTRL_D)
					UpdateTimers();

				e_HandleChanges(&notdone);

			}
		FreePrayerRobe(Robe);
	}
}

static BOOL ButtonFunctions(struct IntuiMessage *imsg, UWORD func, ULONG arg)
{
BOOL notdone = TRUE;
APTR p;

switch(func) {
		case	PRT_BUTTON_PREFS:
			EditPrayerPrefs(Prefs->Screen);
			break;
		case	PRT_BUTTON_ABOUT:
			DisplayAbout(imsg);
			break;
		case	PRT_BUTTON_TEXT:
			ShowPopupText(imsg);
			break;
		case	PRT_BUTTON_QUIT:
			notdone = FALSE;
			break;
		case	PRT_BUTTON_ICONIFY:
			fhide = TRUE;
			break;
		case	PRT_BUTTON_OPENWINDOW: {
					struct PrayerWindow *x;
					if(x = OpenPrayerWindow(Robe, arg)) {
							frefresh = TRUE;
						}
				}
			break;
		case	PRT_BUTTON_PLAYMODE:
			FreezeData.PlayMode = (FreezeData.PlayMode + 1) % PMODE__MAX;
			UpdatePrayerDisplay(Robe, PRT_DISPLAY_PLAYMODE, pmodes[FreezeData.PlayMode]);
			break;
		case	PRT_BUTTON_PREV:
			PlayPrev();
			break;
		case	PRT_BUTTON_NEXT:
			PlayNext();
			break;
		case	PRT_BUTTON_PLAY:
			PlayCurrent();
			break;
		case	PRT_BUTTON_STOP:
			Prayer_Stop();
			break;
		case	PRT_BUTTON_REM:
			if(!arg)
				SelectedPtr = RemNode(SelectedPtr);
			else {
					if(p = PM_MakeMenu(
					PM_Item, (ULONG) PM_MakeItem(PM_Title, (ULONG)"Remove", PM_Underlined, TRUE, PM_NoSelect, TRUE, PM_Center, TRUE, TAG_DONE),

											PM_Item, (ULONG) PM_MakeItem(PM_Title, (ULONG)"Single", PM_UserData, 1, PM_Center, TRUE, TAG_DONE),
											PM_Item, (ULONG) PM_MakeItem(PM_Title, (ULONG)"Pattern", PM_UserData, 2, PM_Center, TRUE, TAG_DONE),
											PM_Item, (ULONG) PM_MakeItem(PM_Title, (ULONG)"All", PM_UserData, 3, PM_Center, TRUE, TAG_DONE),
											TAG_DONE)) {
							int i = PM_OpenPopupMenu(imsg->IDCMPWindow, PM_Menu, (ULONG) p, PM_Code, MENUDOWN, TAG_DONE);
							switch(i) {
									case	1:
										SelectedPtr = RemNode(SelectedPtr);
										break;
									case	2:
										PatternRemove(imsg->IDCMPWindow);
										break;
									case	3:
										FreePlayList();
										SetListerList(Robe, PRT_LISTER_PLAYLIST, &PlayList);
										break;
								}
							PM_FreePopupMenu(p);
						}
				}
			break;
		case	PRT_BUTTON_SORTLIST:
			SortPlayList();
			break;
		case	PRT_BUTTON_LOADLIST:
			LoadPlayList(imsg);
			break;
		case	PRT_BUTTON_SAVELIST:
			SavePlayList();
			break;
		case	PRT_BUTTON_ADDFILES:
			AddFiles();
			break;
		case	PRT_BUTTON_ALARM:
			FreezeData.DayAlarm_Enabled = ((struct Gadget *)(imsg->IAddress))->Flags & GFLG_SELECTED;
			DisplayAlarmTime();
			break;
		case	PRT_BUTTON_SETHOUR:
			if(FreezeData.DayAlarm_Enabled) {
					FreezeData.DayAlarm_Hour++;
					if(FreezeData.DayAlarm_Hour > 23)
						FreezeData.DayAlarm_Hour = 0;
					DisplayAlarmTime();
				}
			break;
		case	PRT_BUTTON_SETMINUTE:
			if(FreezeData.DayAlarm_Enabled) {
					FreezeData.DayAlarm_Minute++;
					if(FreezeData.DayAlarm_Minute > 59)
						FreezeData.DayAlarm_Minute = 0;
					DisplayAlarmTime();
				}
			break;
		case	PRT_BUTTON_PAUSE:
			if(((struct Gadget *)(imsg->IAddress))->Flags & GFLG_SELECTED)
				mPrayerPause();
			else
				mPrayerResume();
			break;
		case	(PRT_SLIDER_VOLUME | 0x0100):
			SetVolume(GetPrayerSlider((APTR) imsg->IAddress));
			break;
		case	(PRT_SLIDER_SLEEP | 0x0100):
			SetSleepSeconds(GetPrayerSlider((APTR) imsg->IAddress));
			break;
		case	(PRT_SLIDER_POSITION | 0x0100):
			Prayer_Skip(GetPrayerSlider((APTR) imsg->IAddress) * 1000);
			dtf |= DTF_TIME;
			SetSignal(0L, SIGBREAKF_CTRL_E);
			ignore_sigbreakf_e = 3;
			break;
		case	PRT_BUTTON_RECORD:
			if(((struct Gadget *)(imsg->IAddress))->Flags & GFLG_SELECTED) {
					GenCapturePath();
					PShared.capture_enabled = TRUE;
				} else {
					PShared.capture_enabled = FALSE;
				}
			break;
		case	PRT_BUTTON_GROUP:
			SwitchGroup((APTR)imsg->IDCMPWindow->UserData, arg);
			break;
	}
return(notdone);
}

static void GenCapturePath(void)
{
struct DateStamp ds;
struct	ClockData cd;
ULONG	isec;

DateStamp(&ds);

isec  = ds.ds_Tick / TICKS_PER_SECOND;
isec += ds.ds_Minute * 60;
isec += ds.ds_Days * 86400;

Amiga2Date(isec, &cd);
//                    YYYYmmddHHMMSS
SPrintf(CapturePath, "%s%04ld%02ld%02ld%02ld%02ld%02ld%02ld",
										Prefs->CapturePath, cd.year, cd.month, cd.mday, cd.hour, cd.min, cd.sec,
										ds.ds_Tick % TICKS_PER_SECOND);
}


static void CalculateLevel(struct WaveBuffer *wbuf, ULONG sampleoffs)
{
LONG nsamples;
int level_l = 0, level_r = 0;
static int l_old = 0, r_old = 0;

if(wbuf) {

		nsamples = wbuf->Length - sampleoffs;

#ifndef __MORPHOS__
		if(nsamples > 0) {
				level_l = calc_level(wbuf->Ch1 + sampleoffs, nsamples);

				if(PShared.dec.channels == 2)
					level_r = calc_level(wbuf->Ch2 + sampleoffs, nsamples);
				else
					level_r = level_l;
			}
#endif

	}

if((l_old != level_l) || (r_old != level_r))
	UpdatePrayerLevel(Robe, level_r, level_l);

l_old = level_l;
r_old = level_r;
}

static void SetPrayerIcon(char *ztitle)
{
static char lasttitle[64] = {0};
static char *states[] = {
		"Sleeping",
		"Boring",
		"Eating",
		"Looking Nice",
		"Reading Book",
		"Waiting",
		"Dying",
	};

if(ztitle)
	strncpy(lasttitle, ztitle, sizeof(lasttitle) - 1);
else
	ztitle = lasttitle;

if(DOpusBase && AppIcon) {
		if(PShared.status == PSTATUS_PRAYING)
			SPrintf(IconText, "Prayer2: %s (%s)", paused ? "Paused" : "Playing", ztitle);
		else
			SPrintf(IconText, "Prayer2: %s", states[SysBase->DispCount % 7]);

		ChangeAppIcon(AppIcon, NULL, NULL, IconText, CAIF_TITLE);
	}
}

static void smallChange()
{
char	ab[64];
char	*zq = "LMH";

if(PShared.tags.Title) {
		UpdatePrayerDisplay(Robe, PRT_DISPLAY_SONGTITLE, PShared.tags.Title);
		SetPrayerIcon(PShared.tags.Title);
	} else SetPrayerIcon(NULL);

if(PShared.tags.Album)
	UpdatePrayerDisplay(Robe, PRT_DISPLAY_ALBUM, PShared.tags.Album);

if(PShared.tags.Other)
	UpdatePrayerDisplay(Robe, PRT_DISPLAY_OTHER, PShared.tags.Other);

if(PShared.dec.frequency) {
		SPrintf(ab, "%02.2ldk %s", PShared.dec.frequency / 1000, (PShared.dec.channels == 2) ? "Stereo" : "Mono" /*, zq[PShared.dec.quality] */);
		UpdatePrayerDisplay(Robe, PRT_DISPLAY_OUTPUT, ab);
	}

if(PShared.tags.Artist && PShared.tags.Artist[0])
	zq = PShared.tags.Artist;
else
	zq = Tag_Artist;
UpdatePrayerDisplay(Robe, PRT_DISPLAY_ARTIST, zq);

if(!PShared.capture_enabled)
	UpdatePrayerButton(Robe, PRT_BUTTON_RECORD, PShared.capture_enabled);
}

static BOOL DisplayTime(ULONG t1, ULONG t2, UWORD dtf)
{
char	ab[32];
static ULONG to1 = 0, to2 = 0;

if((t1 != to1) || (t2 != to2)) {
		if(dtf & DTF_TIME) {
				SPrintf(ab, "%02ld:%02ld %02ld:%02ld", t1 / 60, t1 % 60, t2 / 60, t2 % 60);
				UpdatePrayerDisplay(Robe, PRT_DISPLAY_TIME, ab);
				if(PShared.capture_enabled)
					UpdatePrayerDisplay(Robe, PRT_DISPLAY_RECCURRENT, ab + 6);
				else
					UpdatePrayerDisplay(Robe, PRT_DISPLAY_RECSTART, ab + 6);

				to1 = t1;to2 = t2;
			}
		if(dtf & DTF_POS) {
				UpdatePrayerSlider(Robe, PRT_SLIDER_POSITION, t1, t2);
			}
		return(TRUE);
	}
return(FALSE);
}

static /* __saveds */ int pplAddNodeCallback(char *path)
{

if(AddNode(path))
	return(TRUE);
else
	return(FALSE);
}

static struct Node *AddNode(char *zpath)
{
struct ListNode *node;

if(node = PAlloc(sizeof(*node))) {
		if(node->storage = PAlloc(strlen(zpath) + 1)) {
				strcpy(node->storage, zpath);
				node->node.ln_Name = FilePart(node->storage);
				AddTail(&PlayList, (APTR) node);
				NumTracks++;
			} else {
				PFree(node);
				node = NULL;
			}
	}
return((APTR)node);
}

static ULONG GetItemNumber(struct Node *node)
{
ULONG i = 0;
if(node) {
		while(node->ln_Pred) {
				node = node->ln_Pred;
				i++;
			}
	}
return(i - 1);
}

static void DisplayFileName(char *file)
{
char *zp, *zx;
ULONG i;

i = strlen(file);

if(zp = PAlloc(i + 1)) {
		strcpy(zp, file);
		while(i--) {
				if(zp[i] == '.') {
						zp[i] = 0;
						break;
					}
			}
		if(zx = strchr(zp, '-')) {// group separator
				*zx++ = 0;
				zx = stpblk(zx);
				strncpy(Tag_Artist, zp, sizeof(Tag_Artist) - 1);
			} else {
				Tag_Artist[0] = 0;
				zx = zp;
			}

		UpdatePrayerDisplay(Robe, PRT_DISPLAY_SONGTITLE, zx);
		SetPrayerIcon(zx);
		PFree(zp);
	}
}

static void PlayCurrent()
{
if((CurrentTrack == NULL) || (CurrentTrack->ln_Succ == NULL))
	SelectNext();

if(CurrentTrack->ln_Succ) {
		DisplayFileName(CurrentTrack->ln_Name);
		SetListerHighlight(Robe, PRT_LISTER_PLAYLIST, GetItemNumber(CurrentTrack));

		GenCapturePath();

		Prayer_Play(((struct ListNode *)CurrentTrack)->storage, Prefs->PlayerPri);
		mPrayerResume();

		PlayHistory[ph_Cursor++] = CurrentTrack;
	}
}


static void SelectPrev(void)
{
if(CurrentTrack == NULL) {
		CurrentTrack = (APTR)PlayList.lh_Tail;
		return;
	}
if(CurrentTrack->ln_Pred == (struct Node *)&PlayList.lh_Head) {
		CurrentTrack = (APTR)PlayList.lh_Tail;
		return;
	}
if(CurrentTrack->ln_Pred)
	CurrentTrack = CurrentTrack->ln_Pred;
}

static void SelectNext(void)
{
switch(FreezeData.PlayMode) {
		case	PMODE_RTM:
			if(SelectedPtr) {
					CurrentTrack = SelectedPtr;
					return;
				}
		case	PMODE_SLP:
		case	PMODE_PGM:
			if(CurrentTrack == NULL) {
					CurrentTrack = (APTR) PlayList.lh_Head;
					return;
				}
			if(CurrentTrack->ln_Succ == NULL) {
					CurrentTrack = (APTR) PlayList.lh_Head;
					return;
				}
			if(CurrentTrack->ln_Succ)
				CurrentTrack = CurrentTrack->ln_Succ;
			break;
		case	PMODE_RND:
			CurrentTrack = GetRND();
			break;
		case	PMODE_SIN:
			if((CurrentTrack == NULL) || (CurrentTrack->ln_Succ == NULL))
				CurrentTrack = (APTR) PlayList.lh_Head;
	}
}

static void AddFiles(void)
{
struct rtFileList *qlist, *qnext;
static char	ab[110] = {0};

if(qFileReq) {
		if(qlist = rtFileRequest(qFileReq, ab, "Select Files And Directories", 
													RT_LockWindow, TRUE,
													RTFI_Flags, FREQF_MULTISELECT|FREQF_SELECTDIRS|FREQF_PATGAD,
													RTFI_OkText, (ULONG) "Select",
													RT_PubScrName, (ULONG) Prefs->Screen,
													TAG_DONE)) {
				SetListerList(Robe, PRT_LISTER_PLAYLIST, (struct List *)~0);
				for(qnext = qlist; qnext; qnext = qnext->Next) {
						AddEntry(qFileReq->Dir, qnext->Name, (qnext->StrLen == 0xffffffff) ? TRUE : FALSE);
					}
				SetListerList(Robe, PRT_LISTER_PLAYLIST, &PlayList);
				rtFreeFileList(qlist);
			}
	}
}

static struct Node *AddEntry(char *dir, char *name, BOOL subdir)
{
char *zbuffer;
BPTR lock;
struct FileInfoBlock *fib;
ULONG i;
struct Node *node = NULL, *t;

i = strlen(dir) + 110;

if(zbuffer = PAlloc(i)) {
		strcpy(zbuffer, dir);
		AddPart(zbuffer, name, i);
		if(subdir) {
				if(lock = Lock(zbuffer, ACCESS_READ)) {
						if(fib = PAlloc(sizeof(*fib))) {
								if(Examine(lock, fib)) {
										if(fib->fib_DirEntryType >= 0) {
												while(ExNext(lock, fib)) {
														t = AddEntry(zbuffer, fib->fib_FileName, (fib->fib_DirEntryType >= 0) ? TRUE : FALSE);
														if((node == NULL) && (fib->fib_DirEntryType < 0))
															node = t;
													}
											} else {
												node = (APTR)AddNode(zbuffer);
											}
									}
								PFree(fib);
							}
						UnLock(lock);
					}
			} else {
				node =(APTR) AddNode(zbuffer);
			}
		PFree(zbuffer);
	}
return(node);
}

static struct Node *AddEntryLock(BPTR parent, char *zname)
{
struct Node *node = NULL;
char	*zparent;

if(zparent = GetLockName(parent)) {
		node = AddEntry(zparent, zname, TRUE);
		PFree(zparent);
	}
return(node);
}

static char *GetLockName(BPTR lock)
{
char	*zname = NULL, *rc;
LONG	i = 256;

while(1) {
		if(zname = PAlloc(i)) {
				if(NameFromLock(lock, zname, i))
					break;
				else {
						if(IoErr() == ERROR_LINE_TOO_LONG) {
								i += 256;
								PFree(zname);
							} else break;
					}
			} else break;
	}

rc = PStrDup(zname);
PFree(zname);
return(rc);
}

static void SetVolume(UBYTE vol)
{
char ab[32];

FreezeData.Volume = vol;
Prayer_Volume(vol);
UpdatePrayerSlider(Robe, PRT_SLIDER_VOLUME, 201, vol);

SPrintf(ab, "%ld%%", vol);

UpdatePrayerDisplay(Robe, PRT_DISPLAY_VOLUME, ab);
}

static struct PrayerCommodity *InitPrayerCommodity(void)
{
struct PrayerCommodity *c;
BOOL	rc = FALSE;
int i;

if(c = PAlloc(sizeof(*c))) {
		if(c->CxPort = CreateMsgPort()) {
				c->CxMPFlag = (1L << c->CxPort->mp_SigBit);
				c->NewBroker.nb_Version = NB_VERSION;
				c->NewBroker.nb_Name = "Prayer2";
				c->NewBroker.nb_Title = "Prayer 2.0";
				c->NewBroker.nb_Descr = "Nice and faast MPEG-Audio player, (c) 1999-2001 Tamara^MindNever";
				c->NewBroker.nb_Flags = COF_SHOW_HIDE;
				c->NewBroker.nb_Port = c->CxPort;
				if(c->Broker = CxBroker(&c->NewBroker, NULL)) {

						for(i = 0; i < PRHK_MAX; i++) {
								c->Hotkeys[i] = CxFilter(NULL);
								ActivateCxObj(c->Hotkeys[i], FALSE);

								AttachCxObj(c->Broker, c->Hotkeys[i]);
								AttachCxObj(c->Hotkeys[i], CxSender(c->CxPort, i));
								AttachCxObj(c->Hotkeys[i], CxTranslate(NULL));
							}

						ActivateCxObj(c->Broker, TRUE);
						rc = TRUE;
					}
			}
	}
if(rc)
	return(c);
FreePrayerCommodity(c);
return(NULL);
}

static void SetupHotkeys(struct PrayerCommodity *c, struct PrayerPrefs *p)
{
int i;

for(i = 0; i < PRHK_MAX; i++) {
		ActivateCxObj(c->Hotkeys[i], FALSE);

		c->ix[i].ix_Class = IECLASS_NULL;
		c->ix[i].ix_Version = IX_VERSION;

		ParseIX(p->Hotkeys[i].Key, &c->ix[i]);

		if(c->ix[i].ix_Class != IECLASS_NULL) {
				if(p->Hotkeys[i].Repeat)
					c->ix[i].ix_QualMask &= ~IEQUALIFIER_REPEAT;

				if(p->Hotkeys[i].Global) {
						ClearCxObjError(c->Hotkeys[i]);
						SetFilterIX(c->Hotkeys[i], &c->ix[i]);
						if(!CxObjError(c->Hotkeys[i]))
							ActivateCxObj(c->Hotkeys[i], TRUE);
					}					
			}
	}
}


static void FreePrayerCommodity(struct PrayerCommodity *c)
{
struct	Message *msg;
if(c) {
		DeleteCxObjAll(c->Broker);
		if(c->CxPort) {
				while(msg = GetMsg(c->CxPort))
					ReplyMsg(msg);
				DeleteMsgPort(c->CxPort);
			}
		PFree(c);
	}
}

static void PlayPrev()
{
SelectPrev();
Prayer_Stop();
PlayCurrent();
}

static void PlayNext()
{
SelectNext();
Prayer_Stop();
PlayCurrent();
}

char *PStrDup(char *str)
{
char *rc = NULL;
if(str && (rc = PAlloc(strlen(str) + 1)))
	strcpy(rc, str);
return(rc);
}

char *StrDup(char *str)
{
char *rc = NULL;
if(str && (rc = AllocVec(strlen(str) + 1, MEMF_ANY))) // !!!!!!!!!!
	strcpy(rc, str);
return(rc);
}


static struct PopupMenu *MakePopup(void)
{
ULONG	i = 0, j = 0;
APTR p = NULL;
struct TagItem *Tags, *winTags;
struct PrayerPlugin *ppl;

ObtainSemaphore(&PShared.PluginLock);
Tags = PAlloc((PShared.PluginCount + 2) * sizeof(*Tags));
winTags = PAlloc((Robe->NumWindows + 2) * sizeof(*winTags));

if(Tags && winTags) {
		for(ppl = (APTR) PShared.PluginList.lh_Head; ppl->Node.ln_Succ; ppl = (APTR) ppl->Node.ln_Succ) {
				if(i < PShared.PluginCount) {
						Tags[i].ti_Tag = PM_Item;
						Tags[i].ti_Data = (ULONG)PM_MakeItem(PM_Title, (ULONG) ppl->Title, PM_Checkit, TRUE, PM_Checked, ppl->Enabled, PM_AutoStore, (ULONG) &ppl->Enabled, TAG_DONE);
						i++;
					} else break;
			}

		if(!PShared.PluginCount) {
				Tags[i].ti_Tag = PM_Item;
				Tags[i].ti_Data = (ULONG)PM_MakeItem(PM_Title, (ULONG) "No plugins installed",PM_NoSelect, TRUE, TAG_DONE);
				i++;
			}

		Tags[i].ti_Tag = TAG_DONE;

		ReleaseSemaphore(&PShared.PluginLock);

		for(i = 0; i < Robe->NumWindows; i++) {
				if(Robe->RWindows[i] && (Robe->RWindows[i]->ChunkSize > 24) && Robe->RWindows[i]->Title[0]) {
						winTags[j].ti_Tag = PM_Item;
						winTags[j].ti_Data = (ULONG) PM_MakeItem(PM_Title, (ULONG) Robe->RWindows[i]->Title, PM_UserData, 0xF00D0000 | i, TAG_DONE);
						j++;
					}
			}

		if(j == 0) {
				winTags[j].ti_Tag = PM_Item;
				winTags[j].ti_Data = (ULONG) PM_MakeItem(PM_Title, (ULONG) "(no window titles defined)", TAG_DONE);
				j++;
			}

		winTags[j].ti_Tag = TAG_DONE;

		p = PM_MakeMenu(
			PM_Item, (ULONG) PM_MakeItem(PM_Title, (ULONG) "Prefs", PM_UserData, PM_PREFS, TAG_DONE),
			PM_Item, (ULONG) PM_MakeItem(PM_Title, (ULONG) "Plugins", PM_Sub, (ULONG) PM_MakeMenuA(Tags), TAG_DONE),
			PM_Item, (ULONG) PM_MakeItem(PM_Title, (ULONG) "Windows", PM_Sub, (ULONG) PM_MakeMenuA(winTags), TAG_DONE),
			PM_Item, (ULONG) PM_MakeItem(PM_TitleBar, TRUE, TAG_DONE),
			PM_Item, (ULONG) PM_MakeItem(PM_Title, (ULONG) "Add", PM_UserData, PM_ADD, TAG_DONE),
			PM_Item, (ULONG) PM_MakeItem(PM_Title, (ULONG) "Prev", PM_UserData, PM_PREV, TAG_DONE),
			PM_Item, (ULONG) PM_MakeItem(PM_Title, (ULONG) "Next", PM_UserData, PM_NEXT, TAG_DONE),
			PM_Item, (ULONG) PM_MakeItem(PM_Title, (ULONG) "Stop", PM_UserData, PM_STOP, TAG_DONE),
			PM_Item, (ULONG) PM_MakeItem(PM_Title, (ULONG) (paused ? "Resume" : "Pause"), PM_UserData, PM_PAUSE, TAG_DONE),
			PM_Item, (ULONG) PM_MakeItem(PM_Title, (ULONG) "Play", PM_UserData, PM_PLAY, TAG_DONE),
			PM_Item, (ULONG) PM_MakeItem(PM_TitleBar, TRUE, TAG_DONE),
			PM_Item, (ULONG) PM_MakeItem(PM_Title, (ULONG) "About", PM_UserData, PM_ABOUT, TAG_DONE),
			PM_Item, (ULONG) PM_MakeItem(PM_Title, (ULONG) "Quit",PM_UserData, PM_QUIT, TAG_DONE),
			TAG_DONE);
	} else ReleaseSemaphore(&PShared.PluginLock);

PFree(Tags);
PFree(winTags);
return(p);
}

static void playlist_click(APTR tags, ULONG sec, ULONG msec)
{
static ULONG osec = 0, omsec = 0;
static APTR os = NULL;

if((SelectedPtr == os)
	&& DoubleClick(osec, omsec, sec, msec)) {

		CurrentTrack = SelectedPtr;

		Prayer_Stop();
		PlayCurrent();				
	}

os = SelectedPtr;
osec = sec;
omsec = msec;
}

static struct Node *RemNode(struct Node *node)
{
struct Node *newnode = NULL;
if(node && node->ln_Succ && node->ln_Pred) {
		if(node->ln_Succ->ln_Succ)
			newnode = node->ln_Succ;
		else if(node->ln_Pred->ln_Pred)
			newnode = node->ln_Pred;

		if(node == CurrentTrack) {
				Prayer_Stop();
				CurrentTrack = newnode;
			}
		
		SetListerList(Robe, PRT_LISTER_PLAYLIST, (struct List *)~0);
		Remove(node);
		PFree(node);
		SetListerList(Robe, PRT_LISTER_PLAYLIST, &PlayList);
		NumTracks--;
	}
return(newnode);
}

static void DisplayAbout(struct IntuiMessage *imsg)
{
APTR p;

if(p = PM_MakeMenu(
				PM_Item, (ULONG) PM_MakeItem(PM_Title, (ULONG) "The Prayer  - Release V2.0.GPL", PM_NoSelect, TRUE, PM_Center, TRUE, TAG_DONE),
				PM_Item, (ULONG) PM_MakeItem(PM_Title, (ULONG) "Version " VERSION_STRING, PM_NoSelect, TRUE, PM_Center, TRUE, TAG_DONE),
				PM_Item, (ULONG) PM_MakeItem(PM_Title, (ULONG) "MPEG-Audio Player for 680x0 Amiga computers", PM_NoSelect, TRUE, PM_Center, TRUE, TAG_DONE),
				PM_Item, (ULONG) PM_MakeItem(PM_Title, (ULONG) " Copyright 1999-2001 by Tamara C.", PM_NoSelect, TRUE, PM_Center, TRUE, TAG_DONE),
				PM_Item, (ULONG) PM_MakeItem(PM_Title, (ULONG) " Copyright 2001-2002 by Vladimir Zidar", PM_NoSelect, TRUE, PM_Center, TRUE, TAG_DONE),
				PM_Item, (ULONG) PM_MakeItem(PM_Title, (ULONG) "mailto:<tamara@MindNever.org>", PM_NoSelect, TRUE, PM_Center, TRUE, TAG_DONE),
				PM_Item, (ULONG) PM_MakeItem(PM_Title, (ULONG) "mailto:<mr_W@MindNever.org>", PM_NoSelect, TRUE, PM_Center, TRUE, TAG_DONE),
				PM_Item, (ULONG) PM_MakeItem(PM_Title, (ULONG) "http://www.MindNever.org/theprayer", PM_NoSelect, TRUE, PM_Center, TRUE, TAG_DONE),
				PM_Item, (ULONG) PM_MakeItem(PM_Title, (ULONG) "http://www.Prayer2.org", PM_NoSelect, TRUE, PM_Center, TRUE, TAG_DONE),
				PM_Item, (ULONG) PM_MakeItem(PM_Title, (ULONG) "All rights reserved.", PM_NoSelect, TRUE, PM_Center, TRUE, TAG_DONE),

				TAG_DONE)) {
			PM_OpenPopupMenu(imsg->IDCMPWindow, PM_Menu, (ULONG) p, PM_Code, SELECTDOWN, TAG_DONE);
			PM_FreePopupMenu(p);
		}
}


static void ShowPopupText(struct IntuiMessage *imsg)
{
struct PrayerButton *prbt;
char *zp, *zcopy, *zr;
struct	TagItem *list;
ULONG	i = 0, j = 0;
APTR	p;

if(prbt = (APTR)(((struct Gadget *)(imsg->IAddress))->UserData)) {
		if(zp = GetPrayerText(Robe, prbt->Argument)) {
				for(zcopy = zp; *zcopy; zcopy++) {
						if(*zcopy == '\n')
							j++;
						i++;
					}
				if(zcopy = PAlloc(i + 1)) {
						strcpy(zcopy, zp);
						if(list = PAlloc(sizeof(*list) * (j + 1))) {
								i = 0;
								zr = zcopy;
								for(zp = zcopy; *zp; zp++) {
										if(*zp == '\n') {
												*zp++ = 0;
												list[i].ti_Tag = PM_Item;
												if(zr[0] == '-' && zr[1] == '-' && zr[2] == '-')
													list[i].ti_Data = (ULONG)PM_MakeItem(PM_TitleBar, TRUE, TAG_DONE);
												else
													list[i].ti_Data = (ULONG)PM_MakeItem(PM_Title, (ULONG) zr, PM_NoSelect, TRUE, PM_Center, TRUE, TAG_DONE);
												zr = zp;
												i++;
											}
									}
								list[i].ti_Tag = PM_Item;
								list[i].ti_Data = (ULONG)PM_MakeItem(PM_Title, (ULONG) zr, PM_NoSelect, TRUE, PM_Center, TRUE, TAG_DONE);
								i++;
								list[i].ti_Tag = TAG_DONE;
								if(p = PM_MakeMenuA(list)) {
										PM_OpenPopupMenu(imsg->IDCMPWindow, PM_Menu, (ULONG) p, PM_Code, SELECTDOWN, TAG_DONE);
										PM_FreePopupMenu(p);							
									} else Printf("Can't create popup\n");
								PFree(list);
							}
						PFree(zcopy);
					}
			} else Printf("Can't find text nr %ld\n", prbt->Argument);
	}
}

static long LoadPlayList(struct IntuiMessage *imsg)
{
BPTR lock = NULL;
BPTR file;
char *zbuffer, *zp;
LONG (* pplPlayListLoadFunc)(register (* Callback)(char *path)) = (LONG (*)())SPEC_PTR;

if(imsg == NULL) {
		if(FreezeData.PlayListName[0] == 0)
			return(0);
		lock = Lock(FreezeData.PlayListPath, ACCESS_READ);
	} else {
		if(imsg != SPEC_PTR)
			pplPlayListLoadFunc = (LONG (*)()) SelectPlayListFunc(imsg);
		if(pplPlayListLoadFunc == SPEC_PTR) {
				if(rtFileRequest(qPlayListReq, FreezeData.PlayListName, "Select PlayList to load", 
															RT_LockWindow, TRUE,
															RTFI_OkText, "Load",
															RT_PubScrName, Prefs->Screen,
															TAG_DONE)) {
						lock = Lock(qPlayListReq->Dir, ACCESS_READ);
					}
			} else if(pplPlayListLoadFunc) {
				FreePlayList();

				pplPlayListLoadFunc( pplAddNodeCallback );

				SetListerList(Robe, PRT_LISTER_PLAYLIST, &PlayList);
			}
	}
if(lock) {
		lock = CurrentDir(lock);
		if(zbuffer = PAlloc(1024)) {
				if(file = Open(FreezeData.PlayListName, MODE_OLDFILE)) {
						FreePlayList();
						while(FGets(file, zbuffer, 1023)) {
								if(zp = strchr(zbuffer, '\n'))
									*zp = 0;
								AddNode(zbuffer);
							}
						SetListerList(Robe, PRT_LISTER_PLAYLIST, &PlayList);
						Close(file);
					}
				PFree(zbuffer);
			}
		UnLock(CurrentDir(lock));
	}
return(0);
}

static void SortPlayList(void)
{
struct List temp;
struct ListNode *ln, *lt;

if(IsListEmpty(&PlayList))
	return;

NewList(&temp);
SetListerList(Robe, PRT_LISTER_PLAYLIST, (APTR) ~0);

while(ln = (APTR) RemHead(&PlayList)) {
		for(lt = (APTR) temp.lh_TailPred; lt->node.ln_Pred; lt = (APTR) lt->node.ln_Pred) {
				if(Stricmp(ln->node.ln_Name, lt->node.ln_Name) >= 0) {
						Insert(&temp, (APTR) ln, (APTR) lt);
						ln = NULL;
						break;
					}
			}
		if(ln)
			AddHead(&temp, (APTR) ln);
	}

// Huh, this should copy entire list.

while(ln = (APTR) RemHead(&temp))
	AddTail(&PlayList, (APTR) ln);

SetListerList(Robe, PRT_LISTER_PLAYLIST, &PlayList);
}

static void SavePlayList(void)
{
BPTR pfile, odir;
struct Node *node;

if(rtFileRequest(qPlayListReq, FreezeData.PlayListName, "Select PlayList to Save", 
													RT_LockWindow, TRUE,
													RTFI_Flags, FREQF_SAVE,
													RTFI_OkText, "Save",
													RT_PubScrName, Prefs->Screen,
													TAG_DONE)) {
		if(odir = Lock(qPlayListReq->Dir, ACCESS_READ)) {
				odir = CurrentDir(odir);
				if(pfile = Open(FreezeData.PlayListName, MODE_NEWFILE)) {
						for(node = (APTR) PlayList.lh_Head; node->ln_Succ; node = (APTR) node->ln_Succ) {
								FPrintf(pfile, "%s\n", ((struct ListNode *)node)->storage);
							}
						Close(pfile);
					}
				UnLock(CurrentDir(odir));
			}
	}
}


static void InitPlugins(void)
{
NewList(&PShared.PluginList);
NewList(&plugin_list);
InitSemaphore(&plugin_lock);
InitSemaphore(&PShared.PluginLock);

ReloadPlugins();
}


static void FreePluginsAll(void)
{
struct PrayerPlugin *ppl;
struct MinNode *node;

while(node = (APTR)RemHead(&plugin_list)) {
		ppl = (APTR)(node + 1);
//Printf("FreePluginsAll: %s (opencount = %ld)\n", ppl->Node.ln_Name, ppl->InUse);
		UnLoadSeg(ppl->Segment);
		PFree(node);
	}
}


static void ReloadPlugins(void)
{
struct PrayerPlugin *ppl;
struct Node *node;

ObtainSemaphore(&PShared.PluginLock);
while(ppl = (APTR) RemTail(&PShared.PluginList)) {
		ClosePlugin(ppl);
	}

PShared.PluginCount = 0;

for(node = Prefs->Plugins.lh_Head; node->ln_Succ; node = node->ln_Succ) {
		if(ppl = OpenPlugin((char *)(node + 1))) { // open ...
				Enqueue(&PShared.PluginList, (APTR) ppl);
				PShared.PluginCount++;
			}
	}
ReleaseSemaphore(&PShared.PluginLock);
}



void ClosePlugin(struct PrayerPlugin *ppl)
{
struct MinNode *node;

if(ppl) {
		node = (APTR)ppl;
		node--;
		ObtainSemaphore(&plugin_lock);
		ppl->InUse--;
//Printf("ClosePlugin: %s (opencount=%ld)\n", ppl->Node.ln_Name, ppl->InUse);
		if(ppl->InUse == 0) {
				UnLoadSeg(ppl->Segment);
				Remove((APTR) node);
				PFree(node);
			}
		ReleaseSemaphore(&plugin_lock);
	}
}


struct PrayerPlugin *OpenPlugin(char *path)
{
ULONG	i;
UWORD	*data;
BPTR seg;
struct PrayerPluginTag *ppt = NULL;
struct PrayerPlugin *ppl = NULL;
struct TagItem *ti;
struct MinNode *node;
struct mySegment {
		long space;
		long length;
		BPTR nextseg;
	} *segptr;

//return(NULL); // crippled :-(

ObtainSemaphore(&plugin_lock);

for(node = (APTR) plugin_list.lh_Head; node->mln_Succ; node = node->mln_Succ) {
		ppl = (APTR) (node + 1);
		if(!Stricmp(ppl->Node.ln_Name, path)) {
				break;
			}
	}

if(node->mln_Succ == NULL) {
		ppl = NULL;
		if(seg = NewLoadSeg(path, NULL)) {
				segptr = BADDR(seg - 2);
				while(1) {
					i = segptr->length / 2;
					data = (UWORD *)(segptr + 1);
					while(i--) {
							if(*data == 0x4AFC) {
									ppt = (APTR)data;
									if((ppt->MagicLong == MAKE_ID('P','P','L','G'))
											&& (ppt->Link == ppt)) {
											break;
										} else ppt = NULL;
								}
							data++;
						}
					if(segptr->nextseg)
						segptr = BADDR(segptr->nextseg - 2);
					else
						break;
				}
				if(ppt) {
						if(node = PAlloc(sizeof(*ppl) + sizeof(*node) + strlen(path) + 1)) {
								ppl = (APTR)(node + 1);

								ppl->Node.ln_Name = (char *)(ppl + 1);
								strcpy(ppl->Node.ln_Name, path);

								ppl->Node.ln_Type = ppt->Type;
								ppl->Node.ln_Pri = ppt->Pri;
								ppl->Segment = seg;
								ppl->Enabled = TRUE;

								if(ti = FindTagItem(PPL_Name, ppt->Tags))
									ppl->Name = (char *) ti->ti_Data;
								else
									ppl->Name = FilePart(ppl->Node.ln_Name);
								if(ti = FindTagItem(PPL_Title, ppt->Tags))
									ppl->Title = (char *) ti->ti_Data;
								if(ti = FindTagItem(PPL_Author, ppt->Tags))
									ppl->Author = (char *) ti->ti_Data;
								if(ti = FindTagItem(PPL_Description, ppt->Tags))
									ppl->Description = (char *) ti->ti_Data;
								ppl->Tags = ppt->Tags;
								if(ti = FindTagItem(PPL_PlayListExtra, ppt->Tags))
									ppl->PlayListExtra = (char *) ti->ti_Data;

//Printf("ppl->Title = %s\n", ppl->Title ? ppl->Title : "N/A");
//Printf("ppl->Author = %s\n", ppl->Author ? ppl->Author : "N/A");

								AddTail(&plugin_list, (APTR) node);
							}
					}
				if(!ppl)
					UnLoadSeg(seg);
			}
	}

if(ppl)
	ppl->InUse++;

//Printf("OpenPlugin(%s) = 0x%08lx\n", path, ppl);
ReleaseSemaphore(&plugin_lock);
return(ppl);
}

static APTR SelectPlayListFunc(struct IntuiMessage *imsg)
{
struct PrayerPlugin *ppl;
struct TagItem *ti, *Tags;
APTR rc = SPEC_PTR, p;
ULONG	i = 0;

ObtainSemaphoreShared(&PShared.PluginLock);
for(ppl = (APTR) PShared.PluginList.lh_Head; ppl->Node.ln_Succ; ppl = (APTR) ppl->Node.ln_Succ) {
		if(ppl->PlayListExtra)
			i++;
	}
if(i > 0) {
		if(Tags = PAlloc((i + 2) * sizeof(*Tags))) {
				i = 0;				
				for(ppl = (APTR) PShared.PluginList.lh_Head; ppl->Node.ln_Succ; ppl = (APTR) ppl->Node.ln_Succ) {
						if(ppl->PlayListExtra) {
								Tags[i].ti_Tag = PM_Item;
								Tags[i].ti_Data = (ULONG)PM_MakeItem(PM_Title, ppl->PlayListExtra, PM_UserData, ppl, TAG_DONE);
								i++;								
							}
					}
				Tags[i].ti_Tag = TAG_DONE;

				if(p = PM_MakeMenu( PM_Item, PM_MakeItem(PM_Title, "Normal Playlist", PM_UserData, 1, TAG_DONE),
														TAG_MORE, Tags)) {
						if(ppl = (APTR) PM_OpenPopupMenu(imsg->IDCMPWindow, PM_Menu, p, PM_Code, SELECTDOWN, TAG_DONE)) {
								if((ULONG)ppl != 1) {
										if(ti = FindTagItem(PPL_PlayListLoadFunc, ppl->Tags)) {
												rc = (APTR)ti->ti_Data;
											}
									}
							} else rc = NULL;
					}
				PFree(Tags);
			}
	}
ReleaseSemaphore(&PShared.PluginLock);
return(rc);
}

static void FreePlayList()
{
struct Node *node;
SetListerList(Robe, PRT_LISTER_PLAYLIST, (struct List *)~0);
while(node = RemHead(&PlayList)) {
		PFree(node);
	}
SelectedPtr = NULL;
CurrentTrack = NULL;
NumTracks = 0;
}

APTR PAlloc(ULONG size)
{
ULONG *mem;

size += sizeof(ULONG);

if(mem = AllocPooled(FastMemoryPool, size)) {
		*mem = size;
		mem++;
	}

return(mem);
}

void PFree(APTR mem)
{
ULONG *m;

if(m = mem) {
		m--;
		FreePooled(FastMemoryPool, m, *m);
	}
}

static void PatternRemove(struct Window *win)
{
char	*zbuffer, *zparsed;
struct Node *nn, *next;

zbuffer = PAlloc(256);
zparsed = PAlloc(600);
if(zbuffer && zparsed) {
		if(rtGetString(zbuffer, 255, "Remove Pattern", NULL,
											RTGS_TextFmt, "Enter wildcard pattern to match entries",
											RT_Window, win,
											TAG_DONE)) {
				if(ParsePatternNoCase(zbuffer, zparsed, 600) >= 0) {
						nn = PlayList.lh_Head;
						while(nn->ln_Succ) {
								nn = nn->ln_Succ;
								if(MatchPatternNoCase(zparsed, nn->ln_Pred->ln_Name)) {
										next = RemNode(nn->ln_Pred);
										if(nn->ln_Pred == SelectedPtr) {
												SelectedPtr = next;
											}
									}
							}
					}
			}
	}
PFree(zparsed);
PFree(zbuffer);
}

static struct Node *GetRND(void)
{
ULONG r;
struct Node *ref = NULL;
struct DateStamp	ds;
int i, n = 0;
UBYTE cp;

if(NumTracks == 0)
	return(NULL);

DateStamp(&ds);
r = ((ds.ds_Minute)
		^ (ds.ds_Tick << 2)
		^ (SysBase->IdleCount)
		^ (SysBase->DispCount)
		^ (SysBase->ex_EClockFrequency)
		^ NumTracks) % NumTracks;

while((r > 0) && (n <= NumTracks)) {
		if(ref && ref->ln_Succ)
			ref = ref->ln_Succ;
		else
			ref = PlayList.lh_Head;

		for(i = 0; i < 256; i++) {
				if(PlayHistory[i] == ref) {
						break;
					}
			}

		if(i == 256)
			r--;

		n++;
	}

if(n > NumTracks) { // Delete NumTracks/2 from history
		i = (NumTracks / 2) % 128;
		cp = ph_Cursor - (i * 2);
		while(i--) {
				PlayHistory[cp++] = NULL;
			}
	}
return(ref);
}

static void UpdateTimers(void)
{
static ULONG osec = 0;
static UBYTE sleep_mode_volume = 0, active_flag = FALSE;
static WORD alarm_fadein = 0;
LONG diff, sec;
struct DateStamp ds;
UWORD	dayalarm = 0xffff;

alarm(1, 0);

//ScrollAllDisplays(Robe);

DateStamp(&ds);

if(FreezeData.DayAlarm_Enabled)
	dayalarm = FreezeData.DayAlarm_Minute + (FreezeData.DayAlarm_Hour * 60);

sec = (ds.ds_Tick/TICKS_PER_SECOND)
			+ (ds.ds_Minute*60L)
			+ (ds.ds_Days*60L*60L*24L);

if(diff = (sec - osec)) {
		if(FreezeData.PlayMode == PMODE_SLP) {
				SetSleepSeconds(FreezeData.SleepSeconds - diff);
				if(FreezeData.SleepSeconds < Prefs->AutoFadeOut) {
						if(sleep_mode_volume == 0)
							sleep_mode_volume = FreezeData.Volume;
						if(FreezeData.SleepSeconds <= 0) {
								Prayer_Stop();
								sleep_mode_volume = 0;
							}
						else
							SetVolume((sleep_mode_volume * FreezeData.SleepSeconds) / Prefs->AutoFadeOut);
					}
			}
		if(ds.ds_Minute == dayalarm) {
				if(!active_flag) {
				/* .. do the alarm :-) */
						if(PShared.status != PSTATUS_PRAYING) {
								SetVolume(0);
								alarm_fadein = Prefs->AutoFadeIn;
								if(FreezeData.PlayMode == PMODE_SLP) {
										FreezeData.PlayMode = PMODE_PGM;
										UpdatePrayerDisplay(Robe, PRT_DISPLAY_PLAYMODE, pmodes[FreezeData.PlayMode]);
									}
								PlayCurrent();
								active_flag = TRUE;
							}
					}
			} else active_flag = FALSE;

		if(alarm_fadein > 0) {
				SetVolume(100 / alarm_fadein);
				alarm_fadein -= diff;
			}
		osec = sec;
	}
}

static void SetSleepSeconds(WORD sec)
{
UWORD h;
char	ab[20];

if(sec < 0)
	sec = 0;

FreezeData.SleepSeconds = sec;

UpdatePrayerSlider(Robe, PRT_SLIDER_SLEEP, 10800, sec);
h = sec / 3600;
sec = sec % 3600;
SPrintf(ab, "%02ld:%02ld:%02ld", h, sec / 60, sec % 60);
UpdatePrayerDisplay(Robe, PRT_DISPLAY_SLEEP, ab);
}

static void DisplayAlarmTime(void)
{
char	ab[20];

if(FreezeData.DayAlarm_Enabled) {
		SPrintf(ab, "%02ld:%02ld:00", FreezeData.DayAlarm_Hour, FreezeData.DayAlarm_Minute);
		UpdatePrayerDisplay(Robe, PRT_DISPLAY_ALARM, ab);
	} else {
		UpdatePrayerDisplay(Robe, PRT_DISPLAY_ALARM, "--:--:--");
	}
}

static struct HippoPort *HIPPort = NULL;

static void hipemu_init(void)
{
static UBYTE	smp_r[512], smp_l[512];
static struct PTch PTch_r, PTch_l;

if(HIPPort = PAlloc(sizeof(*HIPPort))) {
		HIPPort->ExecMessage.mp_Node.ln_Type = NT_MSGPORT;
		HIPPort->ExecMessage.mp_Node.ln_Name = "HiP-Port";
		HIPPort->playertype = 33;
		HIPPort->PTch1 = HIPPort->PTch2 = &PTch_l;
		HIPPort->PTch3 = HIPPort->PTch4 = &PTch_r;
		PTch_r.length = 512;
		PTch_r.loopstart = 0;
		PTch_r.replen = 0;
		PTch_r.volume = 64;
		PTch_r.period = 123;
		CopyMem(&PTch_r, &PTch_l, sizeof(PTch_l));
		PTch_r.start = (LONG) &smp_r[0];
		PTch_l.start = (LONG) &smp_l[0];
		AddPort((APTR) HIPPort);
	}
}

static void hipemu_cleanup(void)
{
if(HIPPort) {
		while(HIPPort->opencount) {
//				rtEZRequestA("Please close all HippoPlayer scopes !", "Ok", NULL, NULL, NULL);
Delay(50);
			}	
		RemPort((APTR) HIPPort);
		PFree(HIPPort);
	}
}

static void hipemu_update(void)
{
int i;
struct WaveBuffer *wbuf;
UBYTE *ch_r, *ch_l, *dst_r, *dst_l;
ULONG offs;

if(HIPPort) {
		HIPPort->mainvolume = 64;
		if(wbuf = AudioStatus.CurrentBuffer) {
				offs = AudioStatus.SampleOffs;
				HIPPort->play = TRUE;
				ch_r = (APTR) (wbuf->Ch1 + offs);
				ch_l = (APTR) (wbuf->Ch2 + offs);
				HIPPort->PTch1->length = HIPPort->PTch3->length =wbuf->Length;
				dst_r = (APTR) HIPPort->PTch1->start;
				dst_l = (APTR) HIPPort->PTch3->start;
				for(i = 0; i < 512; i++) {
						*dst_r++ = *ch_r;
						*dst_l++ = *ch_l;
						ch_r += 2;
						ch_l += 2;
					}
				HIPPort->PTtrigger1 =
					HIPPort->PTtrigger2 =
					HIPPort->PTtrigger3 =
					HIPPort->PTtrigger4 = 1;
			} else HIPPort->play = FALSE;
	}
}


static BOOL HandleHotkey(int f)
{
BOOL rc = TRUE;
int i;

switch(f) {
		case PRHK_POPUP:
			fshow = TRUE;
			break;
		case PRHK_HIDE:
			fhide = TRUE;
			break;
		case PRHK_PLAY:
			CurrentTrack = SelectedPtr;
			Prayer_Stop();
			PlayCurrent();
			break;
		case PRHK_STOP:
			Prayer_Stop();
			break;
		case PRHK_PAUSE:
			paused ^= 1;
			if(paused)
				Prayer_Pause();
			else
				Prayer_Resume();
			UpdatePrayerButton(Robe, PRT_BUTTON_PAUSE, paused);
			if(PShared.status == PSTATUS_PRAYING)
				SetPrayerIcon(NULL);
			break;
		case PRHK_NEXT:
			PlayNext();
			break;
		case PRHK_PREV:
			PlayPrev();
			break;
		case PRHK_QUIT:
			rc = FALSE;
			break;
		case PRHK_ADDFILES:
			AddFiles();
			break;
		case PRHK_PLAYMODE:
			FreezeData.PlayMode = (FreezeData.PlayMode + 1) % PMODE__MAX;
			UpdatePrayerDisplay(Robe, PRT_DISPLAY_PLAYMODE, pmodes[FreezeData.PlayMode]);
			break;
		case PRHK_REM:
			SelectedPtr = RemNode(SelectedPtr);
			break;
		case PRHK_EXTREM:
// FIXME SOON !
			break;
		case PRHK_LOADLIST:
			LoadPlayList(SPEC_PTR);
			break;
		case PRHK_EXTLOADLIST:
// FIXME SOON !
			break;
		case PRHK_SAVELIST:
			SavePlayList();
			break;
		case PRHK_SORTLIST:
			SortPlayList();
			break;
		case PRHK_PREFS:
			EditPrayerPrefs(Prefs->Screen);
			break;
		case PRHK_ABOUT:
// FIXME SOON
			break;
		case PRHK_VOLUME_UP:
			if(FreezeData.Volume < 200)
				SetVolume(FreezeData.Volume + 1);
			break;
		case PRHK_VOLUME_DOWN:
			if(FreezeData.Volume > 0)
				SetVolume(FreezeData.Volume - 1);
			break;
		case PRHK_POSITION_UP: {
			struct WaveBuffer *wbuf = AudioStatus.CurrentBuffer;
			if(wbuf)
				Prayer_Skip(wbuf->Position + 1000);
			} break;
		case PRHK_POSITION_DOWN: {
			struct WaveBuffer *wbuf = AudioStatus.CurrentBuffer;
			if(wbuf && (wbuf->Position > 1000))
				Prayer_Skip(wbuf->Position - 1000);
			} break;
		case PRHK_LISTER_UP:
			i = GetListerTop(Robe, PRT_LISTER_PLAYLIST);
			if(i > 0)
				SetListerTop(Robe, PRT_LISTER_PLAYLIST, i - 1);
			break;
		case PRHK_LISTER_DOWN:
			SetListerTop(Robe, PRT_LISTER_PLAYLIST, GetListerTop(Robe, PRT_LISTER_PLAYLIST) + 1);
			break;
		case PRHK_SELECT_DOWN:
			SetListerActive(Robe, PRT_LISTER_PLAYLIST, GetListerActive(Robe, PRT_LISTER_PLAYLIST) + 1);
			break;
		case PRHK_SELECT_UP:
			i = GetListerActive(Robe, PRT_LISTER_PLAYLIST);
			if(i > 0)
				SetListerActive(Robe, PRT_LISTER_PLAYLIST, i - 1);
			break;
	}
return(rc);
}


static BOOL HandleRexxMsg(struct RexxMsg *msg)
{
BOOL rc = TRUE;
#if 0

int i;

Printf("HandleRexxMsg:\n");
for(i = 0; i < 5; i++) {
		Printf("Arg%ld = %s\n", i, msg->rm_Args[i] ? msg->rm_Args[i] : "(NULL)");
	}
#endif
return(rc);
}

static ULONG SendRexxMsg(char *zport, char *ctl, ...)
{
ULONG rc = 0;
va_list args;

char *zcmd;
struct RexxMsg *msg;
struct MsgPort *port, *rp;

va_start(args, ctl);

if(zcmd = PAlloc(4096)) {
		RawDoFmt(ctl, args, (void (*)()) "\x16\xC0\x4E\x75", zcmd);

		if(rp = CreateMsgPort()) {
				if(msg = CreateRexxMsg(rp, NULL, NULL)) {
						if(msg->rm_Args[0] = CreateArgstring(zcmd, strlen(zcmd))) {
								msg->rm_Action = RXCOMM | 1;
								Forbid();
								if(port = FindPort(zport))
									PutMsg(port, (APTR) msg);
								Permit();
								if(port) {
										while(!GetMsg(rp))
											WaitPort(rp);
										rc = (ULONG) msg->rm_Result1;
// Printf("SendRexxMsg: rc1 = %ld, rc2 = %ld\n", msg->rm_Result1, msg->rm_Result2);
									}
								DeleteArgstring(msg->rm_Args[0]);
							}
						DeleteRexxMsg(msg);
					}
				DeleteMsgPort(rp);
			}
		PFree(zcmd);
	}

va_end(args);
return(rc);
}

static void AddPrayerIcon(void)
{
if(AppIcon)
	return;

if(DiskIcon) {
		AppIcon = AddAppIcon(1,
											NULL,
											"Prayer2: Waiting",
											AppPort,
											NULL,
											DiskIcon,
											DAE_Close, TRUE,
											DAE_Info, TRUE,
											DAE_Menu, "Prefs",
											DAE_Menu, "Add Files",
											DAE_Menu, "Play Previous",
											DAE_Menu, "Play Next",
											DAE_Menu, "Play Current",
											DAE_Menu, "Stop",
											DAE_Menu, "Pause / Resume",
											TAG_DONE);
		SetPrayerIcon(NULL);
	}
}

static void RemovePrayerIcon(void)
{
if(AppIcon)
	RemoveAppIcon(AppIcon);

AppIcon = NULL;
}


static void mPrayerPause(void)
{
Prayer_Pause();
paused = TRUE;
UpdatePrayerButton(Robe, PRT_BUTTON_PAUSE, TRUE);
if(PShared.status == PSTATUS_PRAYING)
	SetPrayerIcon(NULL);
}

static void mPrayerResume(void)
{
Prayer_Resume();
paused = FALSE;
UpdatePrayerButton(Robe, PRT_BUTTON_PAUSE, FALSE);
if(PShared.status == PSTATUS_PRAYING)
	SetPrayerIcon(NULL);
}
