/*
 * =-=-=-=-=-=-=-=-=-=-=-  NOTICE  =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * The following is the license that applies to this copy of the Prayer2
 * software.  For a license to use the Prayer2 software under conditions
 * other than those described here, please contact MindNever Software at
 * following addresses:
 *
 * MindNever Software <info@mindnever.org>
 * MindNever Software +381 63 550-161 or +381 64 220-7115
 * 
 * =-=-=-=-=-=-=-=-=-=-  COPYRIGHT  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Prayer2 is:
 *   Copyright (c) 1999-2001 by Tamara Cetinjski <tamara@mindnever.org>
 *   Copyright (c) 2001-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *   Copyright (c) 1999-2002 by MindNever Software Group <info@mindnever.org>
 *
 * Tamara Cetinjski is
 *   Copyright (c) 1999-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *
 *
 * =-=-=-=-=-=-=-=-=-=-  LICENSE  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Redistributions in any form must be accompanied by information on
 *    how to obtain complete source code for the Prayer2 software and any
 *    accompanying software that uses the Prayer2 software.  The source code
 *    must either be included in the distribution or be available for no
 *    more than the cost of distribution plus a nominal fee, and must be
 *    freely redistributable under reasonable conditions.  For an
 *    executable file, complete source code means the source code for all
 *    modules it contains.  It does not include source code for modules or
 *    files that typically accompany the major components of the operating
 *    system on which the executable file runs.
 *
 * =-=-=-=-=-=-=-=-=-=-  NO WARRANTY =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
 *
 * THIS SOFTWARE IS PROVIDED BY MINDNEVER SOFTWARE ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
 * NON-INFRINGEMENT, ARE DISCLAIMED.  IN NO EVENT SHALL MINDNEVER SOFTWARE
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, HAIR, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include	<exec/semaphores.h>
#include	<exec/execbase.h>
#include	<exec/memory.h>

#include	<string.h>
#include	<dos/dosextens.h>
#include	<dos/dostags.h>
#include	<libraries/mpega.h>

#include	<exec/interrupts.h>
#include	<hardware/cia.h>
#include	<resources/cia.h>
#include	<proto/cia.h>

#include	<proto/exec.h>
#include	<proto/mpega.h>
#include	<proto/dos.h>
#include	<proto/asyncio.h>
#include	<proto/utility.h>


#include	"prayer2_audio.h"
#include	"prayer2_audiodriver.h"

#include "opt68k.h"

#define	HIT (*((ULONG *)0) = 0x00F00D00)


static APTR (*  pplExportOpenFunc)(register char *filename ASM("a0"),
														register ULONG frequency ASM("d0"),
														register UBYTE channels ASM("d1"));

static LONG (*  pplExportWriteFunc)(register APTR handle ASM("a0"),
														 register struct dec_buffer *buffer ASM("a1"));

static void (* pplExportCloseFunc)(register APTR handle ASM("a0"));

static APTR (*  pplOpenFunc)(register struct dec_request *req ASM("a0"),
													 register struct dec_status *dec ASM("a1"),
													 register struct dec_tags *tags ASM("a2"));

static LONG (*  pplDecodeFunc)(register APTR stream ASM("a0"), register struct dec_buffer *fb ASM("a1"));
static LONG (*  pplSeekFunc)(register APTR stream ASM("a0"), register ULONG ms_position ASM("d0"));
static LONG (*  pplTimeFunc)(register APTR stream ASM("a0"));
static void (*  pplCloseFunc)(register APTR stream ASM("a0"));
static void (*  pplBoostFunc)(register APTR stream ASM("a0"), register UWORD value ASM("d0"));

static BOOL (*  pplExamineFunc)(register UBYTE *databuffer ASM("a0"), register ULONG size ASM("d0"));


static __inline BOOL ad_AudioInit(struct AudioSettings *as);		// Begin queue-run
static __inline void ad_AudioCleanup(void);											// End everything, free hardware
static __inline void ad_AudioFlush(void);												// Stop all current requests and clear queue
static __inline void ad_AudioUpdate(struct AudioSettings *as);	// Update Volume/Freq
static __inline void ad_AudioQueue(struct WaveBuffer *wbuf);		// queue in

void ma_FreeWBuffer(struct WaveBuffer *wbuf);		// return wbuffer to free-list
void ma_DMAStart(struct WaveBuffer *wbuf);


static APTR  mpgOpenFunc(register struct dec_request *req ASM("a0"),
													 register struct dec_status *dec ASM("a1"),
													 register struct dec_tags *tags ASM("a2"));
static LONG  mpgDecodeFunc(register APTR stream ASM("a0"), register struct dec_buffer *fb ASM("a1"));
static LONG  mpgSeekFunc(register APTR stream ASM("a0"), register ULONG ms_position ASM("d0"));
static LONG  mpgTimeFunc(register APTR stream ASM("a0"));
static void  mpgCloseFunc(register APTR stream ASM("a0"));
static void  mpgBoostFunc(register APTR stream ASM("a0"), register UWORD value ASM("d0"));
static BOOL  mpgExamineFunc(register UBYTE *databuffer ASM("a0"), register ULONG size ASM("d0"));

static void pplCloseFuncMaster(APTR stream);
static APTR pplOpenFuncMaster(struct dec_request *req,
															struct dec_status *dec,
															struct dec_tags *tags);

static void pa_CiaInit(void);
static void pa_CiaCleanup(void);
static void pa_CiaPause(void);
static void pa_CiaResume(void);

static struct PrayerPlugin *CurrentPlugin = NULL;

static APTR FastPool = NULL;

struct AudioMode AudioModes[] = {
	{"Paula 14bit", paula_14bit_ad_AudioInit, paula_14bit_ad_AudioCleanup, paula_14bit_ad_AudioFlush, paula_14bit_ad_AudioQueue, paula_14bit_ad_AudioUpdate},
	{"AHI", ahi_ad_AudioInit, ahi_ad_AudioCleanup, ahi_ad_AudioFlush, ahi_ad_AudioQueue, ahi_ad_AudioUpdate},
	}, *CurrentMode = NULL;

extern struct ExecBase *SysBase;

struct Library *MPEGABase = NULL;
struct Library *AsyncIOBase = NULL;

struct	AudioStatus AudioStatus = {0};
static ULONG new_pos = 0L;

static ULONG def_baccess(register struct Hook  *hook ASM("a0"),
                                         register APTR          handle ASM("a2"),
                                         register MPEGA_ACCESS *access ASM("a1"));
void Prayer2_Audio(void);
static void P_ParseTags(char *tags, struct dec_tags *dt);
static struct WaveBuffer *GetWBuffer(void);

#define PSIGSTOP		SIGBREAKF_CTRL_C
#define	PSIGPAUSE		SIGBREAKF_CTRL_D
#define	PSIGRESUME	SIGBREAKF_CTRL_E
#define	PSIGSKIP		SIGBREAKF_CTRL_F
#define	PSIGMASK (PSIGSTOP|PSIGPAUSE|PSIGRESUME|PSIGSKIP)


static ULONG wbuffs = 0L, dbuffs = 0L;
static BYTE sig_free;
static struct List free_list, buffer_list, capture_list;
static struct AudioSettings AudioSettings = {
		22050, 100, FALSE, FALSE, 0, 0x00020002, 44100,
	};

//#define	WBSIZE			(MPEGA_PCM_SIZE * 5) / 2
static ULONG wbsize;
#define WBSIZE	wbsize

//#define	WBUFFERS		20
static ULONG wbuffers;
#define WBUFFERS wbuffers

static BYTE		sig_setup = -1;
static struct Process *prayer_task = NULL;
static struct MsgPort *prayer_port = NULL;
static struct Message prayer_msg = {0};
static char *prayer_file = NULL;
static struct Task *prayer_parent = NULL;

static struct Hook def_bsaccess_hook = {
   { NULL, NULL }, def_baccess, NULL, NULL
};


////////////
// Public //
////////////

LONG Prayer_Play(char *filename, BYTE pri)
{
LONG rc = PERROR_SYSERR;

prayer_parent = FindTask(NULL);

if(prayer_task)
	return(PERROR_AACTIVE);
else if(prayer_port)
	Prayer_Stop();

if(prayer_port = CreateMsgPort()) {
		prayer_file = filename;

		//Printf("Prayer_Play(%s) --> Port=0x%08lx\n", filename, prayer_port);
		
		if(prayer_task = CreateNewProcTags(NP_Entry, Prayer2_Audio,
																			 NP_Name, "Prayer2_Audio_Processor",
																			 NP_StackSize, 4096, NP_Priority, pri,
																			 NP_Output, Output(),
																			 NP_CloseOutput, FALSE,
																			 TAG_DONE)) {
				return(PERROR_NOERROR);
			} else {
				DeleteMsgPort(prayer_port);
				prayer_port = NULL;
			}
	}
return(rc);
}

////////////
// Public //
////////////

void Prayer_Stop(void)
{
Forbid();
if(prayer_task)
	Signal((APTR) prayer_task, PSIGSTOP);
Permit();

if(prayer_port) {
		while(!GetMsg(prayer_port))
			WaitPort(prayer_port);
		DeleteMsgPort(prayer_port);
		prayer_port = NULL;
	}
prayer_task = NULL;
}

void Prayer_Pause(void)
{
Forbid();
if(prayer_task)
	Signal((APTR) prayer_task, PSIGPAUSE);
Permit();
}

void Prayer_Resume(void)
{
Forbid();
if(prayer_task)
	Signal((APTR) prayer_task, PSIGRESUME);
Permit();
}

void Prayer_Skip(ULONG pos)
{
new_pos = pos;
Forbid();
if(prayer_task)
	Signal((APTR) prayer_task, PSIGSKIP);
Permit();
}

void Prayer_Volume(UBYTE vol)
{
PShared.volume = vol;
Forbid();
if((ULONG)prayer_task && (sig_setup > 0))
	Signal((APTR)prayer_task, 1L << sig_setup);
Permit();
}

void ma_FreeWBuffer(struct WaveBuffer *wbuf)
{
if(PShared.capture_enabled)
	AddTail(&capture_list, (APTR) wbuf);
else
	AddTail(&free_list, (APTR) wbuf);
Signal((APTR) prayer_task, 1L << sig_free);
}


//////////////////////////////////////////
/// PRIVATE
//////////////////////////////////////////

void __saveds Prayer2_Audio(void)
{
struct dec_buffer fb;
struct	WaveBuffer *wbuf;
APTR stream;
ULONG	signals, sigmask;
ULONG current_pos = 0L;
BOOL	happy = TRUE, refresh = FALSE, pr_paused = FALSE;
APTR	capture = NULL;
struct PrayerPlugin *wavedump = NULL;

sig_setup = AllocSignal(-1);
sig_free = AllocSignal(-1);

wbuffs = 0L;
dbuffs = 0L;

NewList(&buffer_list);
NewList(&free_list);
NewList(&capture_list);

memset(&AudioStatus, 0, sizeof(AudioStatus));

MPEGABase = OpenLibrary("mpega.library", 2);
AsyncIOBase = OpenLibrary("asyncio.library", 39);

if(PShared.AudioDriverID == 3)
	CurrentMode = &AudioModes[1];
else
	CurrentMode = &AudioModes[0];

if(MPEGABase && AsyncIOBase) {
		PShared.req.filename = prayer_file;
		if(stream = pplOpenFuncMaster(&PShared.req, &PShared.dec, &PShared.tags)) {
				PShared.status = PSTATUS_INIT;

				WBSIZE = ((PShared.dec.frequency / 3) & ~0x3) + 4;

				WBUFFERS = ((150 * 1024) / WBSIZE) + 1;

				if(FastPool = CreatePool(MEMF_ANY|MEMF_CLEAR|MEMF_PUBLIC, (WBSIZE + sizeof(struct WaveBuffer)) * 10, WBSIZE)) {
						AudioSettings.AhiMode = PShared.AudioID;
						AudioSettings.MixFreq = PShared.MixFreq;
						AudioSettings.BlkSiz = WBSIZE;
						AudioSettings.Stereo = (PShared.dec.channels == 2) ? TRUE : FALSE;
						AudioSettings.Freq = PShared.dec.frequency;
						AudioSettings.Volume = 0;
//						if((AudioSettings.Volume > 100) && pplBoostFunc)
//							pplBoostFunc(stream, AudioSettings.Volume);

						if(ad_AudioInit(&AudioSettings)) {

								pa_CiaInit();

								sigmask = PSIGMASK | (1L << sig_free) | (1L << sig_setup);
								Signal((APTR) prayer_task , (1L << sig_setup));

								pplExportOpenFunc = NULL;
								pplExportCloseFunc = NULL;
								pplExportWriteFunc = NULL;

								if(PShared.WaveDump && PShared.WaveDump[0] &&
										(wavedump = OpenPlugin(PShared.WaveDump))) {
										if(wavedump->Node.ln_Type == PPL_TYPE_EXPORT) {
												struct TagItem *ti;
												if(ti = FindTagItem(PPL_ExportOpenFunc, wavedump->Tags))
													pplExportOpenFunc = (APTR (* )()) ti->ti_Data;
												if(ti = FindTagItem(PPL_ExportWriteFunc, wavedump->Tags))
													pplExportWriteFunc = (LONG (* )()) ti->ti_Data;
												if(ti = FindTagItem(PPL_ExportCloseFunc, wavedump->Tags))
													pplExportCloseFunc = (void (* )()) ti->ti_Data;
											}

										if(pplExportOpenFunc && pplExportWriteFunc && pplExportCloseFunc) {
												if(PShared.capture_enabled)
													capture = pplExportOpenFunc(PShared.capture_name, PShared.dec.frequency, PShared.dec.channels);
											} else {
												ClosePlugin(wavedump);
												wavedump = NULL;
											}
									}

								if(wavedump == NULL)
									PShared.capture_enabled = FALSE;

								PShared.status = PSTATUS_PRAYING;
								Signal(prayer_parent, SIGBREAKF_CTRL_F);

								while(happy) {
										while(wbuf = (APTR) RemHead(&capture_list)) {
												if(capture) {
														fb.Size = wbuf->Length;
														fb.Data_r = wbuf->Ch1;
														fb.Data_l = wbuf->Ch2;

														pplExportWriteFunc(capture, &fb);
													}
												AddTail(&free_list, (APTR) wbuf);
											}

										if(!pr_paused && (wbuf = GetWBuffer())) {

												fb.Size = WBSIZE;
												fb.Data_r = wbuf->Ch1;
												fb.Data_l = wbuf->Ch2;

												if(pplTimeFunc)
													current_pos = pplTimeFunc(stream);

												pplDecodeFunc(stream, &fb);

												if(fb.Size < 0) {
														PShared.status = PSTATUS_SONGEND;
														break;
													}

												fb.Size &= ~3;

												wbuf->Length = fb.Size;
												wbuf->Position = current_pos;

												ad_AudioQueue(wbuf);
//ma_FreeWBuffer(wbuf);

												if(refresh) {
														ad_AudioUpdate(&AudioSettings);
														refresh = FALSE;
													}

												signals = SetSignal(0L, sigmask);

											}  else signals = Wait(sigmask);

										if(signals & PSIGPAUSE) {
												AudioSettings.Freq = 0;
												pa_CiaPause();
												pr_paused = TRUE;
												ad_AudioUpdate(&AudioSettings);
											}
										if(signals & PSIGRESUME) {
												pa_CiaResume();
												pr_paused = FALSE;
												AudioSettings.Freq = PShared.dec.frequency;
												ad_AudioUpdate(&AudioSettings);
											}
										if(signals & PSIGSKIP) {
												if(pplSeekFunc) {
														pplSeekFunc(stream, new_pos);
														ad_AudioFlush();
														refresh = TRUE;
														current_pos = new_pos;
														if(wbuf = AudioStatus.CurrentBuffer)
															wbuf->Position = current_pos;
													}
												if(pplTimeFunc)
													current_pos = pplTimeFunc(stream);
											}
										if(signals & PSIGSTOP) {
												AudioSettings.Volume = 0;
												ad_AudioUpdate(&AudioSettings);
												ad_AudioFlush();
												happy = FALSE;
											}
										if(signals & (1L << sig_setup)) {
												AudioSettings.Volume = PShared.volume;
												ad_AudioUpdate(&AudioSettings);
//												if((AudioSettings.Volume > 100) && pplBoostFunc)
//													pplBoostFunc(stream, AudioSettings.Volume);
											}

										if(PShared.capture_enabled) {
												if(wavedump) {
														if(capture == NULL)
															capture = pplExportOpenFunc(PShared.capture_name, PShared.dec.frequency, PShared.dec.channels);
													} else {
														Signal(prayer_parent, SIGBREAKF_CTRL_F);
														PShared.capture_enabled = NULL;
													}
											}
										if(!PShared.capture_enabled && capture) {
												pplExportCloseFunc(capture);
												capture = NULL;
											}
									}

								pa_CiaCleanup();

								ad_AudioCleanup();

								while(wbuf = (APTR) RemHead(&capture_list)) {
										if(capture) {
												fb.Size = wbuf->Length;
												fb.Data_r = wbuf->Ch1;
												fb.Data_l = wbuf->Ch2;
												pplExportWriteFunc(capture, &fb);
											}
										AddTail(&free_list, (APTR) wbuf);
									}

								if(wavedump) {
										if(capture)
											pplExportCloseFunc(capture);
										ClosePlugin(wavedump);
									}
							}

						memset(&AudioStatus, 0, sizeof(AudioStatus));

						DeletePool(FastPool);
					}
				pplCloseFuncMaster(stream);
			} else PShared.status = PSTATUS_ERROR_F;
	}
CloseLibrary(MPEGABase);
CloseLibrary(AsyncIOBase);

P_ParseTags("", &PShared.tags);

Forbid();
sig_setup = -1;
prayer_task = NULL;
if(PShared.status < PSTATUS_EXITED)
	PShared.status = PSTATUS_EXITED;
Signal(prayer_parent, SIGBREAKF_CTRL_F);
PutMsg(prayer_port, &prayer_msg);
}

static struct WaveBuffer *GetWBuffer(void)
{
struct WaveBuffer *wbuf;

if(wbuf = (APTR) RemHead(&free_list))
	return(wbuf);

if(wbuffs < WBUFFERS) {
		if(wbuf = AllocPooled(FastPool, sizeof(*wbuf))) {
				wbuf->Ch1 = AllocPooled(FastPool, WBSIZE * 2);
				if(AudioSettings.Stereo)
					wbuf->Ch2 = AllocPooled(FastPool, WBSIZE * 2);
				else
					wbuf->Ch2 = wbuf->Ch1;

				if(wbuf->Ch1 && wbuf->Ch2) {
						wbuffs++;
						return(wbuf);
					}
			}
	}
return(NULL);
}


static __saveds ULONG def_baccess( register struct Hook  *hook ASM("a0"),
                                         register APTR          handle ASM("a2"),
                                         register MPEGA_ACCESS *access ASM("a1"))
{
	static char tagdata[128];
	LONG	i;

   switch( access->func ) {

      case MPEGA_BSFUNC_OPEN: {
      		struct FileInfoBlock *qfib;
					struct AsyncFile *qfile;
					access->data.open.buffer_size = 32768;
					
					if(qfib = AllocMem(sizeof(*qfib), MEMF_ANY)) {
							qfib->fib_Size = 0; 
					
							if(qfile = OpenAsync(access->data.open.stream_name, MODE_READ, 65536))
								ExamineFH(qfile->af_File, qfib);

							if(access->data.open.stream_size = qfib->fib_Size) {
									SeekAsync(qfile, -sizeof(tagdata), MODE_END);
									i = ReadAsync(qfile, tagdata, sizeof(tagdata) - 1);
									if(i > 0) {
											tagdata[i] = 0;
											P_ParseTags(tagdata, hook->h_Data);
										}
									SeekAsync(qfile, 0, MODE_START);
								}
							FreeMem(qfib, sizeof(*qfib));
						}

					return((ULONG) qfile);
				}
      case MPEGA_BSFUNC_CLOSE:
         if( handle ) {
						CloseAsync((APTR)handle);
         }
         break;
      case MPEGA_BSFUNC_READ: {
					LONG bytes_read;
					if(handle)
						{
							bytes_read = ReadAsync((APTR) handle, access->data.read.buffer, access->data.read.num_bytes);
							if(bytes_read < 0)
								bytes_read = 0;
							return((ULONG)bytes_read);
						}
					else
						return(0);
				}
      case MPEGA_BSFUNC_SEEK:
         if( !handle ) return 0;
					SeekAsync((APTR) handle, access->data.seek.abs_byte_seek_pos, MODE_START);

					return(0);
   }
   return 0;
}

void strfix(char *zp)
{
int i = 29;

while((i > 0) && (*zp == ' ')) {
		*zp = 0;
		zp--;
		i--;
	}
}


static void P_ParseTags(char *tags, struct dec_tags *dt)
{
char *zt, c;

dt->Title = NULL;
dt->Artist = "";
dt->Album = "";
dt->Other = "";
dt->Year = 0;

if(zt = strstr(tags, "TAG")) {
		zt += 3;
		dt->Title = zt;
		dt->Artist = zt + 30;
		dt->Album = zt + 60;
		dt->Other = zt + 94;
		strfix(zt + 29);
		strfix(zt + 59);
		strfix(zt + 89);
		strfix(zt + 123);
		c = zt[94]; zt[94] = 0;
		StrToLong(zt + 90, (APTR) &dt->Year);
		zt[94] = c;
	}
}

static void pplCloseFuncMaster(APTR stream)
{
if(pplCloseFunc)
	pplCloseFunc(stream);
else
	Printf("pplCloseFunc missing, HOW !?\n");

if(CurrentPlugin) {
		ObtainSemaphore(&PShared.PluginLock);
		CurrentPlugin->InUse--;
		ReleaseSemaphore(&PShared.PluginLock);
		CurrentPlugin = NULL;
	}
}

static APTR set_internal(UBYTE *buffer, ULONG buffsiz, APTR req, APTR dec, APTR tags)
{
APTR rc = NULL;
if(mpgExamineFunc(buffer, buffsiz)) {
		pplOpenFunc = mpgOpenFunc;
		pplDecodeFunc = mpgDecodeFunc;
		pplSeekFunc = mpgSeekFunc;
		pplTimeFunc = mpgTimeFunc;
		pplCloseFunc = mpgCloseFunc;
		pplBoostFunc = mpgBoostFunc;
		rc = pplOpenFunc(req, dec, tags);
	}
return(rc);
}


static APTR pplOpenFuncMaster(struct dec_request *req,
															struct dec_status *dec,
															struct dec_tags *tags)
{
APTR rc = NULL;
struct PrayerPlugin *ppl;
struct TagItem *ti;
BOOL	try_internal = TRUE, p;
UBYTE	*buffer;
ULONG	buffsize = 10240;
BPTR file;

CurrentPlugin = NULL;

if(buffer = AllocVec(buffsize, MEMF_ANY)) {

		if((req->filename[0] == ':')
			&& (req->filename[1] == ':'))
			strncpy(buffer, req->filename, buffsize - 1);
		else if(file = Open(req->filename, MODE_OLDFILE)) {
				Read(file, buffer, buffsize);
//Printf("Read(%s, size=%ld)\n", req->filename, buffsize);
				Close(file);
			} else {
				dec->error = "Can't open file";
				FreeVec(buffer);
				return(NULL);
			}

		ObtainSemaphoreShared(&PShared.PluginLock);

		for(ppl = (APTR) PShared.PluginList.lh_Head; ppl->Node.ln_Succ; ppl = (APTR) ppl->Node.ln_Succ) {
				if((ppl->Node.ln_Pri < 0) && try_internal) {
						if(rc = set_internal(buffer, buffsize, req, dec, tags))
							break;
						try_internal = FALSE;

					}
				if((ppl->Node.ln_Type == PPL_TYPE_DECODER) &&
						(ppl->Enabled)) {
						pplOpenFunc = NULL;
						pplDecodeFunc = NULL;
						pplSeekFunc = NULL;
						pplTimeFunc = NULL;
						pplCloseFunc = NULL;
						pplBoostFunc = NULL;
						pplExamineFunc = NULL;

						if(ti = FindTagItem(PPL_OpenFunc, ppl->Tags))
							pplOpenFunc = (APTR (* )()) ti->ti_Data;
						if(ti = FindTagItem(PPL_DecodeFunc, ppl->Tags))
							pplDecodeFunc = (LONG (* )()) ti->ti_Data;
						if(ti = FindTagItem(PPL_SeekFunc, ppl->Tags))
							pplSeekFunc = (LONG (* )()) ti->ti_Data;
						if(ti = FindTagItem(PPL_TimeFunc, ppl->Tags))
							pplTimeFunc = (LONG (* )()) ti->ti_Data;
						if(ti = FindTagItem(PPL_CloseFunc, ppl->Tags))
							pplCloseFunc = (void (* )()) ti->ti_Data;
						if(ti = FindTagItem(PPL_BoostFunc, ppl->Tags))
							pplBoostFunc = (void (* )()) ti->ti_Data;
						if(ti = FindTagItem(PPL_ExamineFunc, ppl->Tags))
							pplExamineFunc = (BOOL (* )()) ti->ti_Data;

						p = 2;

						if(pplOpenFunc && pplDecodeFunc && pplCloseFunc) {
								if(!(pplExamineFunc && !(p = pplExamineFunc(buffer, buffsize)))) {
										if(rc = pplOpenFunc(req, dec, tags)) {
												ppl->InUse++;
												CurrentPlugin = ppl;
												break;
											}
									}
						  }
					}
			}

		ReleaseSemaphore(&PShared.PluginLock);

		if((rc == NULL) && try_internal)
			rc = set_internal(buffer, buffsize, req, dec, tags);

		FreeVec(buffer);
	}
return(rc);
}

static BOOL  mpgExamineFunc(register UBYTE *databuffer ASM("a0"), register ULONG size ASM("d0"))
{
if(MPEGA_find_sync(databuffer, size) >= 0)
	return(TRUE);
return(FALSE);
}

static APTR  mpgOpenFunc(register struct dec_request *req ASM("a0"),
													 register struct dec_status *dec ASM("a1"),
													 register struct dec_tags *tags ASM("a2"))
{
static MPEGA_CTRL mpc = {0};
MPEGA_STREAM *mps;

mpc.bs_access = &def_bsaccess_hook;

def_bsaccess_hook.h_Data = tags;

mpc.layer_1_2.force_mono = (req->channels2 == 1);
mpc.layer_1_2.mono.freq_div = 
mpc.layer_1_2.stereo.freq_div = req->freqdiv2;
mpc.layer_1_2.mono.quality = 
mpc.layer_1_2.stereo.quality = req->quality2;

mpc.layer_3.force_mono = (req->channels1 == 1);
mpc.layer_3.mono.freq_div = 
mpc.layer_3.stereo.freq_div = req->freqdiv1;
mpc.layer_3.mono.quality = 
mpc.layer_3.stereo.quality = req->quality1;

mpc.check_mpeg = 0;

//Printf("MPEGA_open(%s)\n", req->filename);

if(mps = MPEGA_open(req->filename, &mpc)) {
		dec->duration = mps->ms_duration;
		dec->frequency = mps->dec_frequency;
		dec->channels = mps->dec_channels;
		dec->quality = mps->dec_quality;
	} else {
		//Printf("MPEGA_open() failed\n");
	}
return(mps);
}

static LONG  mpgDecodeFunc(register APTR stream ASM("a0"), register struct dec_buffer *fb ASM("a1"))
{
LONG siz, rc = -1, s;
WORD *pcmb[2];

pcmb[0] = &fb->Data_r[0];
pcmb[1] = &fb->Data_l[0];

s = fb->Size - MPEGA_PCM_SIZE;

while(rc < s) {
		siz = MPEGA_decode_frame(stream, pcmb);
		if(siz >= 0) {
				pcmb[0] += siz;
				pcmb[1] += siz;
			}
		else
			break;
		rc += siz;
	}

fb->Size = rc;
return(rc);
}

static LONG  mpgSeekFunc(register APTR stream ASM("a0"), register ULONG ms_position ASM("d0"))
{
return(MPEGA_seek(stream, ms_position));
}

static LONG  mpgTimeFunc(register APTR stream ASM("a0"))
{
ULONG pos = 0;
if(MPEGA_time(stream, &pos) < 0)
	return(0);
else
	return((LONG)pos);
}

static void  mpgCloseFunc(register APTR stream ASM("a0"))
{
MPEGA_close(stream);
}

static void  mpgBoostFunc(register APTR stream ASM("a0"), register UWORD value ASM("d0"))
{
MPEGA_scale(stream, value);
}

#define STOPA_AND  CIACRAF_TODIN |CIACRAF_PBON | CIACRAF_OUTMODE | CIACRAF_SPMODE
#define STOPB_AND  CIACRBF_ALARM | CIACRBF_PBON | CIACRBF_OUTMODE
#define STARTA_OR  CIACRAF_START
#define STARTB_OR  CIACRBF_START

struct Library *CiaBase = NULL;
static UBYTE	cia_timerbit;
static struct CIA	*cia_base;
static UBYTE	*cia_cr;
static UBYTE	*cia_lo;
static UBYTE	*cia_hi;
static struct Interrupt cia_int = {0};
static UBYTE	cia_stopmask;
static UBYTE	cia_startmask;
static UBYTE	cia_interval_l, cia_interval_h;
static BOOL	cia_enabled = TRUE;
static ULONG	cia_step = 0L;
static UBYTE cia_interval_hh = 0, cia_ih = 0;

static __inline void cia_start(void)
{
if(cia_ih = cia_interval_hh) {
		*cia_lo = 0xff;
		*cia_hi = 0xff;
	} else {
		*cia_lo = cia_interval_l;
		*cia_hi = cia_interval_h;
	}

*cia_cr |= cia_startmask;
}


void ma_DMAStart(struct WaveBuffer *wbuf)
{
*cia_cr &= cia_stopmask;
AudioStatus.CurrentBuffer = wbuf;
AudioStatus.SampleOffs = 0L;

Signal(prayer_parent, SIGBREAKF_CTRL_E);

cia_start();
}

static __saveds void  pa_CiaInterrupt(register APTR dp ASM("a1"))
{
ULONG i;

if(cia_ih) {
		*cia_lo = cia_interval_l;
		*cia_hi = cia_interval_h;
		cia_ih = 0;
		*cia_cr |= cia_startmask;
	} else {

		if(PShared.RefreshRate < 5)
			PShared.RefreshRate = 5;

		i = SysBase->ex_EClockFrequency / PShared.RefreshRate;

		cia_step = PShared.dec.frequency / PShared.RefreshRate;

		cia_interval_l = i & 0xff;
		cia_interval_h = (i >> 8);
		cia_interval_hh = (i >> 16);


		if(cia_enabled) {
				AudioStatus.SampleOffs += cia_step;
				Signal(prayer_parent, SIGBREAKF_CTRL_E);
			}
		cia_start();
	}
}

static void pa_CiaResume(void)
{
cia_enabled = TRUE;
Disable();
*cia_cr &= cia_startmask;
Enable();
}

static void pa_CiaPause(void)
{
cia_enabled = FALSE;
Disable();
*cia_cr &= cia_stopmask;
Enable();
}

static void pa_CiaInit(void)
{
ULONG	i;
CiaBase = OpenResource(CIAANAME);

cia_int.is_Code = (void (*)()) pa_CiaInterrupt;
cia_int.is_Data = NULL;

cia_timerbit = 0;

cia_base = (struct CIA *)0xbfe001;

if(AddICRVector(CiaBase, CIAICRB_TA, &cia_int)) {
		if(AddICRVector(CiaBase, CIAICRB_TB, &cia_int)) {
				cia_base = (struct CIA *)0xbfd000;
				CiaBase = OpenResource(CIABNAME);
				if(AddICRVector(CiaBase, CIAICRB_TA, &cia_int)) {
						if(AddICRVector(CiaBase, CIAICRB_TB, &cia_int)) {
								Printf("warning: can't allocate CIA interrupt\n");
								return;
							} else cia_timerbit = CIAICRB_TB;
					} else cia_timerbit = CIAICRB_TA;
			} else cia_timerbit = CIAICRB_TB;
	} else cia_timerbit = CIAICRB_TA;

if(cia_timerbit == CIAICRB_TA) {
		cia_cr = &cia_base->ciacra;
		cia_lo = &cia_base->ciatalo;
		cia_hi = &cia_base->ciatahi;
		cia_stopmask = STOPA_AND;
		cia_startmask = STARTA_OR;
	} else {
		cia_cr = &cia_base->ciacrb;
		cia_lo = &cia_base->ciatblo;
		cia_hi = &cia_base->ciatbhi;
		cia_stopmask = STOPB_AND;
		cia_startmask = STARTB_OR;
	}

if(PShared.RefreshRate < 5)
	PShared.RefreshRate = 5;

i = SysBase->ex_EClockFrequency / PShared.RefreshRate;

cia_step = PShared.dec.frequency / PShared.RefreshRate;

cia_interval_l = i & 0xff;
cia_interval_h = (i >> 8);
cia_interval_hh = (i >> 16);

Disable();
*cia_cr &= cia_stopmask;
Enable();

Disable();
cia_start();
Enable();
}

static void pa_CiaCleanup(void)
{
if(CiaBase) {
		RemICRVector(CiaBase, cia_timerbit, &cia_int);
		CiaBase = NULL;
	}
}

static __inline BOOL ad_AudioInit(struct AudioSettings *as)
{
if(CurrentMode)
	return(CurrentMode->ad_AudioInit(as));
else
	return(FALSE);
}

static __inline void ad_AudioCleanup(void)
{
if(CurrentMode)
	CurrentMode->ad_AudioCleanup();
}

static __inline void ad_AudioFlush(void)
{
if(CurrentMode)
	CurrentMode->ad_AudioFlush();
}

static __inline void ad_AudioQueue(struct WaveBuffer *wbuf)
{
if(CurrentMode)
	CurrentMode->ad_AudioQueue(wbuf);
}

static __inline void ad_AudioUpdate(struct AudioSettings *as)
{
if(CurrentMode)
	CurrentMode->ad_AudioUpdate(as);
}
