/*
 * =-=-=-=-=-=-=-=-=-=-=-  NOTICE  =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * The following is the license that applies to this copy of the Prayer2
 * software.  For a license to use the Prayer2 software under conditions
 * other than those described here, please contact MindNever Software at
 * following addresses:
 *
 * MindNever Software <info@mindnever.org>
 * MindNever Software +381 63 550-161 or +381 64 220-7115
 * 
 * =-=-=-=-=-=-=-=-=-=-  COPYRIGHT  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Prayer2 is:
 *   Copyright (c) 1999-2001 by Tamara Cetinjski <tamara@mindnever.org>
 *   Copyright (c) 2001-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *   Copyright (c) 1999-2002 by MindNever Software Group <info@mindnever.org>
 *
 * Tamara Cetinjski is
 *   Copyright (c) 1999-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *
 *
 * =-=-=-=-=-=-=-=-=-=-  LICENSE  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Redistributions in any form must be accompanied by information on
 *    how to obtain complete source code for the Prayer2 software and any
 *    accompanying software that uses the Prayer2 software.  The source code
 *    must either be included in the distribution or be available for no
 *    more than the cost of distribution plus a nominal fee, and must be
 *    freely redistributable under reasonable conditions.  For an
 *    executable file, complete source code means the source code for all
 *    modules it contains.  It does not include source code for modules or
 *    files that typically accompany the major components of the operating
 *    system on which the executable file runs.
 *
 * =-=-=-=-=-=-=-=-=-=-  NO WARRANTY =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
 *
 * THIS SOFTWARE IS PROVIDED BY MINDNEVER SOFTWARE ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
 * NON-INFRINGEMENT, ARE DISCLAIMED.  IN NO EVENT SHALL MINDNEVER SOFTWARE
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, HAIR, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
#ifndef _PRAYER2_AUDIO_H_
#define _PRAYER2_AUDIO_H_

#include	<dos/dos.h>
#include	<exec/semaphores.h>


#include	"plugin.h"
#include "prayer2_audiodriver.h"


extern struct Prayer_Shared {
		struct dec_request req;
		struct dec_status dec;
		struct dec_tags tags;
		UBYTE	volume;
		UBYTE	status;
		struct	List PluginList;
		struct	SignalSemaphore	PluginLock;
		ULONG		PluginCount;
		UWORD		RefreshRate;
		char	*WaveDump; // export plugin name
		BOOL	capture_enabled;
		char	*capture_name;
		ULONG	AudioDriverID;
		ULONG MixFreq;
		ULONG AudioID;
	} PShared;


struct PrayerPlugin {
	struct Node Node; // ln_Name is Path
	                  // ln_Type is plugin type
	                  // ln_Pri is plugin priority
	BPTR	Segment;
	BOOL	InUse;      // plugin is in use
	BOOL	Enabled;   // plugin is disabled..
	char	*Title;     // common
	char	*Author;    // common
	char	*Description; // common
	struct TagItem *Tags; // other data
	char	*PlayListExtra; // system use
	char	*Name;
	};

#define	PPLSYSF_AUTOOPEN	(1L << 0)

extern struct AudioMode {
	char	*Name;
	BOOL (* ad_AudioInit)(struct AudioSettings *as);
	void (* ad_AudioCleanup)(void);
	void (* ad_AudioFlush)(void);
	void (* ad_AudioQueue)(struct WaveBuffer *wbuf);
	void (* ad_AudioUpdate)(struct AudioSettings *as);
} AudioModes[];

#define	PERROR_NOERROR	0
#define	PERROR_AACTIVE	1
#define	PERROR_SYSERR		2
#define	PERROR_NACTIVE	3

#define	PSTATUS_NOTBORN	0
#define	PSTATUS_INIT		1
#define	PSTATUS_WAITING	2
#define	PSTATUS_PRAYING	3
#define	PSTATUS_EXITED	4
#define	PSTATUS_SONGEND	5
#define PSTATUS_ERROR_S	6
#define PSTATUS_ERROR_F	7

LONG Prayer_Play(char *filename, BYTE pri);
void Prayer_Stop(void);
void Prayer_Pause(void);
void Prayer_Resume(void);
void Prayer_Skip(ULONG pos);
void Prayer_Volume(UBYTE vol);

/* alarm.c */

BOOL alarm(ULONG secs, ULONG micro);

struct PrayerPlugin *OpenPlugin(char *path);
void ClosePlugin(struct PrayerPlugin *ppl);

#endif /* _PRAYER2_AUDIO_H_ */
