/*
 * =-=-=-=-=-=-=-=-=-=-=-  NOTICE  =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * The following is the license that applies to this copy of the Prayer2
 * software.  For a license to use the Prayer2 software under conditions
 * other than those described here, please contact MindNever Software at
 * following addresses:
 *
 * MindNever Software <info@mindnever.org>
 * MindNever Software +381 63 550-161 or +381 64 220-7115
 * 
 * =-=-=-=-=-=-=-=-=-=-  COPYRIGHT  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Prayer2 is:
 *   Copyright (c) 1999-2001 by Tamara Cetinjski <tamara@mindnever.org>
 *   Copyright (c) 2001-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *   Copyright (c) 1999-2002 by MindNever Software Group <info@mindnever.org>
 *
 * Tamara Cetinjski is
 *   Copyright (c) 1999-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *
 *
 * =-=-=-=-=-=-=-=-=-=-  LICENSE  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Redistributions in any form must be accompanied by information on
 *    how to obtain complete source code for the Prayer2 software and any
 *    accompanying software that uses the Prayer2 software.  The source code
 *    must either be included in the distribution or be available for no
 *    more than the cost of distribution plus a nominal fee, and must be
 *    freely redistributable under reasonable conditions.  For an
 *    executable file, complete source code means the source code for all
 *    modules it contains.  It does not include source code for modules or
 *    files that typically accompany the major components of the operating
 *    system on which the executable file runs.
 *
 * =-=-=-=-=-=-=-=-=-=-  NO WARRANTY =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
 *
 * THIS SOFTWARE IS PROVIDED BY MINDNEVER SOFTWARE ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
 * NON-INFRINGEMENT, ARE DISCLAIMED.  IN NO EVENT SHALL MINDNEVER SOFTWARE
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, HAIR, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
#include	<exec/memory.h>
#include	<string.h>
#include	<stddef.h>
#include	<hardware/blit.h>
#include	<intuition/gadgetclass.h>
#include	<intuition/icclass.h>
#include	<zlib.h>

#include <clib/alib_protos.h>

#include <libraries/reqtools.h>

#include	<proto/exec.h>
#include	<proto/graphics.h>
#include	<proto/intuition.h>
#include	<proto/dos.h>
#include	<proto/layers.h>
#include	<proto/wb.h>
#include	<proto/reqtools.h>


#include	"PrayerListClass.h"
#include	"PrayerImageClass.h"
#include	"PrayerPropClass.h"

#include	"robe.h"
#include	"prayer2_gui.h"
#include	"prayer2_prefs.h"
#include	"pif.h"

#include	"pbar.h"

#include "sprintf.h"


#include <clib/cgxsystem_protos.h>
#include	<inline/cgxsystem.h>

extern struct Library *CGXSystemBase;

APTR PAlloc(ULONG size);
void PFree(APTR mem);

//extern UBYTE	built_in_robe[];
//extern ULONG	built_in_robe_len;

#define DoMethod(MyObject, tags...) \
   	({ULONG _tags[] = { tags }; DoMethodA((MyObject), (APTR)_tags);})


static const ULONG more_prop_tags[] = {
		GA_RelVerify, TRUE,
		GA_Immediate, TRUE,
		GA_FollowMouse, TRUE,
		PGA_Total, 1,
		PGA_Visible, 1,
		PGA_NewLook, TRUE,
		PGA_Borderless, TRUE,
		TAG_DONE,
		};

static LONG MapLVToSlider[] =
{
	PLA_Top, PGA_Top,
	PLA_Total,	PGA_Total,
	PLA_Visible,	PGA_Visible,
	TAG_DONE
};

static LONG MapSliderToLV[] =
{
	PGA_Top,	PLA_Top,
	TAG_DONE
};

struct ImageData	{
		struct	Image	*Image; // intuition Image
		APTR BitMap; // gfx BitMap
	};


struct ImageCache {
		struct ImageCache *next;
		struct ImageData *image;
		UWORD	id;
	};

union robechunk {
	UBYTE		*cptr;
	struct	RobeInfo	*info;
	struct	RobeChunk	*chunk;
	struct	RobeWindow	*window;
	struct	RobeButton	*button;
	struct	RobeSlider	*slider;
	struct	RobeLevel	*level;
	struct	RobeDisplay *display;
	struct	RobeLed	*led;
	struct	RobeFont	*font;
	struct	RobeImage	*image;
	struct	RobeLister *lister;
	struct	RobeText	*text;
	struct	RobeMenu	*menu;
	struct	RobeGroup *group;
};

static struct TextFont *InitFont(struct RobeFont *font);
static void ParseWindowElements(union robechunk chunk, struct PrayerRobe *robe, struct PrayerWindow *win);
static int SetTransRegion(struct PrayerRobe *robe, struct PrayerWindow *win, UWORD image_id);
static struct ImageData *GetImage2(struct PrayerRobe *robe, struct Screen *scr, UWORD id, BOOL mask, char *func, ULONG line);
static struct	ImageData *GetImage(struct PrayerRobe *robe, struct Screen *scr, UWORD id, BOOL remap);

#define GetImage(r,s,i,m) GetImage2(r,s,i,m,__FUNCTION__,__LINE__)

static void FreeImageCached(struct PrayerRobe *robe, struct ImageData *img);
static void ImageCacheCleanup(struct PrayerRobe *robe);

static struct Gadget *GetGadgetGroup(struct PrayerWindow *win, UWORD id, UWORD *num, struct PrayerGroup **pgrp);
static void RefreshDisplays(struct PrayerWindow *win);
static BOOL ReSnipWindow(struct PrayerWindow *win, struct PrayerRobe *robe, APTR bg);

static struct TextFont *GetFont(struct PrayerRobe *robe, UWORD id);
static void safeCloseWindow(struct Window *qwin);
static void privateDisplayLevel(struct RastPort *rp, struct PrayerLevel *level, WORD value);
static void InitTranspElements(UWORD xoff, UWORD yoff, APTR bg, struct PrayerWindow *win);
static BOOL RellocChunks(UBYTE *chunks, LONG chnksize, struct PrayerRobe *robe);
static APTR ReAlloc(APTR olddata, ULONG oldsize, ULONG newsize);
static BOOL ShowPrayerWindowsFlags(struct PrayerRobe *robe, ULONG flags);
static void ConnectBoopsi(struct PrayerWindow *win);
static void ConnectBobs(struct PrayerWindow *win);
static void DisConnectBoopsi(struct PrayerWindow *win);
static APTR	GetModel(struct PrayerWindow *win, UWORD id);
static char *InitText(struct RobeText *text);
static void SnipPositions(struct PrayerRobe *robe);
char *PStrDup(char *str);

static void FreeImage(struct ImageData *img);

#define getcolor(w,p) ObtainBestPenA(w->PubScreen->ViewPort.ColorMap, p[0] << 24, p[1] << 24, p[2] << 24, NULL)
#define	freecolor(w,c) ReleasePen(w->PubScreen->ViewPort.ColorMap,c)

#if 0
static void BM_Request(struct BitMap *bm, UWORD w, UWORD h)
{
struct Window *win;
struct IntuiMessage *msg;
UWORD lb, tb;
BOOL notdone = TRUE;

if(win = OpenWindowTags(NULL, WA_InnerWidth, w + 40, 
														WA_InnerHeight, h + 40,
														WA_Flags, WFLG_CLOSEGADGET|WFLG_DRAGBAR|WFLG_DEPTHGADGET|WFLG_ACTIVATE,
														WA_IDCMP, IDCMP_CLOSEWINDOW,
														WA_Title, "BM_Request",
														TAG_DONE)) {
		lb = win->BorderLeft;
		tb = win->BorderTop;
		SetAPen(win->RPort, 2);
		Move(win->RPort, lb + 19, tb + 19);
		Draw(win->RPort, lb + 21 + w, tb + 19);
		Draw(win->RPort, lb + 21 + w, tb + 21 + h);
		Draw(win->RPort, lb + 19, tb + 21 + h);
		Draw(win->RPort, lb + 19, tb + 19);
		
		BltBitMapRastPort(bm, 0, 0, win->RPort, lb + 20, tb + 20, w, h, 0xc0);
		
		while(notdone) {
				WaitPort(win->UserPort);
				while(msg = (APTR) GetMsg(win->UserPort)) {
						if(msg->Class == IDCMP_CLOSEWINDOW)
							notdone = FALSE;
						ReplyMsg((APTR) msg);
					}
			}
		
		CloseWindow(win);
	}

}
#endif

/////////////
// Public functions
/////////////

struct PrayerRobe *InitPrayerRobe(struct MsgPort *AppPort, struct PrayerPrefs *Prefs)
{
struct	PrayerRobe *robe = NULL;
struct	FileInfoBlock	*fib;
LONG	i;
BPTR file;
BOOL	rc = FALSE;

file = Open(Prefs->Robe, MODE_OLDFILE);
if(!file) {
		PrintFault(IoErr(), Prefs->Robe);
		file = Open("Robes/default.robe", MODE_OLDFILE);
	}
if(file) {
		if(fib = PAlloc(sizeof(*fib))) {
				if(ExamineFH(file, fib)) {
						i = sizeof(*robe) + fib->fib_Size + sizeof(ULONG);
						if(robe = PAlloc(i)) {
								robe->RemapFlag = Prefs->RemapFlag;
								robe->RemapPrecision = Prefs->RemapPrecision;
								robe->RTG = Prefs->RTG;

//								if(RellocChunks(&built_in_robe[0], built_in_robe_len, robe)) {
										i = Read(file, &robe->RobeData[0], fib->fib_Size);
										if(i != fib->fib_Size) {
												Printf("%s: Read error\n", Prefs->Robe);
											} else {
												rc = RellocChunks(&robe->RobeData[0], i, robe);
											}
//									}
							}
					} else PrintFault(IoErr(), "ExamineFH");
				PFree(fib);
			}
		Close(file);
	} else PrintFault(IoErr(), "Robes/default.robe");

if(rc) {
		rc = FALSE;
		robe->PubScreenName = PStrDup(Prefs->Screen);
		robe->AppPort = AppPort;
		if(robe->UserPort = CreateMsgPort()) {
				if(file = Open("ENVARC:Prayer2.wpos", MODE_OLDFILE)) {
						Read(file, robe->Positions, robe->NumWindows * sizeof(struct PrayerPosition));
						Close(file);
					}

//				for(i = 0; i < robe->NumWindows; i++) {
//						if(robe->Positions[i].o && robe->RWindows[i])
//							robe->RWindows[i]->Flags |= PRF_WINDOW_HIDDEN;
//					}

				robe->PrayerListClass = MakePrayerListClass();
				robe->PrayerPropClass = MakePrayerPropClass();
				robe->PrayerImageClass = MakePrayerImageClass();
				robe->ListerData = PAlloc(sizeof(*robe->ListerData) * robe->NumWindows);
				
				if(robe->PrayerListClass && robe->PrayerPropClass && robe->PrayerImageClass && robe->ListerData) {
						robe->InitOk = TRUE;

						rc = ShowPrayerWindowsFlags(robe, ~0);
					}
			}
	}
if(rc)
	return(robe);
	
FreePrayerRobe(robe);
return(NULL);
}


void FreePrayerRobe(struct PrayerRobe *robe)
{
struct	TextFont	**tf;
char	**zp;
BPTR	file;
ULONG	i;

if(robe) {
		for(i = 0; i < PRT_DISPLAY__MAX; i++)
			PFree(robe->DisplayTexts[i]);

		SnipPositions(robe);


		if(robe->InitOk) {
				if(file = Open("ENVARC:Prayer2.wpos", MODE_NEWFILE)) {
						Write(file, robe->Positions, robe->NumWindows * sizeof(struct PrayerPosition));
						Close(file);
					}
			}

		HidePrayerWindows(robe);

		PFree(robe->Windows);

		if(tf = robe->Fonts) {
				i = robe->NumFonts;
				while(i--) {
						PFree(*tf++);
					}
				PFree(robe->Fonts);
			}

		if(zp = robe->Texts) {
				i = robe->NumTexts;
				while(i--) {
						PFree(*zp++);
					}
				PFree(robe->Texts);
			}

		if(robe->Images)
			PFree(robe->Images);
		if(robe->RWindows)
			PFree(robe->RWindows);
		if(robe->Positions)
			PFree(robe->Positions);
		PFree(robe->PubScreenName);
		DeleteMsgPort(robe->UserPort);

		ImageCacheCleanup(robe);

		FreePrayerListClass(robe->PrayerListClass);
		FreePrayerPropClass(robe->PrayerPropClass);
		FreePrayerImageClass(robe->PrayerImageClass);

		PFree(robe);
	}
}

struct PrayerWindow *OpenPrayerWindow(struct PrayerRobe *robe, UWORD id)
{
struct PrayerWindow *win = NULL;
union robechunk c;
BOOL	rc = FALSE;
APTR pool;
struct ImageData *bg;

if((id < 1) || (id > robe->NumWindows)) {
		Printf("OpenPrayerWindow: WindowID out of range (%ld)\n", id);
	} else {
		id--; // make index :-)
		if(win = robe->Windows[id]) {
				WindowToFront(win->Window);
				ActivateWindow(win->Window);
				return(win);
			} else {
				if(c.window = robe->RWindows[id]) {
						if(pool = CreatePool(MEMF_CLEAR, 2048, 1024)) {
								if(win = AllocPooled(pool, sizeof(*win))) {
										win->Link = robe;
										win->PubScreen = LockPubScreen(robe->PubScreenName);
										if(win->PubScreen == NULL)
											win->PubScreen = LockPubScreen(NULL);
										if(bg = GetImage(robe, win->PubScreen, c.window->ImageID, TRUE)) {
												win->MemoryPool = pool;
												win->WindowID = id;
												win->WindowWidth = bg->Image->Width;
												win->WindowHeight = bg->Image->Height;

												ParseWindowElements(c, robe, win);
												win->GadgetList = (APTR) GetGadgetGroup(win, 0L, NULL, NULL);

												if(c.window->Flags & PRF_WINDOW_CENTERED) {
														win->WindowX = (win->PubScreen->Width - win->WindowWidth) / 2;
														win->WindowY = (win->PubScreen->Height - win->WindowHeight) / 2;
													} else {
														win->WindowX = robe->Positions[id].x;
														win->WindowY = robe->Positions[id].y;
													}

												win->Flags |= WFLG_ACTIVATE;

												if(c.window->Flags & PRF_WINDOW_BORDERLESS)
														win->Flags |= WFLG_BORDERLESS|WFLG_RMBTRAP;
													else
														win->Flags |= WFLG_CLOSEGADGET|WFLG_DRAGBAR|WFLG_DEPTHGADGET|WFLG_RMBTRAP;

												if(c.window->Flags & PRF_WINDOW_BACKDROP)
													win->Flags |= WFLG_BACKDROP;

												if(c.window->ActiveImageID)
													win->ActiveImage = GetImage(robe, win->PubScreen, c.window->ActiveImageID, TRUE);
												if(c.window->InActiveImageID)
													win->InActiveImage = GetImage(robe, win->PubScreen, c.window->InActiveImageID, TRUE);

												win->StateImageX = c.window->StateImageX;
												win->StateImageY = c.window->StateImageY;

												if((c.window->ChunkSize > 24) && !(c.window->Flags & PRF_WINDOW_BORDERLESS))
													win->Title = c.window->Title;
												else
													win->Title = NULL;

												ConnectBoopsi(win);

												if(CGXSystemBase && (c.window->ChunkSize > 22))
													SetTransRegion(robe, win, c.window->TranspMaskID);

												if(win->ButtonList)
													win->Flags |= WFLG_REPORTMOUSE;

												if(ReSnipWindow(win, robe, bg->BitMap)) {
														c.window->Flags &= ~PRF_WINDOW_HIDDEN;

														win->Drag = CreateDrag(win->Window->RPort);

														ConnectBobs(win);

														robe->Positions[id].o = TRUE;

														rc = TRUE;
													} else Printf("Can't open window (%ld,%ld,%ld,%ld)\n", win->WindowX, win->WindowY, win->WindowWidth, win->WindowHeight);

												FreeImageCached(robe, bg);
											} else Printf("Can't get image %ld\n", c.window->ImageID);
									} else DeletePool(pool);
							} else Printf("Can't create memory pool\n");
					} else Printf("No such window (%ld) ?\n", id + 1);
			}
	}
if(rc) {
		robe->Windows[id] = win;
		return(win);
	}
ClosePrayerWindow(robe, win);
return(NULL);
}

static BOOL ReSnipWindow(struct PrayerWindow *win, struct PrayerRobe *robe, APTR bg)
{
BOOL rc = FALSE;

if(win->Window = OpenWindowTags(NULL, WA_Left, win->WindowX,
																			WA_Top, win->WindowY,
																			WA_InnerWidth, win->WindowWidth,
																			WA_InnerHeight, win->WindowHeight,
																			WA_Flags, win->Flags,
																			WA_PubScreen, win->PubScreen,
																			WA_BackFill, LAYERS_NOBACKFILL,
																			WA_Title, win->Title,
																			TAG_DONE)) {
		win->Window->UserData = (APTR) win; // backlink
//		HelpControl(win->Window, HC_GADGETHELP);

		SetMouseQueue(win->Window, 2);

		win->Window->UserPort = robe->UserPort;
		ModifyIDCMP(win->Window, IDCMP_CLOSEWINDOW
														 | IDCMP_GADGETDOWN
														 | IDCMP_GADGETUP
														 | IDCMP_MOUSEMOVE
														 | IDCMP_MOUSEBUTTONS
														 | IDCMP_ACTIVEWINDOW
														 | IDCMP_CHANGEWINDOW
														 | IDCMP_INACTIVEWINDOW
														 | IDCMP_RAWKEY
														 | IDCMP_INTUITICKS
														 | IDCMP_IDCMPUPDATE);

		if(CGXSystemBase && win->TranspRegion) {
				InstallTransparentRegion(win->Window->WLayer, win->TranspRegion);
				MoveWindow(win->Window, 0, 0);
			}

		BltBitMapRastPort(bg, 0, 0, win->Window->RPort, win->Window->BorderLeft, win->Window->BorderTop, win->WindowWidth, win->WindowHeight, 0xc0);
		InitTranspElements(win->WindowWidth, win->WindowHeight, bg, win);

		if(win->GadgetList) {
				AddGList(win->Window, (struct Gadget *)win->GadgetList, 0, -1, NULL);
				RefreshGadgets((struct Gadget *) win->GadgetList, win->Window, NULL);
			}

		if(win->currentGroupList) {
				AddGList(win->Window, (struct Gadget *)win->currentGroupList, 0, -1, NULL);
				RefreshGadgets((struct Gadget *) win->currentGroupList, win->Window, NULL);
			}

		RefreshDisplays(win);

		if(robe->AppPort)
			win->App = AddAppWindowA(win->WindowID, (ULONG)win, win->Window, robe->AppPort, NULL);

		rc = TRUE;
	}
return(rc);
}


void ClosePrayerWindow(struct PrayerRobe *robe, struct PrayerWindow *win)
{
struct	PrayerButton	*pbutton; //images
struct	PrayerSlider	*pslider; //images
struct	PrayerLevel		*plevel;  //colors
struct	PrayerDisplay *pdisplay; //colors
struct	PrayerLed			*pled;  //colors
struct	PrayerLister	*plister; //colors & boopsi
struct	PrayerModel		*pmodel, *pmodel2;
struct	PrayerGroup		*pgroup;


if(robe && win) {
		DisConnectBoopsi(win);

		if(win->Window) {
				robe->Positions[win->WindowID].x = win->Window->LeftEdge;
				robe->Positions[win->WindowID].y = win->Window->TopEdge;
				if(win->App)
					RemoveAppWindow(win->App);

				FreeImageCached(robe, win->ActiveImage);
				FreeImageCached(robe, win->InActiveImage);

				if(CGXSystemBase)
					InstallTransparentRegion(win->Window->WLayer, NULL);

				safeCloseWindow(win->Window);

				robe->RWindows[win->WindowID]->Flags |= PRF_WINDOW_HIDDEN;

				if(win->TranspRegion)
					DisposeRegion(win->TranspRegion);
			}

		for(pbutton = win->ButtonList; pbutton; pbutton = pbutton->Next) {
				FreeImageCached(robe, pbutton->ImageOff);
				FreeImageCached(robe, pbutton->ImageOn);
				FreeImageCached(robe, pbutton->ImageAlt);
			}
		for(pslider = win->SliderList; pslider; pslider = pslider->Next) {
				if(pslider->Gadget)
					DisposeObject(pslider->Gadget);
				FreeBitMap(pslider->Background);
				FreeImageCached(robe, pslider->ImageOff);
				FreeImageCached(robe, pslider->ImageOn);
			}
		for(plevel = win->LevelList; plevel; plevel = plevel->Next) {
				freecolor(win, plevel->ColorOff);
				freecolor(win, plevel->ColorOnLow);
				freecolor(win, plevel->ColorOnHigh);
			}
		for(pdisplay = win->DisplayList; pdisplay; pdisplay = pdisplay->Next) {
				freecolor(win, pdisplay->BgColor);
				freecolor(win, pdisplay->FgColor);
				FreeBitMap(pdisplay->Background);
				if(pdisplay->Region)
					DisposeRegion(pdisplay->Region);
			}
		for(pled = win->LedList; pled; pled = pled->Next) {
				freecolor(win, pled->ColorOn);
				freecolor(win, pled->ColorOff);
			}
		for(pgroup = win->GroupList; pgroup; pgroup = pgroup->Next) {
				freecolor(win, pgroup->Color);
				FreeImageCached(robe, pgroup->Image);
			}
		for(plister = win->ListerList; plister; plister = plister->Next) {
				freecolor(win, plister->FgPen);
				freecolor(win, plister->BgPen);
				freecolor(win, plister->FgPenS);
				freecolor(win, plister->BgPenS);
				freecolor(win, plister->FgPenH);
				freecolor(win, plister->BgPenH);

				DeleteBob(plister->Bob);

				if(plister->Gadget) {
						GetAttr(PLA_Top, plister->Gadget, &robe->ListerData[win->WindowID].Top);
						GetAttr(PLA_Selected, plister->Gadget, &robe->ListerData[win->WindowID].Selected);
						DisposeObject(plister->Gadget);
					}
			}
		pmodel = win->ModelList;
		while(pmodel) {
				DisposeObject(pmodel->Model);
				pmodel2 = pmodel;
				pmodel = pmodel->Next;
				PFree(pmodel2);
			}

		DeleteDrag(win->Drag);

		robe->Positions[win->WindowID].o = FALSE;
		robe->Windows[win->WindowID] = NULL;
		UnlockPubScreen(NULL, win->PubScreen);
		DeletePool(win->MemoryPool);
	}
}

void UpdatePrayerButton(struct PrayerRobe *robe, UWORD type, BOOL state)
{
ULONG	i, pos;
struct	PrayerWindow	*win;
struct	PrayerButton	*pbutton;
BOOL	flg;

if(robe) {
		for(i = 0; i < robe->NumWindows; i++) {
				if(win = robe->Windows[i]) {
						for(pbutton = win->ButtonList; pbutton; pbutton = pbutton->Next) {
								if((pbutton->Function == type) && ((pbutton->GroupID == 0) || (pbutton->GroupID == win->currentGroupID))) {
										if(pbutton->Gadget.Flags & GFLG_SELECTED)
											flg = TRUE;
										else
											flg = FALSE;
										if(state != flg) {
												pos = RemoveGadget(win->Window, (APTR) &pbutton->Gadget);
												if(state)
													pbutton->Gadget.Flags |= GFLG_SELECTED;
												else
													pbutton->Gadget.Flags &= ~GFLG_SELECTED;
												AddGadget(win->Window, (APTR) &pbutton->Gadget, pos);
												RefreshGList((APTR) &pbutton->Gadget, win->Window, NULL, 1);
											}
									}
							}
					}
			}
	}
}


void UpdatePrayerLevel(struct	PrayerRobe *robe, UWORD lev_r, UWORD lev_l)
{
ULONG	i;
struct	PrayerWindow	*win;
struct	PrayerLevel	*plevel;

if(robe) {
		for(i = 0; i < robe->NumWindows; i++) {
				if(win = robe->Windows[i]) {
						for(plevel = win->LevelList; plevel; plevel = plevel->Next) {
								if((plevel->GroupID == 0) || (plevel->GroupID == win->currentGroupID)) {
										switch(plevel->Type) {
												case	PRT_LEVEL_LEFT:
													privateDisplayLevel(win->Window->RPort, plevel, lev_l);
													break;
												case	PRT_LEVEL_RIGHT:
													privateDisplayLevel(win->Window->RPort, plevel, lev_r);
													break;
											}
									}
							}
					}
			}
	}
}

void UpdatePrayerSlider(struct	PrayerRobe *robe, UWORD type, UWORD total, UWORD top)
{
ULONG	i, t;
struct	PrayerWindow	*win;
struct PrayerSlider *pslider;

if(robe) {
		for(i = 0; i < robe->NumWindows; i++) {
				if(win = robe->Windows[i]) {
						for(pslider = win->SliderList; pslider; pslider = pslider->Next) {
								if((pslider->Function == type) &&
									!(pslider->Gadget->Flags & GFLG_SELECTED)) {
										if(pslider->Flags & PRF_SLIDER_VERTICAL)
											t = (total - top - 1);
										else
											t = top;
										if((pslider->GroupID == 0) || (pslider->GroupID == win->currentGroupID))
											SetGadgetAttrs(pslider->Gadget, win->Window, NULL, PGA_Top, t, PGA_Total, total, TAG_DONE);
										else
											SetAttrs(pslider->Gadget, PGA_Top, t, PGA_Total, total, TAG_DONE);
										pslider->Total = total;
									}
							}
					}
			}
	}
}
static void DrawDisplayText(struct RastPort *rp, struct PrayerDisplay *pdisplay, char *text, ULONG j)
{
ULONG w, x, i;
// first erase old field ..
struct	Region *or;

w = pdisplay->Rect.MaxX - pdisplay->Rect.MinX;
if(pdisplay->Background) {
		BltBitMapRastPort(pdisplay->Background,
													0, 0, rp,
													pdisplay->Rect.MinX,
													pdisplay->Rect.MinY,
													w,
													pdisplay->Rect.MaxY - pdisplay->Rect.MinY,
													0xc0);
	} else {
		SetABPenDrMd(rp, pdisplay->BgColor, pdisplay->FgColor, JAM2);
		RectFill(rp, pdisplay->Rect.MinX,
								 pdisplay->Rect.MinY,
								 pdisplay->Rect.MaxX,
								 pdisplay->Rect.MaxY);
	}
SetABPenDrMd(rp, pdisplay->FgColor, pdisplay->BgColor, JAM1);
SetFont(rp, pdisplay->Font);

if(pdisplay->Flags & PRF_DISPLAY_CENTERED) {
		x = ((w - TextLength(rp, text, j)) / 2);
	} else if(pdisplay->Flags & PRF_DISPLAY_SCROLLING) {
		i = (w - TextLength(rp, text, j));
		
		x = pdisplay->CurrentX;
		if((pdisplay->StepX == 0) && (i < 0))
			pdisplay->StepX = -1;
		if(pdisplay->StepX) {
				pdisplay->CurrentX += pdisplay->StepX;
				if(pdisplay->CurrentX < i)
					pdisplay->StepX = 1;
				else if(pdisplay->CurrentX == 0)
					pdisplay->StepX = -1;
			}
	} else {
		x = 0;
	}
Move(rp, pdisplay->Rect.MinX + x, pdisplay->Rect.MinY);
or = InstallClipRegion(rp->Layer, pdisplay->Region);
Text(rp, text, j);
InstallClipRegion(rp->Layer, or);
}

void UpdatePrayerDisplay(struct PrayerRobe *robe, UWORD type, char *text)
{
LONG	i, j;
struct	PrayerWindow	*win;
struct	PrayerDisplay	*pdisplay;

if(robe && text) {
		j = strlen(text);
		if(type < PRT_DISPLAY__MAX) {
				if(robe->DisplayTexts[type] != text) {
						PFree(robe->DisplayTexts[type]);
						if(robe->DisplayTexts[type] = PAlloc(j + 1))
							strcpy(robe->DisplayTexts[type], text);
					}
			}
		
		for(i = 0; i < robe->NumWindows; i++) {
				if(win = robe->Windows[i]) {
						for(pdisplay = win->DisplayList; pdisplay; pdisplay = pdisplay->Next) {
								if((pdisplay->Type == type) && ((pdisplay->GroupID == 0) || (pdisplay->GroupID == win->currentGroupID)))
									DrawDisplayText(win->Window->RPort, pdisplay, text, j);
							}
					}
			}
	}
}



void HighLightPrayerButton(struct PrayerWindow *win, struct PrayerButton *pbutton)
{
if((win == NULL))
	return;

if(pbutton && pbutton->GroupID && (pbutton->GroupID != win->currentGroupID))
	return;

if(win->LastHelp && (win->LastHelp != pbutton)) {
		if(win->LastHelp->ImageOff)
			win->LastHelp->Gadget.GadgetRender = win->LastHelp->ImageOff->Image;
		else
			win->LastHelp->Gadget.GadgetRender = NULL;

		RefreshGList((APTR)&win->LastHelp->Gadget, win->Window, NULL, 1);
		win->LastHelp = NULL;
	}

if(pbutton && (pbutton != win->LastHelp) && pbutton->ImageAlt) {
		pbutton->Gadget.GadgetRender = pbutton->ImageAlt->Image;
		RefreshGList((APTR)&pbutton->Gadget, win->Window, NULL, 1);
		win->LastHelp = pbutton;
	}
}

struct Gadget *GetActiveGadget(struct PrayerWindow *win)
{
struct PrayerSlider *pslider;

if(win) {
		for(pslider = win->SliderList; pslider; pslider = pslider->Next) {
				if(pslider->Gadget->Flags & GFLG_SELECTED)
					return((APTR) pslider->Gadget);
			}
	}
return(NULL);
}

struct Gadget *GetHitGadget(struct IntuiMessage *imsg)
{
struct PrayerWindow *win;
struct PrayerButton	*pbutton;
struct Gadget *g;
WORD	x, y;

if(win = (APTR)imsg->IDCMPWindow->UserData) {
		x = imsg->MouseX;
		y = imsg->MouseY;
		for(pbutton = win->ButtonList; pbutton; pbutton = pbutton->Next) {
				if((pbutton->GroupID == 0) || (pbutton->GroupID == win->currentGroupID)) {
						g = (struct Gadget *)&pbutton->Gadget;
						if(((x >= g->LeftEdge) && (x <= (g->LeftEdge + g->Width))) 
								&& (y >= g->TopEdge) && (y <= (g->TopEdge + g->Height)))
							return(g);
					}
			}
	}
return(NULL);
}


void HidePrayerWindows(struct PrayerRobe *robe)
{
struct PrayerWindow **pwin;
LONG	i, j = 0;

if(pwin = robe->Windows) {
		i = robe->NumWindows;
		while(i--) {
				if(*pwin) {
						ClosePrayerWindow(robe, *pwin);
						robe->Positions[j].o = TRUE;
					}	else {
						if(robe->RWindows[j])
							robe->RWindows[j]->Flags &= ~PRF_WINDOW_HIDDEN;
			 		}
				pwin++;
				j++;
			}
	}
}

BOOL ShowPrayerWindows(struct PrayerRobe *robe)
{
return(ShowPrayerWindowsFlags(robe, PRF_WINDOW_HIDDEN));
}


BOOL SetPrayerScreen(struct PrayerRobe *robe, char *newscreen)
{
char *oldscreen;
BOOL rc;

oldscreen = robe->PubScreenName;
HidePrayerWindows(robe);
robe->PubScreenName = PStrDup(newscreen);
if(rc = ShowPrayerWindows(robe))
	PFree(oldscreen);
else {
		PFree(robe->PubScreenName);
		robe->PubScreenName = oldscreen;
		ShowPrayerWindows(robe);
	}
return(rc);
}

void SwitchGroup(struct PrayerWindow *win, UWORD GroupID)
{
struct Gadget *gl;
struct PrayerGroup *pgrp;
UWORD cnt;

if((GroupID == win->currentGroupID) || (GroupID == 0) || (win == NULL))
	return;

if(gl = GetGadgetGroup(win, GroupID, &cnt, &pgrp)) {
		if(win->currentGroupList)
			RemoveGList(win->Window, win->currentGroupList, win->currentGroupCount);

		if(pgrp->Image) {
				BltBitMapRastPort(pgrp->Image->BitMap, 0, 0, win->Window->RPort,
												pgrp->Rect.MinX,
												pgrp->Rect.MinY,
												pgrp->Image->Image->Width,
												pgrp->Image->Image->Height,
												0xc0);
			} else {
				SetAPen(win->Window->RPort, pgrp->Color);
				RectFill(win->Window->RPort, pgrp->Rect.MinX, pgrp->Rect.MinY, pgrp->Rect.MaxX, pgrp->Rect.MaxY);
			}

		win->currentGroupList = gl;
		win->currentGroupCount = cnt;
		win->currentGroupID = GroupID;
		AddGList(win->Window, (struct Gadget *)gl, 0, -1, NULL);
		RefreshGadgets((struct Gadget *) gl, win->Window, NULL);
		RefreshDisplays(win);
	}
}

static struct Gadget *GetGadgetGroup(struct PrayerWindow *win, UWORD id, UWORD *num, struct PrayerGroup **xpgrp)
{
struct Gadget *glist = NULL, *g = (APTR) &glist;
struct PrayerButton *pbutton;
struct PrayerLister *plister;
struct PrayerSlider *pslider;
struct PrayerGroup *pgrp;
UWORD cnt = 0;

for(pslider = win->SliderList; pslider; pslider = pslider->Next) {
		if((pslider->GroupID == id) && (pslider->Gadget)) {
				g->NextGadget = pslider->Gadget;
				g = g->NextGadget;
			}
	}

for(plister = win->ListerList; plister; plister = plister->Next) {
		if((plister->GroupID == id) && (plister->Gadget)) {
				g->NextGadget = plister->Gadget;
				g = g->NextGadget;
			}
	}

for(pbutton = win->ButtonList; pbutton; pbutton = pbutton->Next) {
		if(pbutton->GroupID == id) {
				g->NextGadget = (APTR) &pbutton->Gadget;
				g = g->NextGadget;
			}
	}

g->NextGadget = NULL;

if(num) {
		for(g = glist; g; g = g->NextGadget) {cnt++;}
		*num = cnt;
	}

if(xpgrp) {
		*xpgrp = NULL;
		for(pgrp = win->GroupList; pgrp; pgrp = pgrp->Next) {
				if(pgrp->GroupID == id) {
						*xpgrp = pgrp;
						break;
					}
			}
	}

return(glist);
}

void SetListerAttr(struct PrayerRobe *robe, UWORD type, ULONG attr, ULONG val)
{
ULONG	i;
struct PrayerLister *plister;
struct	PrayerWindow	*win;
ULONG more_tags[] = { PLA_Top, 0, PLA_Selected, 0, TAG_DONE };
ULONG tag_done = TAG_DONE;

//if(attr == PLA_Items)
//	tag_done = TAG_MORE;

if(robe) {
		for(i = 0; i < robe->NumWindows; i++) {
				if(win = robe->Windows[i]) {
						more_tags[1] = robe->ListerData[i].Top; // PLA_Top
						more_tags[3] = robe->ListerData[i].Selected; // PLA_Selected
						for(plister = win->ListerList; plister; plister = plister->Next) {
								if(plister->Function == type) {
										if((plister->GroupID == 0) || (plister->GroupID == win->currentGroupID))
											SetGadgetAttrs(plister->Gadget, win->Window, NULL, attr, val, tag_done, &more_tags[0]);
										else
											SetAttrs(plister->Gadget, attr, val, tag_done, &more_tags[0]);
									}
							}
					}
			}
	}
}


ULONG GetListerAttr(struct PrayerRobe *robe, UWORD type, ULONG attr)
{
ULONG	i, rc = 0;
struct PrayerLister *plister;
struct	PrayerWindow	*win;

if(robe) {
		for(i = 0; i < robe->NumWindows; i++) {
				if(win = robe->Windows[i]) {
						for(plister = win->ListerList; plister; plister = plister->Next) {
								if(plister->Function == type) {
										if((plister->GroupID == 0) || (plister->GroupID == win->currentGroupID)) {
												GetAttr(attr, plister->Gadget, &rc);
												return(rc);
											}
									}
							}
					}
			}
	}
return(rc);
}

// global settings for listers
void SetListerList(struct PrayerRobe *robe, UWORD type, struct List *list)
{
SetListerAttr(robe, type, PLA_Items, (ULONG) list);
}

void SetListerActive(struct PrayerRobe *robe, UWORD type, ULONG val)
{
SetListerAttr(robe, type, PLA_MakeVisible, val);
SetListerAttr(robe, type, PLA_Selected, val);
}

void SetListerHighlight(struct PrayerRobe *robe, UWORD type, ULONG val)
{
SetListerAttr(robe, type, PLA_Highlighted, val);
}

void SetListerTop(struct PrayerRobe *robe, UWORD type, ULONG val)
{
SetListerAttr(robe, type, PLA_Top, val);
}

ULONG GetListerActive(struct PrayerRobe *robe, UWORD type)
{
return(GetListerAttr(robe, type, PLA_Selected));
}

ULONG GetListerTop(struct PrayerRobe *robe, UWORD type)
{
return(GetListerAttr(robe, type, PLA_Top));
}


UWORD GetPrayerSlider(struct Gadget *g)
{
struct PrayerSlider *pslider = (struct PrayerSlider *)(g->UserData);
ULONG rc = 0;

GetAttr(PGA_Top, g, &rc);
if(pslider->Flags & PRF_SLIDER_VERTICAL)
	rc = (pslider->Total - rc) - 1;
return((UWORD)rc);
}

char *GetPrayerText(struct PrayerRobe *robe, UWORD id)
{
if(robe && robe->Texts) {
		if((id > 0) && (id <= robe->NumTexts)) {
				return(robe->Texts[id - 1]);
			}
	}
return(NULL);
}

void SetWindowState(struct PrayerWindow *win)
{
struct ImageData *img;
struct Window *w;

if(win) {
		w = win->Window;
		if(w->Flags & WFLG_WINDOWACTIVE) {
				if(win->ListerList)
					ActivateGadget(win->ListerList->Gadget, w, NULL);
				img = win->ActiveImage;
			} else {
				img = win->InActiveImage;
			}
		if(img)
			BltBitMapRastPort(img->BitMap, 0, 0, w->RPort,
												win->StateImageX + w->BorderLeft,
												win->StateImageY + w->BorderTop,
												img->Image->Width,
												img->Image->Height,
												0xc0);
	}
}


void ScrollAllDisplays(struct PrayerRobe *robe)
{
struct PrayerWindow *win;
struct PrayerDisplay *pdisplay;
int i;
char *text;

for(i = 0; i < robe->NumWindows; i++) {
		if(win = robe->Windows[i]) {
				for(pdisplay = win->DisplayList; pdisplay; pdisplay = pdisplay->Next) {
						if((pdisplay->Flags == PRF_DISPLAY_SCROLLING)
								&& ((pdisplay->GroupID == 0) || (pdisplay->GroupID == win->currentGroupID))
								&& (pdisplay->Type < PRT_DISPLAY__MAX) && (text = robe->DisplayTexts[pdisplay->Type])) {
								DrawDisplayText(win->Window->RPort, pdisplay, text, strlen(text));
							}
					}
			}
	}
}


//////////////
// Private Functions
//////////////

static void RefreshDisplays(struct PrayerWindow *win)
{
struct PrayerDisplay *pdisplay;
struct PrayerRobe *robe = win->Link;
char *text;

for(pdisplay = win->DisplayList; pdisplay; pdisplay = pdisplay->Next) {
		if((pdisplay->Type < PRT_DISPLAY__MAX) && (text = robe->DisplayTexts[pdisplay->Type])) {
				if((pdisplay->GroupID == 0) || (pdisplay->GroupID == win->currentGroupID))
					DrawDisplayText(win->Window->RPort, pdisplay, text, strlen(text));
			}
	}
}


static void SnipPositions(struct PrayerRobe *robe)
{
ULONG i;
for(i = 0; i < robe->NumWindows; i++) {
		if(robe->Windows[i]) {
				robe->Positions[i].x = robe->Windows[i]->Window->LeftEdge;
				robe->Positions[i].y = robe->Windows[i]->Window->TopEdge;
			}
	}
}


static struct TextFont *InitFont(struct RobeFont *font)
{
struct TextFont *tf = NULL;
UBYTE	*rawdata;
ULONG	d, e, n;

if(tf = PAlloc(sizeof(*tf) + font->RawLen)) {
		rawdata = (APTR) (tf + 1);
		d = font->RawLen;
		if(e = uncompress(rawdata, &d, font->CompressedData, font->ChunkSize - sizeof(*font))) {
				Printf("Z uncompress error: %ld\n", e);
				PFree(tf);
				tf = NULL;
			} else {
				tf->tf_YSize = font->YSize;
				tf->tf_Style = font->Style;
				tf->tf_Flags = font->Flags;
				tf->tf_XSize = font->XSize;
//				tf->tf_Baseline = font->Baseline;
//	tf->tf_Baseline = font->YSize;
				tf->tf_BoldSmear = font->BoldSmear;
				tf->tf_Modulo = font->Modulo;
				tf->tf_LoChar = font->LoChar;
				tf->tf_HiChar = font->HiChar;
				tf->tf_CharData = rawdata;

				n = (font->HiChar - font->LoChar) + 2;

//Printf("InitFont: ID %ld, nchars = %ld, datasize = %ld, spaceoffset = %ld, kernoffser = %ld\n",
//				font->FontID, n, (font->Modulo * font->YSize), (n * sizeof(ULONG)), (n * sizeof(ULONG)) + (n * sizeof(UWORD)));

				rawdata += (font->Modulo * font->YSize);
				if((ULONG)rawdata & 1) {
						rawdata++;
					}
				tf->tf_CharLoc = rawdata;
				rawdata += (n * sizeof(ULONG));

				if(font->HaveCharSpace) {
						tf->tf_CharSpace = rawdata;
						rawdata += (n * sizeof(UWORD));
					}
				if(font->HaveCharKern) {
						tf->tf_CharKern = rawdata;
					}
			}
	}
return(tf);
}

static void ParseWindowElements(union robechunk c, struct PrayerRobe *robe, struct PrayerWindow *win)
{
BOOL notdone = TRUE, rc;
UWORD	xoff = 0, yoff = 0, typ, x, y, xroff = 0, yboff = 0;
ULONG	i;
struct	Rectangle	*rect;
struct	PrayerButton	*pbutton = (APTR)&win->ButtonList;
struct	PrayerSlider	*pslider = (APTR)&win->SliderList;
struct	PrayerLevel		*plevel = (APTR)&win->LevelList;
struct	PrayerDisplay *pdisplay = (APTR)&win->DisplayList;
struct	PrayerLed			*pled = (APTR)&win->LedList;
struct	PrayerGroup 	*pgroup = (APTR)&win->GroupList;
struct	PrayerLister	*plister = (APTR)&win->ListerList;
APTR	tmp;


if(!(c.window->Flags & PRF_WINDOW_BORDERLESS)) {
		xoff = win->PubScreen->WBorLeft;
		yoff = win->PubScreen->WBorTop + win->PubScreen->Font->ta_YSize + 1;
		xroff = win->PubScreen->WBorRight;
		yboff = win->PubScreen->WBorBottom;
	}

c.cptr += c.window->ChunkSize;

while(notdone && c.chunk->ChunkType) {
//Printf("Parsing window elements, got chunk %4.4s\n", &c.chunk->ChunkType);
		switch(c.chunk->ChunkType) {
				case	ID_PRAYER_WINDOW:
					notdone = FALSE;
					break;

				case	ID_PRAYER_LISTER:
					if(plister->Next = AllocPooled(win->MemoryPool, sizeof(*plister))) {
							rc = FALSE;
							tmp = plister;
							plister = plister->Next;
							/* .. init draw info .. */
							plister->BgPen = getcolor(win, c.lister->BgColor);
							plister->FgPen = getcolor(win, c.lister->FgColor);
							plister->BgPenS = getcolor(win, c.lister->BgSelect);
							plister->FgPenS = getcolor(win, c.lister->FgSelect);
							plister->BgPenH = getcolor(win, c.lister->BgActive);
							plister->FgPenH = getcolor(win, c.lister->FgActive);

							plister->Argument = c.lister->Argument;
							plister->Function = c.lister->Function;
							if(plister->Function >= PRT_LISTER__MAX)
								plister->Function = PRT_LISTER__MAX - 1;

							if(c.chunk->ChunkSize >= (offsetof(struct RobeLister, GroupID) + sizeof(UWORD)))
								plister->GroupID = c.lister->GroupID;

							if(plister->Font = GetFont(robe, c.lister->FontID)) {
									if(plister->Gadget = NewObject(robe->PrayerListClass, NULL,
											GA_ID, (c.lister->Function | 0x0200),
											GA_UserData, plister,
											GA_Left, c.lister->Left + xoff,
											GA_Top, c.lister->Top + yoff,
											GA_Width, c.lister->Width ,
											GA_Height, c.lister->Height,
											GA_RelVerify, TRUE,
											PLA_FgPen, plister->FgPen,
											PLA_BgPen, plister->BgPen,
											PLA_FgPenS, plister->FgPenS,
											PLA_BgPenS, plister->BgPenS,
											PLA_FgPenH, plister->FgPenH,
											PLA_BgPenH, plister->BgPenH,
											PLA_TextFont, plister->Font,
											TAG_DONE)) {
											rc = TRUE;
										} else Printf("NewObject(): Can't create PrayerListClass object\n");
								}
							if(!rc) {
									freecolor(win, plister->FgPen);
									freecolor(win, plister->BgPen);
									freecolor(win, plister->FgPenS);
									freecolor(win, plister->BgPenS);
									freecolor(win, plister->FgPenH);
									freecolor(win, plister->BgPenH);

									FreePooled(plister, win->MemoryPool, sizeof(*plister));
									plister = tmp;
									plister->Next = NULL;
								}
						}
					break;

				case	ID_PRAYER_BUTTON:
					if(pbutton->Next = AllocPooled(win->MemoryPool, sizeof(*pbutton))) {
							pbutton = pbutton->Next;
							if(pbutton->ImageOff = GetImage(robe, win->PubScreen, c.button->ImageOffID, TRUE)) {
									pbutton->Gadget.Flags |= GFLG_GADGIMAGE;
									pbutton->Gadget.Width = pbutton->ImageOff->Image->Width;
									pbutton->Gadget.Height = pbutton->ImageOff->Image->Height;
									pbutton->Gadget.GadgetRender = pbutton->ImageOff->Image;
								} else {
									if(c.button->Width > 1000)
										pbutton->Gadget.Width = win->WindowWidth - c.button->Left;
									else
										pbutton->Gadget.Width = c.button->Width;

									if(c.button->Height > 1000)
										pbutton->Gadget.Height = win->WindowHeight - c.button->Top;
									else
										pbutton->Gadget.Height = c.button->Height;
								}
							if(pbutton->ImageOn = GetImage(robe, win->PubScreen, c.button->ImageOnID, TRUE)) {
									pbutton->Gadget.Flags |= GFLG_GADGHIMAGE;
									pbutton->Gadget.SelectRender = pbutton->ImageOn->Image;
								} else
								pbutton->Gadget.Flags |= GFLG_GADGHNONE;
							if(pbutton->ImageAlt = GetImage(robe, win->PubScreen, c.button->ImageAltID, TRUE)) {
									pbutton->Gadget.Flags |= GFLG_EXTENDED;
									pbutton->Gadget.MoreFlags = GMORE_GADGETHELP;
								}
							pbutton->Gadget.LeftEdge = c.button->Left + xoff;
							pbutton->Gadget.TopEdge = c.button->Top + yoff;
							pbutton->Gadget.Activation = GACT_RELVERIFY;
							pbutton->Function = c.button->Function;
							pbutton->Argument = c.button->Argument;

							if(c.chunk->ChunkSize >= (offsetof(struct RobeButton, Argument_RMB) + sizeof(UWORD))) {
									pbutton->Function_RMB = c.button->Function_RMB;
									pbutton->Argument_RMB = c.button->Argument_RMB;
								} else {
									pbutton->Function_RMB = 0xffff;
									pbutton->Argument_RMB = 0xffff;
								}

							if(c.chunk->ChunkSize >= (offsetof(struct RobeButton, GroupID) + sizeof(UWORD)))
								pbutton->GroupID = c.button->GroupID;

							switch(pbutton->Function) {
									case	PRT_BUTTON_WDEPTH:
										typ = GTYP_WDEPTH;
										break;
									case	PRT_BUTTON_WCLOSE:
										typ = GTYP_CLOSE;
										break;
									case	PRT_BUTTON_WDRAG:
										typ = GTYP_WDRAGGING;
										pbutton->Gadget.Activation = GACT_IMMEDIATE|GACT_RELVERIFY;
//										Printf("Wdragging = %ld,%ld,%ld,%ld\n", pbutton->Gadget.LeftEdge, pbutton->Gadget.TopEdge, pbutton->Gadget.Width, pbutton->Gadget.Height);
										break;
									case	PRT_BUTTON_ALARM:
									case	PRT_BUTTON_PAUSE:
									case	PRT_BUTTON_RECORD:
										pbutton->Gadget.Activation |= GACT_TOGGLESELECT;
									default:
										typ = GTYP_BOOLGADGET;
								}
							pbutton->Gadget.GadgetType = typ;
							pbutton->Gadget.GadgetID = pbutton->Function;
							pbutton->Gadget.UserData = pbutton;
						}
					break;
				case	ID_PRAYER_SLIDER:
					if(pslider->Next = AllocPooled(win->MemoryPool, sizeof(*pslider))) {
							tmp = pslider;
							pslider = pslider->Next;
							pslider->ImageOff = GetImage(robe, win->PubScreen, c.slider->ImageOffID, TRUE);
							pslider->ImageOn = GetImage(robe, win->PubScreen, c.slider->ImageOnID, TRUE);
							pslider->Function = c.slider->Function;
							pslider->Argument = c.slider->Argument;
							pslider->Flags |= c.slider->Flags;
							if(pslider->Flags & PRF_SLIDER_TRANSPARENT)
								pslider->Background = AllocBitMap(c.slider->Width, c.slider->Height, win->PubScreen->RastPort.BitMap->Depth, BMF_MINPLANES, win->PubScreen->RastPort.BitMap);

							if(c.chunk->ChunkSize >= (offsetof(struct RobeSlider, GroupID) + sizeof(UWORD)))
								pslider->GroupID = c.slider->GroupID;

							pslider->Gadget = NewObject(robe->PrayerPropClass, NULL,
									GA_Left, c.slider->Left + xoff,
									GA_Top, c.slider->Top + yoff,
									GA_Width, c.slider->Width,
									GA_Height, c.slider->Height,
									GA_ID, pslider->Function | 0x0100,
									GA_UserData, pslider,
									PGA_Freedom, (pslider->Flags & PRF_SLIDER_VERTICAL) ? FREEVERT : FREEHORIZ,
									pslider->ImageOff ? GA_Image : TAG_IGNORE, pslider->ImageOff ? pslider->ImageOff->Image : NULL,
									pslider->ImageOn ? GA_SelectRender : TAG_IGNORE, pslider->ImageOn ? pslider->ImageOn->Image : NULL,
									PPA_BitMap, pslider->Background,
									TAG_MORE, &more_prop_tags[0]);

							if(pslider->Gadget == NULL) {
									FreeBitMap(pslider->Background);
									FreePooled(win->MemoryPool, pslider, sizeof(*pslider));
									pslider = tmp;
									pslider->Next = NULL;
								}
						}
					break;
				case	ID_PRAYER_LEVEL:
					if(plevel->Next = AllocPooled(win->MemoryPool, sizeof(*plevel) + (sizeof(struct Rectangle) * c.level->NumLeds))) {
							plevel = plevel->Next;
							plevel->NumLeds = c.level->NumLeds;
							plevel->LedHigh = c.level->LedHigh - 1;
							plevel->ColorOff = getcolor(win, c.level->ColorOff);
							plevel->ColorOnLow = getcolor(win, c.level->ColorOnLow);
							plevel->ColorOnHigh = getcolor(win, c.level->ColorOnHigh);
							plevel->Type = c.level->Type;

							if(c.chunk->ChunkSize >= (offsetof(struct RobeLevel, GroupID) + sizeof(UWORD)))
								plevel->GroupID = c.level->GroupID;

							x = c.level->Left + xoff;
							y = c.level->Top + yoff;

							for(i = 0; i < plevel->NumLeds; i++) {
									if(c.level->Flags & PRF_LEVEL_VERTICAL) {
											rect = &plevel->Coords[plevel->NumLeds - i - 1];
										} else {
											rect = &plevel->Coords[i];
										}
									rect->MinX = x;
									rect->MinY = y;
									rect->MaxX = x + c.level->LedWidth - 1;
									rect->MaxY = y + c.level->LedHeight - 1;

									if(c.level->Flags & PRF_LEVEL_VERTICAL)
										y += (c.level->LedHeight + c.level->LedSpace);
									else
										x += (c.level->LedWidth + c.level->LedSpace);
								}
						}
					break;
				case	ID_PRAYER_DISPLAY:
					if(pdisplay->Next = AllocPooled(win->MemoryPool, sizeof(*pdisplay))) {
							if(pdisplay->Next->Font = GetFont(robe, c.display->FontID)) {
									pdisplay = pdisplay->Next;
									pdisplay->Rect.MinX = c.display->Left + xoff;
									pdisplay->Rect.MinY = c.display->Top + yoff;
									pdisplay->Rect.MaxX = pdisplay->Rect.MinX + c.display->Width;
									pdisplay->Rect.MaxY = pdisplay->Rect.MinY + c.display->Height;
									pdisplay->BgColor = getcolor(win, c.display->BgColor);
									pdisplay->FgColor = getcolor(win, c.display->FgColor);
									pdisplay->Type = c.display->Type;
									pdisplay->Flags = c.display->Flags;

									if(c.chunk->ChunkSize >= (offsetof(struct RobeDisplay, GroupID) + sizeof(UWORD)))
										pdisplay->GroupID = c.display->GroupID;


									if(pdisplay->Region = NewRegion()) {
											if(!OrRectRegion(pdisplay->Region, &pdisplay->Rect)) {
													DisposeRegion(pdisplay->Region);
													pdisplay->Region = NULL;
												}
										}
									if(pdisplay->Flags & PRF_DISPLAY_TRANSPARENT)
										pdisplay->Background = AllocBitMap(c.display->Width, c.display->Height, win->PubScreen->RastPort.BitMap->Depth, BMF_MINPLANES, win->PubScreen->RastPort.BitMap);
								} else {
									FreePooled(win->MemoryPool, pdisplay->Next, sizeof(*pdisplay));
									pdisplay->Next = NULL;
								}
						}
					break;
				case	ID_PRAYER_LED:
					if(pled->Next = AllocPooled(win->MemoryPool, sizeof(*pled))) {
							pled = pled->Next;
							pled->Rect.MinX = c.led->Left + xoff;
							pled->Rect.MinY = c.led->Top + yoff;
							pled->Rect.MaxX = pled->Rect.MinX + c.led->Width;
							pled->Rect.MaxY = pled->Rect.MinY + c.led->Height;
							pled->ColorOff = getcolor(win, c.led->ColorOff);
							pled->ColorOn = getcolor(win, c.led->ColorOn);
							pled->Pattern = c.led->Pattern;

							if(c.chunk->ChunkSize >= (offsetof(struct RobeLed, GroupID) + sizeof(UWORD)))
								pled->GroupID = c.led->GroupID;

						}
					break;
				case	ID_PRAYER_GROUP:
					if(pgroup->Next = AllocPooled(win->MemoryPool, sizeof(*pgroup))) {
							pgroup = pgroup->Next;
							pgroup->Rect.MinX = c.group->Left + xoff;
							pgroup->Rect.MinY = c.group->Top + yoff;
							pgroup->Rect.MaxX = pgroup->Rect.MinX + c.group->Width;
							pgroup->Rect.MaxY = pgroup->Rect.MinY + c.group->Height;
							pgroup->Color = getcolor(win, c.group->Color);
							pgroup->Image = GetImage(robe, win->PubScreen, c.group->ImageID, TRUE);
							pgroup->GroupID = c.group->GroupID;
						}
					break;
			}
		c.cptr += c.chunk->ChunkSize;
	}
}

#undef GetImage

static struct ImageData *GetImage2(struct PrayerRobe *robe, struct Screen *scr, UWORD id, BOOL remap, char *func, ULONG line)
{
struct ImageData *img;
img = GetImage(robe, scr, id, remap);
if((img == NULL) && (id != 0))
	Printf("GetImage2: %s()/%ld\n", func, line);
return(img);
}


static struct	ImageData *GetImage(struct PrayerRobe *robe, struct Screen *scr, UWORD id, BOOL remap)
{
struct	PIFHdr	*pif;
struct	ImageData	*img = NULL;
struct	ImageCache *ic;

if(id == 0)
	return(NULL); // no image

for(ic = robe->ICache; ic; ic = ic->next) {
		if(ic->id == id)
			return(ic->image);
	}

if(id > robe->NumImages) {
		Printf("GetImage: ImageID %ld out of range\n", id);
		return(NULL);
	}

if(pif = robe->Images[id - 1]) {
		if(img = PAlloc(sizeof(*img))) {
				if(img->Image = NewObject(robe->PrayerImageClass, NULL, 
																		PIA_PIF, pif,
																		PIA_DestScr, scr,
																		PIA_Remap, remap, // mask ? FALSE : robe->RemapFlag,
																		PIA_Precision, robe->RemapPrecision,
																		TAG_DONE)) {
						GetAttr(PIA_BitMap, img->Image, (ULONG *)&img->BitMap);
					} else Printf("GetImage: NewObject() error\n");
			} else Printf("GetImage: PAlloc error\n");
	} else Printf("GetImage: No such image (%ld)\n", id);

if(img && (ic = PAlloc(sizeof(*ic)))) {
		ic->next = robe->ICache;
		robe->ICache = ic;
		ic->image = img;
		ic->id = id;
	}

return(img);
}

static struct TextFont *GetFont(struct PrayerRobe *robe, UWORD id)
{
if(id == 0)
	return(NULL); // no font
if(id > robe->NumFonts) {
		Printf("GetFont: FontID %ld out of range\n", id);
		return(NULL);
	}
if(robe->Fonts[id - 1])
	return(robe->Fonts[id - 1]);

Printf("GetFont: No such font (%ld)\n", id);
return(NULL);
}

static void safeCloseWindow(struct Window *qwin)
{
struct	IntuiMessage	*qmsg;
struct	Node	*qsucc;

if(qwin->UserPort)
	{
		Forbid();
		qmsg = (struct IntuiMessage *) qwin->UserPort->mp_MsgList.lh_Head;
		while(qsucc = qmsg->ExecMessage.mn_Node.ln_Succ)
			{
				if(qmsg->IDCMPWindow == qwin)
					{
						Remove((struct Node *) qmsg);
						ReplyMsg((struct Message *) qmsg);
					}
				qmsg = (struct IntuiMessage *) qsucc;
			}
		qwin->UserPort = NULL;
		ModifyIDCMP(qwin, 0L);
		Permit();
	}

CloseWindow(qwin);
}


static void privateDisplayLevel(struct RastPort *rp, struct PrayerLevel *level, WORD value)
{
register long i = 0, step, v = 0, h, j;
register long p;
register struct	Rectangle *rect;
UWORD	col,cof,coh;
UWORD leh;

step = (18000 / level->NumLeds);

if(step == 0)
	return;

i = (level->PeakFreeze / step);
h = (level->PeakCurrent / step);
v = (value / step);

if(i < v)
	i = v;
if(i < h)
	i = h;

i++;

if(i > level->NumLeds)
	i = level->NumLeds;

if(i > 0) {
		rect = &level->Coords[0];

		if(level->PeakCounter == 0) {
				level->PeakFreeze = level->PeakCurrent;
				level->PeakCurrent = 0;
				level->PeakCounter = 5;
			}

		level->PeakCounter--;

		p = level->PeakFreeze / step;

		if(value > level->PeakCurrent)
			level->PeakCurrent = value;

		h = (rect->MaxY - rect->MinY);

		col = level->ColorOnLow;
		coh = level->ColorOnHigh;
		cof = level->ColorOff;
		leh = level->LedHigh;

		for(j = 0; j < i; j++) {
				v += step;
				if((v <= value) || (j == p)) {
						if(j < leh) {
								if(rp->FgPen != col)
									SetAPen(rp, col);
							} else {
								if(rp->FgPen != coh)
									SetAPen(rp, coh);
							}
					} else {
						if(rp->FgPen != cof)
							SetAPen(rp, cof);
					}
				if(h) {
						RectFill(rp, (long)rect->MinX, (long)rect->MinY, (long)rect->MaxX, (long)rect->MaxY);
					} else {
						Move(rp, (long)rect->MinX, (long)rect->MinY);
						Draw(rp, (long)rect->MaxX, (long)rect->MaxY);
					}
				rect++;
			}
	}
}

static void InitTranspElements(UWORD w, UWORD h, APTR bg, struct PrayerWindow *win)
{
struct PrayerDisplay *pdisplay;
struct PrayerSlider *pslider;

for(pdisplay = win->DisplayList; pdisplay; pdisplay = pdisplay->Next) {
		if(pdisplay->Background) {
				BltBitMap(bg, pdisplay->Rect.MinX,
											pdisplay->Rect.MinY,
											pdisplay->Background,
											0, 0,
											pdisplay->Rect.MaxX - pdisplay->Rect.MinX,
											pdisplay->Rect.MaxY - pdisplay->Rect.MinY,
											0xc0, 0xff, NULL);
			}
	}

for(pslider = win->SliderList; pslider; pslider = pslider->Next) {
		if(pslider->Background) {
//Printf("InitTranspElements(... slider->Background=0x%08lx ...), bg=0x%08lx (x=%ld, y=%ld, width=%ld, height=%ld, depth=%ld)\n", pslider->Background, bg, pslider->Gadget->LeftEdge, pslider->Gadget->TopEdge, pslider->Gadget->Width, pslider->Gadget->Height, GetBitMapAttr(pslider->Background, BMA_DEPTH));
//BM_Request(bg, w, h);
				BltBitMap(bg, pslider->Gadget->LeftEdge,
											pslider->Gadget->TopEdge,
											pslider->Background,
											0, 0,
											pslider->Gadget->Width,
											pslider->Gadget->Height,
											0xc0, 0xff, NULL);
//				BM_Request(pslider->Background, pslider->Gadget->Width, pslider->Gadget->Height);
			}
	}
}

static APTR ReAlloc(APTR olddata, ULONG oldsize, ULONG newsize)
{
APTR newdata;
if(newdata = PAlloc(newsize)) {
		if(olddata && oldsize)
			CopyMem(olddata, newdata, (newsize < oldsize) ? newsize : oldsize);
	}
PFree(olddata);
return(newdata);
}


static BOOL RellocChunks(UBYTE *chunks, LONG chnklen, struct PrayerRobe *robe)
{
BOOL rc = TRUE;
UWORD	baseFontID = 0, baseImageID = 0, baseWindowID = 0, baseTextID = 0, baseMenuID = 0;
union robechunk c;
APTR bar;
LONG mlen = chnklen, i;
char ab[32];

c.cptr = chunks;

bar = CreateProgressBar(NULL, "Loading robe ...");

while(rc && (chnklen > 0)) {
		if((ULONG)c.cptr & 1) {
				Printf("Segfault: robedata corrupted\n");
				rc = FALSE;
				break;
			}

		if(c.chunk->ChunkType == 0L)
			break;

		switch(c.chunk->ChunkType) {
				case	ID_PRAYER_ROBEINFO:
					baseFontID = robe->NumFonts;
					baseImageID = robe->NumImages;
					baseWindowID = robe->NumWindows;
					baseTextID = robe->NumTexts;
					baseMenuID = robe->NumMenus;
					
					if(c.info->RobeVersion >= 2) {
							if(c.info->NumMenus) {
									robe->NumMenus += c.info->NumMenus;
									robe->MenuItems = ReAlloc(robe->MenuItems, baseMenuID * sizeof(APTR), (robe->NumMenus) * sizeof(APTR));
									if(!robe->MenuItems)
										rc = FALSE;
								}
						}

					if(c.info->NumTexts) {
							robe->NumTexts += c.info->NumTexts;
							robe->Texts = ReAlloc(robe->Texts, baseTextID * sizeof(APTR), robe->NumTexts * sizeof(APTR));
							if(!robe->Texts)
								rc = FALSE;
						}
					if(c.info->NumFonts) {
							robe->NumFonts += c.info->NumFonts;
							robe->Fonts = ReAlloc(robe->Fonts, baseFontID * sizeof(APTR), robe->NumFonts * sizeof(APTR));
							if(!robe->Fonts)
								rc = FALSE;
						}
					if(c.info->NumWindows) {
							robe->NumWindows += c.info->NumWindows;
							robe->RWindows = ReAlloc(robe->RWindows, baseWindowID * sizeof(APTR), robe->NumWindows * sizeof(APTR));
							robe->Windows = ReAlloc(robe->Windows, baseWindowID * sizeof(APTR), robe->NumWindows * sizeof(APTR));
							robe->Positions = ReAlloc(robe->Positions, baseWindowID * sizeof(struct PrayerPosition), robe->NumWindows * sizeof(struct PrayerPosition));
							if(!(robe->Windows && robe->RWindows && robe->Positions))
								rc = FALSE;
						}

					if(c.info->NumImages) {
							robe->NumImages += c.info->NumImages;
							robe->Images = ReAlloc(robe->Images, baseImageID * sizeof(APTR), robe->NumImages * sizeof(APTR));
							if(!robe->Images)
								rc = FALSE;
						}
					break;
				case	ID_PRAYER_WINDOW:
					c.window->WindowID += baseWindowID;
					if(c.window->ImageID)
						c.window->ImageID += baseImageID;
					if(c.window->ActiveImageID)
						c.window->ActiveImageID += baseImageID;
					if(c.window->InActiveImageID)
						c.window->InActiveImageID += baseImageID;
					if((c.window->ChunkSize > 22) &&
						(c.window->TranspMaskID)) {
							c.window->TranspMaskID += baseImageID;
							if(CGXSystemBase == NULL) {
/*									if(rc)
										rtEZRequestA("This robe file requires cgxsystem.library v42 or better\n(Cybergraphics release 4)", "Sorry", NULL, NULL, NULL);
*/
									rc = FALSE;
								}
						}

					if((c.window->WindowID < 1) || (c.window->WindowID > robe->NumWindows)) {
							Printf("Error: WindowID out of range\n");
							rc = FALSE;
						} else {
							if(robe->RWindows)
								robe->RWindows[c.window->WindowID - 1] = c.window;
							else {
									Printf("Error: Window index not initialized\n");
									rc = FALSE;
								}
						}
					break;
				case	ID_PRAYER_TEXT:
					c.text->TextID += baseTextID;
					if((c.text->TextID < 1) || (c.text->TextID > robe->NumTexts)) {
							Printf("Error: TextID out of range\n");
						} else {
							if(robe->Texts) {
									robe->Texts[c.text->TextID - 1] = InitText(c.text);
//									Printf("Got %08lx for id %ld\n", robe->Texts[c.text->TextID - 1], c.text->TextID);
								}
							else {
									Printf("Error: Text index not initialized\n");
									rc = FALSE;
								}
						}
					break;
				case	ID_PRAYER_FONT:
					c.font->FontID += baseFontID;
					if((c.font->FontID < 1) || (c.font->FontID > robe->NumFonts)) {
							Printf("Error: FontID out of range\n");
							rc = FALSE;
						} else {
							if(robe->Fonts)
								robe->Fonts[c.font->FontID - 1] = InitFont(c.font);
							else {
									Printf("Error: Font index not initialized\n");
									rc = FALSE;
								}
						}
					break;
				case	ID_PRAYER_IMAGE:
					c.image->ImageID += baseImageID;
					if((c.image->ImageID < 1) || (c.image->ImageID > robe->NumImages)) {
							Printf("Error: ImageID out of range\n");
							rc = FALSE;
						} else {
							if(robe->Images)
								robe->Images[c.image->ImageID - 1] = &c.image->Image;
							else {
									Printf("Error: Image index not initialized\n");
									rc = FALSE;
								}
						}
					break;
				case	ID_PRAYER_MENU:
/*
					if(c.menu->Function == PRT_BUTTON_OPENWINDOW)
						c.menu->Argument += baseWindowID;
					if(c.menu->Function == PRT_BUTTON_TEXT)
						c.menu->Argument += baseTextID;
					if(robe->ItemsInUse < robe->NumMenus) {
							if(robe->MenuItems[robe->ItemsInUse] = PM_MakeItem(PM_Title, c.menu->Title, PM_UserData, (c.menu->Function | 0x0300) | (c.menu->Argument << 16) ,TAG_DONE))
								robe->ItemsInUse++;
							else
								Printf("Error while creating item..\n");
						} else Printf("MenuBuffer Overflow: Robe corrupted\n");
*/
					break;
				case	ID_PRAYER_GROUP:
					if(c.group->ImageID)
						c.group->ImageID += baseImageID;
					break;
				case	ID_PRAYER_BUTTON:
					if(c.button->ImageOffID)
						c.button->ImageOffID += baseImageID;
					if(c.button->ImageOnID)
						c.button->ImageOnID += baseImageID;
					if(c.button->ImageAltID)
						c.button->ImageAltID += baseImageID;
					if(c.button->Function == PRT_BUTTON_OPENWINDOW)
						c.button->Argument += baseWindowID;
					if(c.button->Function == PRT_BUTTON_TEXT)
						c.button->Argument += baseTextID;
					if(c.button->Function_RMB == PRT_BUTTON_OPENWINDOW)
						c.button->Argument_RMB += baseWindowID;
					if(c.button->Function_RMB == PRT_BUTTON_TEXT)
						c.button->Argument_RMB += baseTextID;
					break;
				case	ID_PRAYER_SLIDER:
					if(c.slider->ImageOffID)
						c.slider->ImageOffID += baseImageID;
					if(c.slider->ImageOnID)
						c.slider->ImageOnID += baseImageID;
					if(c.slider->ImageAltID)
						c.slider->ImageAltID += baseImageID;
					break;
				case	ID_PRAYER_DISPLAY:
					if(c.display->FontID)
						c.display->FontID += baseFontID;
					break;
				case	ID_PRAYER_LISTER:
					if(c.lister->FontID)
						c.lister->FontID += baseFontID;
					break;
			}
		chnklen -= c.chunk->ChunkSize;
		c.cptr += c.chunk->ChunkSize;

		if(mlen && bar) {
				i = ((mlen - chnklen) * 100) / mlen;
				SPrintf(ab, "%ld%%", i);
				UpdateProgressBar(bar, i, ab);
			}
	}

if(bar)
	EndProgressBar(bar);

return(rc);
}

static BOOL ShowPrayerWindowsFlags(struct PrayerRobe *robe, ULONG flags)
{
BOOL rc = TRUE;
LONG i;

i = robe->NumWindows;
while(i--) {
		if(((i == 1) || robe->Positions[i].o) && robe->RWindows[i] && (robe->RWindows[i]->Flags & flags)) {
				if(!OpenPrayerWindow(robe, i + 1)) {
						Printf("AutoOpen window initalization failure\n");
						rc = FALSE;
						break;
					} 
			}
	}

return(rc);
}

static void ConnectBobs(struct PrayerWindow *win)
{
ULONG w, h, d;
struct PrayerLister *plister;

d = GetBitMapAttr(win->PubScreen->RastPort.BitMap, BMA_DEPTH);

for(plister = win->ListerList; plister; plister = plister->Next) {

		GetAttr(PLA_ItemWidth, plister->Gadget, &w);
		GetAttr(PLA_ItemHeight, plister->Gadget, &h);

		plister->Bob = CreateBob(win->Drag, w, h, d);

		SetGadgetAttrs(plister->Gadget, win->Window, NULL, PLA_Bob, plister->Bob, TAG_DONE);
	}
}

static void ConnectBoopsi(struct PrayerWindow *win)
{
struct PrayerSlider *pslider;
struct PrayerLister *plister;

for(pslider = win->SliderList; pslider; pslider = pslider->Next) {
		if(pslider->Function == PRT_SLIDER_LISTER) {
				if(pslider->Model = GetModel(win, pslider->Argument)) {
						SetAttrs(pslider->Gadget, ICA_TARGET, pslider->Model,
																			ICA_MAP, MapSliderToLV,
																			TAG_DONE);
						DoMethod(pslider->Model, OM_ADDMEMBER, pslider->Gadget);
					}
			}
	}

for(plister = win->ListerList; plister; plister = plister->Next) {
		if(plister->Model = GetModel(win, plister->Argument)) {
				SetAttrs(plister->Gadget, ICA_TARGET, plister->Model,
																	ICA_MAP, MapLVToSlider,
																	TAG_DONE);
				DoMethod(plister->Model, OM_ADDMEMBER, plister->Gadget);
			}
	}
}


static void DisConnectBoopsi(struct PrayerWindow *win)
{
struct PrayerSlider *pslider;
struct PrayerLister *plister;

for(pslider = win->SliderList; pslider; pslider = pslider->Next) {
		if(pslider->Model)
			DoMethod(pslider->Model, OM_REMMEMBER, pslider->Gadget);
	}

for(plister = win->ListerList; plister; plister = plister->Next) {
		if(plister->Model)
			DoMethod(plister->Model, OM_REMMEMBER, plister->Gadget);
	}
}

static APTR	GetModel(struct PrayerWindow *win, UWORD id)
{
struct PrayerModel *pmodel;
APTR ico;

for(pmodel = win->ModelList; pmodel; pmodel = pmodel->Next) {
		if(pmodel->Id == id)
			return(pmodel->Model);
	}
if(pmodel = PAlloc(sizeof(*pmodel))) {
		pmodel->Next = win->ModelList;
		win->ModelList = pmodel;
		pmodel->Id = id;
		if(pmodel->Model = NewObjectA(NULL, "modelclass", NULL)) {
				if(ico = NewObject(NULL, "icclass",
															ICA_TARGET, ICTARGET_IDCMP,
//															ICA_MAP, MapLVToIDCMP,
															TAG_DONE)) {
						DoMethod(pmodel->Model, OM_ADDMEMBER, ico);
					}
			}
		return(pmodel->Model);
	}
return(NULL);
}

static char *InitText(struct RobeText *text)
{
ULONG	e, d;
char	*zp;
if(zp = PAlloc(text->RawLen)) {
		d = text->RawLen;
		if(e = uncompress(zp, &d, text->CompressedData, text->ChunkSize - sizeof(*text))) {
				Printf("Z uncompress error: %ld\n", e);
				PFree(zp);
				zp = NULL;
			}
	}
return(zp);
}

static void FreeImageCached(struct PrayerRobe *robe, struct ImageData *img)
{
struct ImageCache *ic;

for(ic = robe->ICache; ic; ic = ic->next) {
		if(ic->image == img) {
				/* free or not ? */
			}
	}
}

static void ImageCacheCleanup(struct PrayerRobe *robe)
{
struct ImageCache *ic;

while(robe->ICache) {
		ic = robe->ICache->next;
		FreeImage(robe->ICache->image);
		PFree(robe->ICache);
		robe->ICache = ic;
	}
}

static void FreeImage(struct ImageData *img)
{
if(img) {
		DisposeObject(img->Image);
		PFree(img);
	}
}


static int SetTransRegion(struct PrayerRobe *robe, struct PrayerWindow *win, UWORD id)
{
ULONG rawlen;
UBYTE *raw, *chunky;
struct PIFHdr *pif;
int i = 0;
struct Rectangle rect = {0};
UWORD w,h,x,y;
BOOL r = FALSE;
struct RegionRectangle *rr;
struct Region *Region;

if(id == 0 || id > robe->NumImages)
	return(0); // no image

if(pif = robe->Images[id - 1]) {
		if(IS_PIF(pif->Id) && (PIF_VERSION(pif->Id) <= 0)) {
				rawlen = (pif->Width * pif->Height) + (pif->NumColors * sizeof(struct PrayerPalette));
				if(raw = PAlloc(rawlen)) {
						if(!uncompress(raw, &rawlen, pif->CompressedData, pif->DataLength)) {
								chunky = raw + (pif->NumColors * sizeof(struct PrayerPalette));
								w = pif->Width;
								h = pif->Height;

								if(Region = NewRegion()) {
										for(y = 0; y < h; y++) {
												for(x = 0; x < w; x++) {
														if(!*chunky++) {
																if(!r) {
																		rect.MinX = x;
																		rect.MinY = y;
																		rect.MaxX = x;
																		rect.MaxY = y;
																		r = TRUE;
																	} else {
																		rect.MaxX = x;
																	}
															} else {
																if(r) {
																		OrRectRegion(Region, &rect);
																		r = FALSE;
																	}
															}
													}
												if(r) {
														OrRectRegion(Region, &rect);
														r = FALSE;
													}
											}
										for(rr = Region->RegionRectangle; rr; rr = rr->Next) {i++;}
										win->TranspRegion = Region;
									}
							}
						PFree(raw);
					}
			}
	}
return(i);
}
