/*
 * =-=-=-=-=-=-=-=-=-=-=-  NOTICE  =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * The following is the license that applies to this copy of the Prayer2
 * software.  For a license to use the Prayer2 software under conditions
 * other than those described here, please contact MindNever Software at
 * following addresses:
 *
 * MindNever Software <info@mindnever.org>
 * MindNever Software +381 63 550-161 or +381 64 220-7115
 * 
 * =-=-=-=-=-=-=-=-=-=-  COPYRIGHT  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Prayer2 is:
 *   Copyright (c) 1999-2001 by Tamara Cetinjski <tamara@mindnever.org>
 *   Copyright (c) 2001-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *   Copyright (c) 1999-2002 by MindNever Software Group <info@mindnever.org>
 *
 * Tamara Cetinjski is
 *   Copyright (c) 1999-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *
 *
 * =-=-=-=-=-=-=-=-=-=-  LICENSE  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Redistributions in any form must be accompanied by information on
 *    how to obtain complete source code for the Prayer2 software and any
 *    accompanying software that uses the Prayer2 software.  The source code
 *    must either be included in the distribution or be available for no
 *    more than the cost of distribution plus a nominal fee, and must be
 *    freely redistributable under reasonable conditions.  For an
 *    executable file, complete source code means the source code for all
 *    modules it contains.  It does not include source code for modules or
 *    files that typically accompany the major components of the operating
 *    system on which the executable file runs.
 *
 * =-=-=-=-=-=-=-=-=-=-  NO WARRANTY =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
 *
 * THIS SOFTWARE IS PROVIDED BY MINDNEVER SOFTWARE ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
 * NON-INFRINGEMENT, ARE DISCLAIMED.  IN NO EVENT SHALL MINDNEVER SOFTWARE
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, HAIR, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
#ifndef _PRAYER2_GUI_H_
#define _PRAYER2_GUI_H_

#include	<intuition/intuition.h>
#include	<graphics/gfx.h>

#include "prayer2_prefs.h"
#include "PrayerListClass.h"
#include "drag.h"


enum	{
	PRT_DISPLAY_SONGTITLE,
	PRT_DISPLAY_ARTIST,
	PRT_DISPLAY_ALBUM,
	PRT_DISPLAY_OTHER,
	PRT_DISPLAY_TIME,
	PRT_DISPLAY_OUTPUT,
	PRT_DISPLAY_PLAYMODE,
	PRT_DISPLAY_ALARM,
	PRT_DISPLAY_SLEEP,

	PRT_DISPLAY_RECSTART,
	PRT_DISPLAY_RECCURRENT,
	PRT_DISPLAY_VOLUME,

	PRT_DISPLAY__MAX
};

enum {
	PRT_BUTTON_PLAY,
	PRT_BUTTON_STOP,
	PRT_BUTTON_PAUSE,
	PRT_BUTTON_NEXT,
	PRT_BUTTON_PREV,
	PRT_BUTTON_OPENWINDOW,
	PRT_BUTTON_QUIT,
	PRT_BUTTON_ICONIFY,
	PRT_BUTTON_PLAYMODE,
	PRT_BUTTON_WDEPTH,
	PRT_BUTTON_WCLOSE,
	PRT_BUTTON_WDRAG,
	PRT_BUTTON_ADDFILES,
	PRT_BUTTON_REM,
	PRT_BUTTON_LOADLIST,
	PRT_BUTTON_SAVELIST,
	PRT_BUTTON_SORTLIST,
	PRT_BUTTON_TEXT,
	PRT_BUTTON_PREFS,
	PRT_BUTTON_ABOUT,

	PRT_BUTTON_SETHOUR,
	PRT_BUTTON_SETMINUTE,
	PRT_BUTTON_SNOOZE,
	PRT_BUTTON_ALARM,

	PRT_BUTTON_RECORD,
	PRT_BUTTON_GROUP,

	PRT_BUTTON__MAX
};

enum {
	PRT_SLIDER_VOLUME,
	PRT_SLIDER_POSITION,
	PRT_SLIDER_LISTER,
	PRT_SLIDER_SLEEP,
	PRT_SLIDER__MAX
	};

enum {
		PRT_LEVEL_LEFT,
		PRT_LEVEL_RIGHT,
		PRT_LEVEL__MAX
	};

enum {
		PRT_LISTER_PLAYLIST,
		PRT_LISTER__MAX
	};

struct	PrayerPosition {
		UWORD	x,y,o;
	};

struct	PrayerRobe {
		ULONG	NumFonts;
		ULONG	NumImages;
		ULONG	NumWindows;
		ULONG	NumTexts;
		ULONG	NumMenus;
		struct	PIFHdr	**Images;		//packed images
		struct	TextFont	**Fonts;   //unpacked fonts
		struct	RobeWindow	**RWindows;  //window definition index
		struct	PrayerWindow	**Windows;	//active windows index
		struct	PrayerPosition *Positions;
		APTR		**MenuItems;
		ULONG		ItemsInUse;
		APTR		SubMenu;
		char		**Texts;
		char		*PubScreenName;
		struct	MsgPort	*UserPort;
		struct	MsgPort	*AppPort;
		APTR		PrayerListClass;
		APTR		PrayerPropClass;
		BYTE		RemapFlag;
		BYTE		RemapPrecision;
		BOOL		InitOk;
		BOOL		RTG;
		struct	ImageCache *ICache;
		char		*DisplayTexts[PRT_DISPLAY__MAX];
		struct	{
				ULONG	Top;
				ULONG	Selected;
			} *ListerData;
		APTR		PrayerImageClass;
		UBYTE		RobeData[0];	//raw data...
	};

struct	PrayerWindow {
		struct	Window *Window;
		struct	ExtGadget	*GadgetList;
		struct	PrayerDisplay	*DisplayList;
		struct	PrayerButton	*ButtonList;
		struct	PrayerSlider	*SliderList;
		struct	PrayerLed			*LedList;
		struct	PrayerLevel		*LevelList;
		struct	PrayerLister	*ListerList;
		struct	PrayerModel		*ModelList;
		struct	PrayerGroup		*GroupList;
		APTR		MemoryPool;
		UWORD		WindowID;
		UWORD		WindowWidth;
		UWORD		WindowHeight;
		UWORD		WindowX;
		UWORD		WindowY;
		struct	Screen	*PubScreen;
		struct	AppWindow *App;
		struct	ImageData	*ActiveImage;
		struct	ImageData	*InActiveImage;
		UWORD		StateImageX;
		UWORD		StateImageY;
		struct	PrayerButton	*LastHelp;
		ULONG		Flags;
		char	*Title;
		struct	Gadget *currentGroupList;
		UWORD		currentGroupCount;
		UWORD		currentGroupID;
		struct	PrayerRobe *Link;
		struct	Region *TranspRegion;
		DRAG *Drag;
	};

struct	PrayerModel {
	struct PrayerModel *Next;
	APTR	Model;
	ULONG	Id;
	};

struct	PrayerGroup {
		struct PrayerGroup *Next;
		struct	Rectangle	Rect;
		struct	ImageData *Image;
		WORD	Color;
		UWORD	GroupID;
	};

struct	PrayerDisplay	{
	struct	PrayerDisplay *Next;
	struct	TextFont	*Font;
	struct	Rectangle	Rect;
	WORD		BgColor;
	WORD		FgColor;
	UWORD		Type;
	UWORD		Flags;
	struct	Region	*Region;
	struct	BitMap	*Background;
	UWORD		GroupID;
	WORD		CurrentX;
	WORD		StepX;
};

struct	PrayerButton	{
	struct	PrayerButton	*Next;
	struct	ExtGadget	Gadget;
	struct	ImageData	*ImageOff;
	struct	ImageData	*ImageOn;
	struct	ImageData	*ImageAlt;
	UWORD		Function;
	UWORD		Argument;
	UWORD		Function_RMB;
	UWORD		Argument_RMB;
	UWORD		GroupID;
};

struct	PrayerSlider	{ // NOTE: This 'device' can be used as output too !
	struct	PrayerSlider *Next;
	struct	Gadget	*Gadget;
	struct	ImageData	*ImageOff;
	struct	ImageData	*ImageOn;
	UWORD		Flags;
	UWORD		Total;
	UWORD		Function;
	UWORD		Argument;
	APTR	Model;
	struct	BitMap *Background;
	UWORD	GroupID;
};

struct PrayerLister {
		struct PrayerLister *Next;
		struct Gadget *Gadget;
		struct TextFont *Font;
		LONG	FgPen,BgPen,FgPenS,BgPenS,FgPenH,BgPenH;
		UWORD	Function;
		UWORD	Argument;
		APTR	Model;
		UWORD	GroupID;
		UWORD	ListerID;
		BOB *Bob;
	};

struct	PrayerLevel	{
	struct	PrayerLevel	*Next;
	UBYTE		NumLeds;
	UBYTE		LedHigh;
	UWORD		PeakFreeze;
	UWORD		PeakCurrent;
	UBYTE		PeakCounter;
	UBYTE		Dummy1;
	WORD		ColorOff;
	WORD		ColorOnLow;
	WORD		ColorOnHigh;
	UWORD		Type;
	UWORD		GroupID;
	UWORD		LastPeak;
	struct	Rectangle	Coords[0]; // in fact 'NumLeds'
};

struct	PrayerLed	{
		struct	PrayerLed	*Next;
		struct	Rectangle	Rect;
		WORD	ColorOff;
		WORD	ColorOn;
		ULONG	Pattern;
		UBYTE	PatCnt; // current pattern bit number
		UBYTE	State;  // current state
		UWORD	GroupID;
	};

#define	PRF_RUN_ALTIMAGE	(1L << 0)

struct PrayerRobe *InitPrayerRobe(struct MsgPort *AppPort, struct PrayerPrefs *Prefs);
void FreePrayerRobe(struct PrayerRobe *robe);

struct PrayerWindow *OpenPrayerWindow(struct PrayerRobe *robe, UWORD id);
void ClosePrayerWindow(struct PrayerRobe *robe, struct PrayerWindow *prayer);

void UpdatePrayerLevel(struct	PrayerRobe *robe, UWORD lev_r, UWORD lev_l);
void UpdatePrayerSlider(struct	PrayerRobe *robe, UWORD type, UWORD total, UWORD top);
UWORD GetPrayerSlider(struct Gadget *g);
void UpdatePrayerDisplay(struct PrayerRobe *robe, UWORD type, char *text);
void UpdatePrayerButton(struct PrayerRobe *robe, UWORD type, BOOL state);

void HighLightPrayerButton(struct PrayerWindow *win, struct PrayerButton *button);
struct Gadget *GetActiveGadget(struct PrayerWindow *win);

void HidePrayerWindows(struct PrayerRobe *robe);
BOOL ShowPrayerWindows(struct PrayerRobe *robe);
BOOL SetPrayerScreen(struct PrayerRobe *robe, char *scr);
void UpdatePrayerButton(struct PrayerRobe *robe, UWORD type, BOOL state);


// global settings for listers
void SetListerList(struct PrayerRobe *robe, UWORD type, struct List *list);
void SetListerHighlight(struct PrayerRobe *robe, UWORD type, ULONG val);
void SetListerActive(struct PrayerRobe *robe, UWORD type, ULONG val);
void SetListerTop(struct PrayerRobe *robe, UWORD type, ULONG val);

ULONG GetListerActive(struct PrayerRobe *robe, UWORD type);
ULONG GetListerTop(struct PrayerRobe *robe, UWORD type);

ULONG GetListerAttr(struct PrayerRobe *robe, UWORD type, ULONG attr);
void SetListerAttr(struct PrayerRobe *robe, UWORD type, ULONG attr, ULONG val);

char *GetPrayerText(struct PrayerRobe *robe, UWORD id);

//
void SetWindowState(struct PrayerWindow *win);
struct Gadget *GetHitGadget(struct IntuiMessage *imsg);

//
void SwitchGroup(struct PrayerWindow *win, UWORD GroupID);
//

void ScrollAllDisplays(struct PrayerRobe *robe);

#endif /* _PRAYER2_GUI_H_ */
