/*
 * =-=-=-=-=-=-=-=-=-=-=-  NOTICE  =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * The following is the license that applies to this copy of the Prayer2
 * software.  For a license to use the Prayer2 software under conditions
 * other than those described here, please contact MindNever Software at
 * following addresses:
 *
 * MindNever Software <info@mindnever.org>
 * MindNever Software +381 63 550-161 or +381 64 220-7115
 * 
 * =-=-=-=-=-=-=-=-=-=-  COPYRIGHT  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Prayer2 is:
 *   Copyright (c) 1999-2001 by Tamara Cetinjski <tamara@mindnever.org>
 *   Copyright (c) 2001-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *   Copyright (c) 1999-2002 by MindNever Software Group <info@mindnever.org>
 *
 * Tamara Cetinjski is
 *   Copyright (c) 1999-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *
 *
 * =-=-=-=-=-=-=-=-=-=-  LICENSE  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Redistributions in any form must be accompanied by information on
 *    how to obtain complete source code for the Prayer2 software and any
 *    accompanying software that uses the Prayer2 software.  The source code
 *    must either be included in the distribution or be available for no
 *    more than the cost of distribution plus a nominal fee, and must be
 *    freely redistributable under reasonable conditions.  For an
 *    executable file, complete source code means the source code for all
 *    modules it contains.  It does not include source code for modules or
 *    files that typically accompany the major components of the operating
 *    system on which the executable file runs.
 *
 * =-=-=-=-=-=-=-=-=-=-  NO WARRANTY =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
 *
 * THIS SOFTWARE IS PROVIDED BY MINDNEVER SOFTWARE ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
 * NON-INFRINGEMENT, ARE DISCLAIMED.  IN NO EVENT SHALL MINDNEVER SOFTWARE
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, HAIR, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
#include	<libraries/triton.h>
#include	<exec/memory.h>
#include	<string.h>
#include	<exec/semaphores.h>
#include	<dos/dostags.h>
#include	<libraries/reqtools.h>
#include	<devices/ahi.h>

#include	<proto/ahi.h>
#include	<proto/reqtools.h>
#include	<proto/triton.h>
#include	<proto/dos.h>
#include	<proto/intuition.h>
#include	<proto/exec.h>
#include	<proto/utility.h>

#include	<stddef.h>

#define GET_HK_NAMES 1
#define GET_HK_DEFAULTS 1

#include	"prayer2_prefs.h"
#include	"prayer2_audio.h"

#include "sprintf.h"

#ifndef MAKE_ID
#define	MAKE_ID(a,b,c,d) (((a) << 24) | ((b) << 16) | ((c) << 8) | (d))
#endif

#define	SCREEN_DEFAULT	" default "

/// static data

struct PluginInfo {
	struct Node Node; // ln_Name points to Name[32];
	char	Name[32];
	char	Title[32];
	char	Description[64];
	char	Author[64];
	char	Path[2];
};

static APTR prj_Audio = NULL, prj_Visual = NULL, prj_Plugins = NULL, prj_Hotkeys = NULL;

static struct List PrefsPluginList;
static struct List PrefsHotkeyList;

static void pr_OpenHotkeysPrj(struct PrayerPrefs *prefs);
static void pr_OpenPluginPrj(struct PrayerPrefs *prefs);
static void pr_OpenVisualPrj(struct PrayerPrefs *prefs);
static void pr_OpenAudioPrj(struct PrayerPrefs *prefs);
static void PrjToFront(APTR prj);

static char *fps_str(UWORD fps);

static struct PrayerPrefs *LocalPrefs = NULL;
static ULONG	NotifyFlag = 0;
static struct SignalSemaphore PrefsLock;
static struct Process *PrefsProc = NULL;
static struct MsgPort *PrefsPort = NULL;
static struct List NullList;
static struct TR_Dimensions prefs_dim = {0}, hotkeys_dim = {0}, plugin_dim = {0}, visual_dim = {0}, audio_dim = {0};
static struct Task *PrefsParent = NULL;
static char *Screen = NULL;

static char	*cy_mode[] = {"Stereo", "Mono", NULL};
static char	*cy_freq[] = {"44.1 kHz", "22.05 kHz", "11.02 kHz", NULL};
static char	*cy_qual[] = {"Low", "Medium", "High", NULL};
static char	*cy_remap[] = {"No Remap", "Exact", "Image", "Gui", NULL};
static char *cy_icon[] = {"Only When Hidden", "Always Active", NULL};
static BYTE	prec_values[] = {-1, 0, 32};
static char hotkey_bf[256];

enum {
	ID_LAYER3_MODE,
	ID_LAYER3_FREQ,
	ID_LAYER3_QUAL,
	ID_LAYER12_MODE,
	ID_LAYER12_FREQ,
	ID_LAYER12_QUAL,
	ID_OUTPUT_DRIVER_CONFIGURE,
	ID_ROBE_TITLE,
	ID_ROBE_SELECT,
	ID_PUBSCREEN_TITLE,
	ID_PUBSCREEN_SELECT,
	ID_REMAP,
	ID_PLUGIN_LIST,
	ID_PLUGIN_DESCR,
	ID_PLUGIN_AUTHOR,
	ID_PLUGIN_ADD,
	ID_PLUGIN_REM,
	ID_PREFS_SAVE,
	ID_PREFS_USE,
	ID_PREFS_CANCEL,
	ID_AUTOFADE_IN,
	ID_AUTOFADE_OUT,
	ID_RTG,
	ID_ICON,
	ID_AUDIO_PREFS,
	ID_VISUAL_PREFS,
	ID_PLUGINS_PREFS,
	ID_HOTKEYS_PREFS,
	ID_AUTOFADEIN_INT,
	ID_AUTOFADEOUT_INT,
	ID_REFRESH_RATE,
	ID_REFRESH_RATE_TX,

	ID_CAPTURE_PLUGIN,
	ID_CAPTURE_PLUGIN_PICK,
	ID_CAPTURE_PATH,
	ID_CAPTURE_PATH_PICK,

	ID_CAPTURE_PLUGIN_DESC,
	ID_CAPTURE_PLUGIN_AUTHOR,

	ID_PLAYER_PRI,
	ID_PLAYER_PRI_INT,

	ID_CHECKBOX_AHI,
	ID_STRING_HOTKEY,
	ID_CHECKBOX_GLOBAL,
	ID_CHECKBOX_REPEAT,
	ID_HOTKEY_LIST,
};

static struct PluginInfo *GetPluginInfo(char *path);
static APTR GetNode(struct List *list, UWORD i);
static void __saveds prayer2_prefs(void);
static struct Node *CreateNode(char *zp, struct List *list);
static void pubscreen_select(struct TR_Project *project, struct PrayerPrefs *prefs);
static UWORD PickEntry(struct List *list, UWORD top, UWORD value);
static void SetPrayerPrefs(struct PrayerPrefs *prefs);
static struct PrayerPrefs *CopyPrayerPrefs(struct PrayerPrefs *src);
static void SavePrayerPrefs(void);
static void LoadPrayerPrefs(void);
static void add_plugin(struct TR_Project *proj, struct PrayerPrefs *prefs, char *dir, char *file);
static void update_plugin_display(UWORD cplugin, APTR project);
static void set_capture_plugin(struct PrayerPrefs *prefs, char *dir, char *name);

char	*StrDup(char *str);

static char	robe_name[109] = {0};
static char	plugin_name[109] = {0};
static char	robe_path[512] = "Robes";
static char	plugin_path[512] = "Plugins";


#define PREFS_HEADER_ID MAKE_ID('P','R','p','r')
#define	PREFS_PLUGIN_ID	MAKE_ID('P','R','p','l')
#define	PREFS_ROBE_ID		MAKE_ID('P','R','r','o')
#define	PREFS_SCREEN_ID	MAKE_ID('P','R','s','c')
#define PREFS_CPATH_ID	MAKE_ID('C','A','p','t')
#define PREFS_CPLUG_ID	MAKE_ID('C','A','p','l')
#define PREFS_HOTKEY_ID	MAKE_ID('P','R','h','k')

struct	PrefsChunk	{
	ULONG	ChunkType;
	ULONG	ChunkSize;
	UWORD	Version;
	UBYTE	Layer3ForceMono, Layer3FreqDiv, Layer3Qual;
	UBYTE	Layer12ForceMono, Layer12FreqDiv, Layer12Qual;
	UWORD	AudioDriverId;
	UBYTE	RemapFlag;
	BYTE	RemapPrecision;
	UWORD	AutoFadeIn;
	UWORD	AutoFadeOut;
	BOOL	RTG;
	UWORD	RefreshRate;
	ULONG	Flags;
	BYTE	PlayerPri;
	BYTE	Dummy0;
	ULONG AudioID;
	ULONG MixFreq;
};

struct HotkeyChunk {
		ULONG	ChunkType;
		ULONG	ChunkSize;
		UWORD	ID;
		UBYTE	Global;
		UBYTE Repeat;
		char 	Name[0];
	};

struct	NameChunk {
		ULONG	ChunkType;
		ULONG	ChunkSize;
		char	Name[0];
	};

///////////
// PUBLIC
///////////

void InitPrayerPrefs(BYTE notifybit)
{
InitSemaphore(&PrefsLock);
NewList(&NullList);

LoadPrayerPrefs();

if(notifybit >= 0) {
		NotifyFlag = 1L << notifybit;
		PrefsParent = FindTask(0);
	}
}

struct PrayerPrefs *GetPrayerPrefs()
{
struct PrayerPrefs *prefs = NULL;

if(LocalPrefs) {
		ObtainSemaphoreShared(&PrefsLock);
		prefs = CopyPrayerPrefs(LocalPrefs);
		ReleaseSemaphore(&PrefsLock);
	}
return(prefs);
}

void FreePrayerPrefs(struct PrayerPrefs *prefs)
{
struct Node *node;

if(prefs) {
		FreeVec(prefs->Robe);
		FreeVec(prefs->Screen);
		FreeVec(prefs->CapturePath);
		FreeVec(prefs->CapturePlugin);
		FreeVec(prefs->Hotkeys);
		while(node = RemTail(&prefs->Plugins)) {
				FreeVec(node);
			}
		FreeVec(prefs);
	}
}


void EditPrayerPrefs(char *Where)
{
BOOL fspawn = TRUE;
Forbid();
if(PrefsProc && PrefsPort) {
		Signal((APTR) PrefsProc, SIGBREAKF_CTRL_D);
		fspawn = FALSE;
	}
Permit();
if(fspawn) {
		if(LocalPrefs) {
				if(PrefsPort == NULL)
					PrefsPort = CreateMsgPort();
				if(PrefsPort) {
						while(GetMsg(PrefsPort)); // Clean port !
						FreeVec(Screen);
						Screen = StrDup(Where);

						PrefsProc = CreateNewProcTags(NP_Entry, prayer2_prefs, 
																							NP_Name, "Prayer2_Prefs_Task",
																							NP_StackSize, 10240,
																							NP_Output, Output(),
																							NP_CloseOutput, FALSE,
																							TAG_DONE);
					}
			}
	}
}


void ExitPrayerPrefs()
{
BOOL fwait = FALSE;

Forbid();
if(PrefsProc && PrefsPort) {
		Signal((APTR) PrefsProc, SIGBREAKF_CTRL_C);
		fwait = TRUE;
	}
Permit();
if(fwait) {
		while(!GetMsg(PrefsPort)) {
				WaitPort(PrefsPort);
			}
	}
if(PrefsPort)
	DeleteMsgPort(PrefsPort);

FreeVec(Screen);
FreePrayerPrefs(LocalPrefs);
}



// private functions

static void SetPrayerPrefs(struct PrayerPrefs *prefs)
{
struct PrayerPrefs *copy = NULL;
if(copy = CopyPrayerPrefs(prefs)) {
		ObtainSemaphore(&PrefsLock);
		FreePrayerPrefs(LocalPrefs);
		LocalPrefs = copy;
		ReleaseSemaphore(&PrefsLock);
		if(PrefsParent)
			Signal(PrefsParent, NotifyFlag);
	}
}


static struct PrayerPrefs *CopyPrayerPrefs(struct PrayerPrefs *src)
{	
struct PrayerPrefs *prefs = NULL;
struct Node *node;

if(src) {
		if(prefs = AllocVec(sizeof(*prefs), MEMF_ANY)) {
				CopyMem(src, prefs, sizeof(*prefs));
				prefs->Robe = StrDup(src->Robe);
				prefs->Screen = StrDup(src->Screen);
				prefs->CapturePath = StrDup(src->CapturePath);
				prefs->CapturePlugin = StrDup(src->CapturePlugin);
				NewList(&prefs->Plugins);
				for(node = src->Plugins.lh_Head; node->ln_Succ; node = node->ln_Succ) {
						CreateNode((char *)(node + 1), &prefs->Plugins);
					}
				if(prefs->Hotkeys = AllocVec(sizeof(*prefs->Hotkeys) * PRHK_MAX, MEMF_CLEAR))
					CopyMem(src->Hotkeys, prefs->Hotkeys, sizeof(*prefs->Hotkeys) * PRHK_MAX);
			}
	}
return(prefs);
}

static void __saveds prayer2_prefs()
{
static struct Message	DeathMsg = {0};

#ifdef __MORPHOS__

// Do nothing yet

#else


struct TR_Message *trmsg;
struct TR_Project *project;
struct rtFileRequester	*qRobeReq, *qPluginReq, *qPathReq;
struct Node *node = NULL;
UWORD	cplugin = 0xffff, chotkey = 0xffff;
ULONG signals, i;
BOOL	fnotdone = TRUE;
struct PrayerPrefs *prefs;
struct PluginInfo *pnode;

if(prefs = GetPrayerPrefs()) {
		NewList(&PrefsPluginList);
		NewList(&PrefsHotkeyList);

		for(node = prefs->Plugins.lh_Head; node->ln_Succ; node = node->ln_Succ) {
				GetPluginInfo((char *) (node + 1));
			}

		for(i = 0; hk_names[i]; i++) {
				CreateNode(hk_names[i], &PrefsHotkeyList);
			}

//Printf("prefs->capturepath @%08lx\n", prefs->CapturePath);

		if(qPathReq = rtAllocRequestA(RT_FILEREQ, NULL)) {
				rtChangeReqAttr(qPathReq, RTFI_Dir, prefs->CapturePath, TAG_DONE);
			}
		if(qRobeReq = rtAllocRequestA(RT_FILEREQ, NULL))
			rtChangeReqAttr(qRobeReq, RTFI_Dir, robe_path, RTFI_MatchPat, "#?.robe", TAG_DONE);
		if(qPluginReq = rtAllocRequestA(RT_FILEREQ, NULL))
			rtChangeReqAttr(qPluginReq, RTFI_Dir, plugin_path, RTFI_MatchPat, "#?.ppl", TAG_DONE);

		if(TR_OpenTriton(TRITON20VERSION, TRCA_Name, "Prayer2", TRCA_Version, "2.0", TAG_END)) {
				if(project = TR_OpenProjectTags(Application,
									WindowID(1),
									QuickHelpOn(TRUE),
									WindowTitle("Prayer2 Preferences"),
									WindowPosition(TRWP_CENTERDISPLAY),
									WindowDimensions(&prefs_dim),
									TRWI_PubScreenName, Screen,
									WindowFlags(TRWF_NOSIZEGADGET|TRWF_NOESCCLOSE),
									WindowBackfillReq,
									HorizGroupA,
										Space,
										VertGroupA,
											Space,
											GroupBox,
											ObjectBackfillB,
											HorizGroupA,
												Space,
												VertGroupA,
													Space,

													Button("_Audio Settings", ID_AUDIO_PREFS), QuickHelp("MPEG Decoding quality, mixing frequency,\nand AHI settings"),
													Space,
													Button("_Visual Settings", ID_VISUAL_PREFS), QuickHelp("Look & Feel, screen selection, and visualization refresh rate"),
													Space,
													Button("_Plugins Settings", ID_PLUGINS_PREFS), QuickHelp("Plugin manager for decoder and output capture plugins"),
													Space,
													Button("_Hotkeys", ID_HOTKEYS_PREFS), QuickHelp("Configure Hotkeys"),
													Space,

													HorizGroupE, // komandna grupa
														Button("_Save", ID_PREFS_SAVE), QuickHelp("Save and use current settings"),
														Space,
														ButtonR("_Use", ID_PREFS_USE), QuickHelp("Use but do not save current settings on disk"),
														Space,
														ButtonE("_Cancel", ID_PREFS_CANCEL), QuickHelp("Cancel all changes"),
														EndGroup, // kraj komandne grupe

													Space,
													EndGroup,
												Space,
												EndGroup,
											Space,
											EndGroup,
										Space,
										EndGroup,
									EndProject)) {

						while(fnotdone) {
								signals = TR_Wait(Application, SIGBREAKF_CTRL_C|SIGBREAKF_CTRL_D|SIGBREAKF_CTRL_E|SIGBREAKF_CTRL_F);
								if(signals & SIGBREAKF_CTRL_C)
									fnotdone = FALSE;
								if(signals & SIGBREAKF_CTRL_D)
									PrjToFront(project);
								while(trmsg = TR_GetMsg(Application)) {
										switch(trmsg->trm_Class) {
												case	TRMS_CLOSEWINDOW:

													if(trmsg->trm_Project == project)
														fnotdone = FALSE;
													else {
															if(trmsg->trm_Project == prj_Audio)
																prj_Audio = NULL;
															else if(trmsg->trm_Project == prj_Visual)
																prj_Visual = NULL;
															else if(trmsg->trm_Project == prj_Hotkeys) {
																if(chotkey != 0xffff && chotkey < PRHK_MAX) {
																		char *zp;
																		if(zp = (char *)TR_GetAttribute(prj_Hotkeys, ID_STRING_HOTKEY, 0L))
																			strncpy(prefs->Hotkeys[chotkey].Key, zp, sizeof(prefs->Hotkeys[0].Key) - 1);
																		prefs->Hotkeys[chotkey].Global = TR_GetAttribute(prj_Hotkeys, ID_CHECKBOX_GLOBAL, TRAT_Value);
																		prefs->Hotkeys[chotkey].Repeat = TR_GetAttribute(prj_Hotkeys, ID_CHECKBOX_REPEAT, TRAT_Value);
																	}
																	chotkey = 0xffff;
																	prj_Hotkeys = NULL;
																}	else if(trmsg->trm_Project == prj_Plugins) {
																		FreeVec(prefs->CapturePath);
																		prefs->CapturePath = StrDup((char *)TR_GetAttribute(prj_Plugins, ID_CAPTURE_PATH, 0));
																		prj_Plugins = NULL;
																	}

															TR_CloseProject(trmsg->trm_Project);
														}
													break;
												case	TRMS_ERROR:
													PutStr(TR_GetErrorString(trmsg->trm_Data));
													break;
												case	TRMS_NEWVALUE:
													switch(trmsg->trm_ID) {
															case	ID_LAYER3_MODE:
																prefs->Layer3ForceMono = trmsg->trm_Data;
																break;
															case	ID_LAYER12_MODE:
																prefs->Layer12ForceMono = trmsg->trm_Data;
																break;
															case	ID_LAYER3_FREQ:
																prefs->Layer3FreqDiv = 1 << trmsg->trm_Data;
																break;
															case	ID_LAYER12_FREQ:
																prefs->Layer12FreqDiv = 1 << trmsg->trm_Data;
																break;
															case	ID_LAYER3_QUAL:
																prefs->Layer3Qual = trmsg->trm_Data;
																break;
															case	ID_LAYER12_QUAL:
																prefs->Layer12Qual = trmsg->trm_Data;
																break;
															case	ID_HOTKEY_LIST:
																if(chotkey != 0xffff && chotkey < PRHK_MAX) {
																		char *zp;
																		if(zp = (char *)TR_GetAttribute(prj_Hotkeys, ID_STRING_HOTKEY, 0L))
																			strncpy(prefs->Hotkeys[chotkey].Key, zp, sizeof(prefs->Hotkeys[0].Key) - 1);
																		prefs->Hotkeys[chotkey].Global = TR_GetAttribute(prj_Hotkeys, ID_CHECKBOX_GLOBAL, TRAT_Value);
																		prefs->Hotkeys[chotkey].Repeat = TR_GetAttribute(prj_Hotkeys, ID_CHECKBOX_REPEAT, TRAT_Value);
																	}
																chotkey = trmsg->trm_Data;
																if(chotkey != 0xffff && chotkey < PRHK_MAX) {
																		TR_SetAttribute(prj_Hotkeys, ID_STRING_HOTKEY, 0L, (ULONG) prefs->Hotkeys[chotkey].Key);
																		TR_SetAttribute(prj_Hotkeys, ID_CHECKBOX_GLOBAL, TRAT_Value, prefs->Hotkeys[chotkey].Global);
																		TR_SetAttribute(prj_Hotkeys, ID_CHECKBOX_REPEAT, TRAT_Value, prefs->Hotkeys[chotkey].Repeat);
																	}
																break;
															case	ID_PLUGIN_LIST:
																cplugin = trmsg->trm_Data;
																update_plugin_display(cplugin, prj_Plugins);
																break;
															case	ID_REMAP:
																if(trmsg->trm_Data) {
																		prefs->RemapPrecision = prec_values[trmsg->trm_Data - 1];
																		prefs->RemapFlag = TRUE;
																	} else {
																		prefs->RemapFlag = FALSE;
																	}
																break;
/* FIX THIS ONE TOO */
/*
															case	ID_AUTOFADE_IN:
																prefs->AutoFadeIn = trmsg->trm_Data;
																TR_SetAttribute(prj_Misc, ID_AUTOFADEIN_INT, TRAT_Value, trmsg->trm_Data);
																break;
															case	ID_AUTOFADE_OUT:
																prefs->AutoFadeOut = trmsg->trm_Data;
																TR_SetAttribute(prj_Misc, ID_AUTOFADEOUT_INT, TRAT_Value, trmsg->trm_Data);
																break;
*/

															case	ID_REFRESH_RATE:
																prefs->RefreshRate = trmsg->trm_Data;
																TR_SetAttribute(prj_Visual, ID_REFRESH_RATE_TX, TRAT_Text, (ULONG) fps_str(trmsg->trm_Data));
																break;
															case	ID_PLAYER_PRI:
																prefs->PlayerPri = (WORD)trmsg->trm_Data;
																TR_SetAttribute(prj_Audio, ID_PLAYER_PRI_INT, TRAT_Value, prefs->PlayerPri);
																break;
															case	ID_CHECKBOX_AHI:
																if(trmsg->trm_Data)
																	prefs->AudioDriverId = 3;
																else
																	prefs->AudioDriverId = 0;
																break;
															case	ID_RTG:
																prefs->RTG = trmsg->trm_Data;
																break;
															case	ID_ICON:
																if(trmsg->trm_Data)
																	prefs->Flags |= PRF_ALWAYS_ADD_ICON;
																else
																	prefs->Flags &= ~PRF_ALWAYS_ADD_ICON;
														}
													break;
												case	TRMS_ACTION:
													switch(trmsg->trm_ID) {
															case	ID_AUDIO_PREFS:
																pr_OpenAudioPrj(prefs);
																break;
															case	ID_VISUAL_PREFS:
																pr_OpenVisualPrj(prefs);
																break;
															case	ID_PLUGINS_PREFS:
																pr_OpenPluginPrj(prefs);
																break;
															case	ID_HOTKEYS_PREFS:
																pr_OpenHotkeysPrj(prefs);
																break;
															case	ID_PUBSCREEN_SELECT:
																pubscreen_select(prj_Visual, prefs);
																break;
															case	ID_ROBE_SELECT:
																if(rtFileRequest(qRobeReq, robe_name, "Select Prayer Robe", RTFI_Flags, FREQF_PATGAD, RT_PubScrName, Screen, TAG_DONE)) {
																		FreeVec(prefs->Robe);
																		i = strlen(qRobeReq->Dir) + strlen(robe_name) + 4;
																		if(prefs->Robe = AllocVec(i, MEMF_ANY)) {
																				strcpy(prefs->Robe, qRobeReq->Dir);
																				AddPart(prefs->Robe, robe_name, i);
																				TR_SetAttribute(prj_Visual, ID_ROBE_TITLE, TRAT_Text, (ULONG) prefs->Robe);
																			}
																	}
																break;
															case	ID_CAPTURE_PLUGIN_PICK:
																if(rtFileRequest(qPluginReq, plugin_name, "Select Output Capture Plugin", RTFI_Flags, FREQF_PATGAD, RT_PubScrName, Screen, TAG_DONE)) {
																		set_capture_plugin(prefs, qPluginReq->Dir, plugin_name);
																	}
																break;
															case	ID_CAPTURE_PATH_PICK:
																if(rtFileRequest(qPathReq, plugin_name, "Select Output Capture Path", RTFI_Flags, FREQF_NOFILES , RT_PubScrName, Screen, TAG_DONE)) {
																		FreeVec(prefs->CapturePath);
																		prefs->CapturePath = StrDup(qPathReq->Dir);
																		TR_SetAttribute(prj_Plugins, ID_CAPTURE_PATH, 0L, (ULONG) prefs->CapturePath);
																	}
																break;
															case	ID_PLUGIN_ADD:
																if(rtFileRequest(qPluginReq, plugin_name, "Select Plug-In to add", RTFI_Flags, FREQF_PATGAD, RT_PubScrName, Screen, TAG_DONE)) {
																		add_plugin(prj_Plugins, prefs, qPluginReq->Dir, plugin_name);
																		cplugin = TR_GetAttribute(prj_Plugins, ID_PLUGIN_LIST, TRAT_Value);
																		update_plugin_display(cplugin, prj_Plugins);
																	}
																break;
															case	ID_PLUGIN_REM:
																if(pnode = (APTR) GetNode(&PrefsPluginList, cplugin)) {
																		TR_SetAttribute(prj_Plugins, ID_PLUGIN_LIST, 0, (ULONG) &NullList);
																		Remove((APTR) pnode);
																		TR_SetAttribute(prj_Plugins, ID_PLUGIN_LIST, 0, (ULONG) &PrefsPluginList);
																		TR_SetAttribute(prj_Plugins, ID_PLUGIN_LIST, TRAT_Value, cplugin);
																		FreeVec(pnode);
																		cplugin = TR_GetAttribute(prj_Plugins, ID_PLUGIN_LIST, TRAT_Value);
																		update_plugin_display(cplugin, prj_Plugins);
																	}
																break;
															case	ID_PREFS_CANCEL:
																fnotdone = FALSE;
																break;
															case	ID_PREFS_SAVE:
															case	ID_PREFS_USE:
																while(node = RemTail(&prefs->Plugins))
																	FreeVec(node);
																for(pnode = (APTR) PrefsPluginList.lh_Head; pnode->Node.ln_Succ; pnode = (APTR) pnode->Node.ln_Succ) {
																		CreateNode(pnode->Path, &prefs->Plugins);
																	}
																SetPrayerPrefs(prefs);
																fnotdone = FALSE;
																if(trmsg->trm_ID == ID_PREFS_SAVE)
																	SavePrayerPrefs();
																break;
															case	ID_OUTPUT_DRIVER_CONFIGURE: {
																	struct AHIAudioModeRequester *req;

																	if(req = AHI_AllocAudioRequest(
																									AHIR_PubScreenName, Screen,
																									AHIR_TitleText,"Select a mode and rate",
																									AHIR_InitialMixFreq, prefs->MixFreq,
																									AHIR_DoMixFreq, TRUE,
																									AHIR_InitialAudioID, prefs->AudioID,
																									TAG_DONE)) {
																			if(AHI_AudioRequestA(req, NULL)) {
																					prefs->MixFreq = req->ahiam_MixFreq;
																					prefs->AudioID = req->ahiam_AudioID;
																				}
																			AHI_FreeAudioRequest(req);
																		}
																}																
																break;
														}
													break;
											}
										TR_ReplyMsg(trmsg);
									}
							}
						if(prj_Audio)
							TR_CloseProject(prj_Audio);
						if(prj_Visual)
							TR_CloseProject(prj_Visual);
						if(prj_Hotkeys)
							TR_CloseProject(prj_Hotkeys);
						if(prj_Plugins)
							TR_CloseProject(prj_Plugins);
						
						prj_Hotkeys = prj_Plugins = prj_Visual = prj_Audio = NULL;
						
						TR_CloseProject(project);
					} else Printf("Can't open project\n");
				TR_CloseTriton();
			} else Printf("Can't open triton\n");

		if(qPathReq)
			rtFreeRequest(qPathReq);

		if(qRobeReq) {
				strncpy(robe_path, qRobeReq->Dir, sizeof(robe_path) - 1);
				rtFreeRequest(qRobeReq);
			}
		if(qPluginReq) {
				strncpy(plugin_path, qPluginReq->Dir, sizeof(plugin_path) - 1);
				rtFreeRequest(qPluginReq);
			}

		while(node = (APTR) RemTail(&PrefsPluginList))
			FreeVec(node);
		while(node = (APTR) RemTail(&PrefsHotkeyList))
			FreeVec(node);

		FreePrayerPrefs(prefs);
	}

#endif

Forbid();
PutMsg(PrefsPort, &DeathMsg);
PrefsProc = NULL;
}

static void set_capture_plugin(struct PrayerPrefs *prefs, char *dir, char *name)
{
struct PrayerPlugin *ppl;
char *path;
ULONG i;

i = strlen(dir) + strlen(name) + 5;

if(path = AllocVec(i, MEMF_ANY)) {
		strcpy(path, dir);
		AddPart(path, name, i);
		if(ppl = OpenPlugin(path)) {
				if(ppl->Node.ln_Type == PPL_TYPE_EXPORT) {
						FreeVec(prefs->CapturePlugin);
						prefs->CapturePlugin = StrDup(path);
						if(prj_Plugins) {
								TR_SetAttribute(prj_Plugins, ID_CAPTURE_PLUGIN, TRAT_Text, (ULONG) prefs->CapturePlugin);
								TR_SetAttribute(prj_Plugins, ID_CAPTURE_PLUGIN_DESC, TRAT_Text, (ULONG)(ppl->Description ? ppl->Description : "N/A"));
								TR_SetAttribute(prj_Plugins, ID_CAPTURE_PLUGIN_AUTHOR, TRAT_Text, (ULONG)(ppl->Author ? ppl->Author : "N/A"));
							}
						i = 0;
					}
				ClosePlugin(ppl);
			}
		if(i)
			DisplayBeep(NULL);
		FreeVec(path);
	}
}

static void pubscreen_select(struct TR_Project *project, struct PrayerPrefs *prefs)
{
struct List *list;
struct	Node	*node;
struct	List	list_copy;
UWORD	i;

NewList(&list_copy);

CreateNode(SCREEN_DEFAULT, &list_copy);

if(list = LockPubScreenList()) {
		for(node = list->lh_Head; node->ln_Succ; node = node->ln_Succ) {
				CreateNode(node->ln_Name, &list_copy);
			}
		UnlockPubScreenList();
	}

i = PickEntry(&list_copy, 1, 0xffff);

if(node = GetNode(&list_copy, i)) {
		FreeVec(prefs->Screen);
		if(i)
			prefs->Screen = StrDup(node->ln_Name);
		else
			prefs->Screen = NULL;

		TR_SetAttribute(project, ID_PUBSCREEN_TITLE, TRAT_Text, (ULONG) (prefs->Screen ? prefs->Screen : SCREEN_DEFAULT));
	}
while(node = RemTail(&list_copy)) {
		FreeVec(node);
	}
}


static UWORD PickEntry(struct List *list, UWORD top, UWORD value)
{
struct TR_Message *trmsg;
struct TR_Project *project;
ULONG signals, sl = 0, ml = 0;
BOOL	fnotdone = TRUE;
UWORD	olv = 0xffff;

if(project = TR_OpenProjectTags(Application,
							WindowID(2),
							QuickHelpOn(TRUE),
							WindowTitle("Select"),
							WindowPosition(TRWP_MOUSEPOINTER),
							WindowFlags(TRWF_NOSIZEGADGET),
							TRWI_PubScreenName, Screen,
							WindowBackfillReq,
							HorizGroupA,
								Space,
								VertGroupA,
									Space,
									ListSS(list, 1, top, value), TRAT_MinWidth, 25, TRAT_MinHeight, 7,
									SpaceS,
									HorizGroupE,
										Button("Select", 2),
										Button("Cancel", 0),
										EndGroup,
									Space,
									EndGroup,
								Space,
								EndGroup,
							EndProject)) {
		while(fnotdone) {
				signals = TR_Wait(Application, SIGBREAKF_CTRL_C);
				if(signals & SIGBREAKF_CTRL_C) {
						Signal(FindTask(0), SIGBREAKF_CTRL_C);
						fnotdone = FALSE;
					}
				while(trmsg = TR_GetMsg(Application)) {
						switch(trmsg->trm_Class) {
								case	TRMS_CLOSEWINDOW:
									fnotdone = FALSE;
									value = 0xffff;
									break;
								case	TRMS_ERROR:
									PutStr(TR_GetErrorString(trmsg->trm_Data));
									break;
								case	TRMS_NEWVALUE:
									if(trmsg->trm_ID == 1) {
											value = trmsg->trm_Data;
											if((olv == value) && DoubleClick(sl, ml, trmsg->trm_Seconds, trmsg->trm_Micros)) {
													fnotdone = FALSE;
												} else {
													sl = trmsg->trm_Seconds;
													ml = trmsg->trm_Micros;
													olv = trmsg->trm_Data;
												}
										}
									break;
								case	TRMS_ACTION:
									switch(trmsg->trm_ID) {
											case	0:
												fnotdone = FALSE;
												value = 0xffff;
												break;
											case	2:
												fnotdone = FALSE;
												break;
										}
									break;
							}
						TR_ReplyMsg(trmsg);
					}
			} 
		TR_CloseProject(project);
	} else Printf("PickEntry: Can't create triton project\n");

return(value);
}

static struct Node *CreateNode(char *zp, struct List *list)
{
struct Node *node;
if(node = AllocVec(sizeof(*node) + strlen(zp) + 1, MEMF_ANY)) {
		node->ln_Name = (APTR) (node + 1);
		strcpy(node->ln_Name, zp);
		AddTail(list, node);
	}
return(node);
}

static void LoadPrayerPrefs(void)
{
struct PrayerPrefs *prefs = NULL;
struct FileInfoBlock *fib;
struct Node *node;
LONG	i;
UBYTE	*buffer;
union pchunk {
		UBYTE	*cptr;
		struct PrefsChunk *prefs;
		struct NameChunk *name;
		struct HotkeyChunk *hkey;
	} c;
BPTR pfile;

// first set defaults...
if(prefs = AllocVec(sizeof(*prefs), MEMF_CLEAR)) {
		prefs->Layer3ForceMono = 
		prefs->Layer12ForceMono = TRUE;
		prefs->Layer3FreqDiv = 
		prefs->Layer12FreqDiv = 4;
		prefs->Layer3Qual =
		prefs->Layer12Qual = 0;
		prefs->Robe = StrDup("Robes/Default.robe");
		prefs->CapturePath = StrDup("");
		prefs->CapturePlugin = StrDup("");
		prefs->Screen = NULL;
		prefs->RemapFlag = TRUE;
		prefs->RemapPrecision = -1;
		prefs->AutoFadeIn = 180;
		prefs->AutoFadeOut = 180;
		prefs->RTG = FALSE;
		prefs->RefreshRate = 15;
		prefs->Flags = 0L;
		prefs->PlayerPri = 0;
		prefs->Hotkeys = AllocVec(sizeof(*prefs->Hotkeys) * PRHK_MAX, MEMF_CLEAR);

		for(i = 0; (i < PRHK_MAX) && hk_defaults[i].Key; i++) {
				strcpy(prefs->Hotkeys[i].Key, hk_defaults[i].Key);
				prefs->Hotkeys[i].Global = hk_defaults[i].Global;
				prefs->Hotkeys[i].Repeat = hk_defaults[i].Repeat;
			}
		
		NewList(&prefs->Plugins);

		if(pfile = Open("ENVARC:Prayer2.prefs", MODE_OLDFILE)) {
				if(fib = AllocVec(sizeof(*fib), MEMF_ANY)) {
						if(ExamineFH(pfile, fib)) {
								if(buffer = AllocVec(fib->fib_Size, MEMF_CLEAR)) {
										c.cptr = buffer;
										i = Read(pfile, buffer, fib->fib_Size);
										while(i > 0) {
												switch(c.prefs->ChunkType) {
														case	PREFS_HEADER_ID:
															prefs->Layer3ForceMono = c.prefs->Layer3ForceMono;
															prefs->Layer3FreqDiv = c.prefs->Layer3FreqDiv;
															prefs->Layer3Qual = c.prefs->Layer3Qual;
															prefs->Layer12ForceMono = c.prefs->Layer12ForceMono;
															prefs->Layer12FreqDiv = c.prefs->Layer12FreqDiv;
															prefs->Layer12Qual = c.prefs->Layer12Qual;
															prefs->AudioDriverId = c.prefs->AudioDriverId;
															prefs->RemapFlag = c.prefs->RemapFlag;
															prefs->RemapPrecision = c.prefs->RemapPrecision;
															if(c.prefs->ChunkSize >= (offsetof(struct PrefsChunk, AutoFadeOut) + sizeof(UWORD))) {
																	prefs->AutoFadeIn = c.prefs->AutoFadeIn;
																	prefs->AutoFadeOut = c.prefs->AutoFadeOut;
																}
															if(c.prefs->ChunkSize >= (offsetof(struct PrefsChunk, RTG) + sizeof(BOOL))) {
																	prefs->RTG = c.prefs->RTG;
																}
															if(c.prefs->ChunkSize >= (offsetof(struct PrefsChunk, RefreshRate) + sizeof(UWORD))) {
																	prefs->RefreshRate = c.prefs->RefreshRate;
																}
															if(c.prefs->ChunkSize >= (offsetof(struct PrefsChunk, Flags) + sizeof(ULONG))) {
																	prefs->Flags = c.prefs->Flags;
																}
															if(c.prefs->ChunkSize >= (offsetof(struct PrefsChunk, PlayerPri) + sizeof(BYTE))) {
																	prefs->PlayerPri = c.prefs->PlayerPri;
																}
															if(c.prefs->ChunkSize >= (offsetof(struct PrefsChunk, AudioID) + sizeof(ULONG))) {
																	prefs->AudioID = c.prefs->AudioID;
																}
															if(c.prefs->ChunkSize >= (offsetof(struct PrefsChunk, MixFreq) + sizeof(ULONG))) {
																	prefs->MixFreq = c.prefs->MixFreq;
																}
															break;
														case	PREFS_PLUGIN_ID:
															if(node = CreateNode(c.name->Name, &prefs->Plugins)) {
																	node->ln_Name = FilePart(node->ln_Name);
																}
															break;
														case	PREFS_ROBE_ID:
															FreeVec(prefs->Robe); prefs->Robe = StrDup(c.name->Name);
															break;
														case	PREFS_SCREEN_ID:
															FreeVec(prefs->Screen); prefs->Screen = StrDup(c.name->Name);
															break;
														case	PREFS_CPATH_ID:
															FreeVec(prefs->CapturePath); prefs->CapturePath = StrDup(c.name->Name);
															break;
														case	PREFS_CPLUG_ID:
															FreeVec(prefs->CapturePlugin); prefs->CapturePlugin = StrDup(c.name->Name);
															break;
														case	PREFS_HOTKEY_ID:
															if(c.hkey->ID != 0xffff && c.hkey->ID < PRHK_MAX) {
																		strncpy(prefs->Hotkeys[c.hkey->ID].Key, c.hkey->Name, sizeof(prefs->Hotkeys[0].Key) - 1);
																		prefs->Hotkeys[c.hkey->ID].Global = c.hkey->Global;
																		prefs->Hotkeys[c.hkey->ID].Repeat = c.hkey->Repeat;
																}
															break;
													}
												i -= c.prefs->ChunkSize;
												c.cptr += c.prefs->ChunkSize;
											}
										FreeVec(buffer);
									}
							}
						FreeVec(fib);
					}
				Close(pfile);
			}
		SetPrayerPrefs(prefs);
	}
}

static void WriteNameChunk(BPTR pfile, ULONG id, char *data)
{
struct NameChunk Chunk;
LONG i = strlen(data) + 1;

if(i & 1)
	i++;

Chunk.ChunkType = id;
Chunk.ChunkSize = sizeof(Chunk) + i;
Write(pfile, &Chunk, sizeof(Chunk));
Write(pfile, data, i);
}

static void WriteHotkeyChunk(BPTR pfile, UWORD id, UWORD gl, UWORD rpt, char *data)
{
struct HotkeyChunk Chunk;
LONG i = strlen(data) + 1;

if(i & 1)
	i++;

Chunk.ChunkType = PREFS_HOTKEY_ID;
Chunk.ChunkSize = sizeof(Chunk) + i;
Chunk.ID = id;
Chunk.Global = gl;
Chunk.Repeat = rpt;
Write(pfile, &Chunk, sizeof(Chunk));
Write(pfile, data, i);
}


static void SavePrayerPrefs(void)
{
struct PrayerPrefs *prefs;
struct PrefsChunk *pr;
struct Node *node;
int i;

BPTR pfile;

if(prefs = GetPrayerPrefs()) {
		if(pfile = Open("ENVARC:Prayer2.prefs", MODE_NEWFILE)) {
				if(pr = AllocVec(sizeof(*pr), MEMF_CLEAR)) {
						pr->ChunkType = PREFS_HEADER_ID;
						pr->ChunkSize = sizeof(*pr);

						pr->Layer3ForceMono = prefs->Layer3ForceMono;
						pr->Layer3FreqDiv = prefs->Layer3FreqDiv;
						pr->Layer3Qual = prefs->Layer3Qual;
						pr->Layer12ForceMono = prefs->Layer12ForceMono;
						pr->Layer12FreqDiv = prefs->Layer12FreqDiv;
						pr->Layer12Qual = prefs->Layer12Qual;
						pr->AudioDriverId = prefs->AudioDriverId;
						pr->RemapFlag = prefs->RemapFlag;
						pr->RemapPrecision = prefs->RemapPrecision;
						pr->AutoFadeIn = prefs->AutoFadeIn;
						pr->AutoFadeOut = prefs->AutoFadeOut;
						pr->RTG = prefs->RTG;
						pr->RefreshRate = prefs->RefreshRate;
						pr->PlayerPri = prefs->PlayerPri;
						pr->Flags = prefs->Flags;
						pr->MixFreq = prefs->MixFreq;
						pr->AudioID = prefs->AudioID;

						Write(pfile, pr, pr->ChunkSize);
						
						if(prefs->Robe)
							WriteNameChunk(pfile, PREFS_ROBE_ID, prefs->Robe);
						if(prefs->Screen)
							WriteNameChunk(pfile, PREFS_SCREEN_ID, prefs->Screen);
						if(prefs->CapturePlugin)
							WriteNameChunk(pfile, PREFS_CPLUG_ID, prefs->CapturePlugin);
						if(prefs->CapturePath)
							WriteNameChunk(pfile, PREFS_CPATH_ID, prefs->CapturePath);

						for(node = prefs->Plugins.lh_Head; node->ln_Succ; node = node->ln_Succ) {
								WriteNameChunk(pfile, PREFS_PLUGIN_ID, (char *)(node + 1));
							}
						for(i = 0; i < PRHK_MAX; i++) {
								WriteHotkeyChunk(pfile, i, prefs->Hotkeys[i].Global,
																					prefs->Hotkeys[i].Repeat,
																					prefs->Hotkeys[i].Key);
							}
					}
				Close(pfile);
			}
		FreePrayerPrefs(prefs);
	}
}


static APTR GetNode(struct List *list, UWORD i)
{
struct Node *node;

if(i == 0xffff)
	return(NULL);
	
node = list->lh_Head;

while(i--) {
		if(node->ln_Succ)
			node = node->ln_Succ;
		else
			break;
	}
if(node->ln_Succ)
	return(node);
else
	return(NULL);
}

static void add_plugin(struct TR_Project *proj, struct PrayerPrefs *prefs, char *dir, char *file)
{
struct PluginInfo *pnode = NULL;
char *pname;
LONG i;

i = strlen(dir) + strlen(file) + 4;
if(pname = AllocVec(i, MEMF_ANY)) {
		strcpy(pname, dir);
		AddPart(pname, file, i);
		if(pnode = GetPluginInfo(pname))
			TR_SetAttribute(proj, ID_PLUGIN_LIST, 0, (ULONG)&PrefsPluginList);
		else
			DisplayBeep(NULL);
	}
}

static void update_plugin_display(UWORD cplugin, APTR project)
{
struct PluginInfo *pnode;

if(pnode = GetNode(&PrefsPluginList, cplugin)) {
		TR_SetAttribute(project, ID_PLUGIN_DESCR, TRAT_Text, (ULONG)(pnode->Description));
		TR_SetAttribute(project, ID_PLUGIN_AUTHOR, TRAT_Text, (ULONG)(pnode->Author));
	} else {
		TR_SetAttribute(project, ID_PLUGIN_DESCR, TRAT_Text, (ULONG)"");
		TR_SetAttribute(project, ID_PLUGIN_AUTHOR, TRAT_Text, (ULONG)"");
	}
}

static struct PluginInfo *GetPluginInfo(char *path)
{
struct PrayerPlugin *ppl;
struct PluginInfo *pnode = NULL;

if(ppl = OpenPlugin(path)) {
		if(ppl->Node.ln_Type == PPL_TYPE_DECODER) {
				for(pnode = (APTR) PrefsPluginList.lh_Head; pnode->Node.ln_Succ; pnode = (APTR) pnode->Node.ln_Succ) {
						if(!Stricmp(pnode->Name, ppl->Name)) {
								ClosePlugin(ppl);
								return(NULL);
							}
					}

				if(pnode = AllocVec(sizeof(*pnode) + strlen(path), MEMF_CLEAR)) {
						strncpy(pnode->Name, ppl->Name, sizeof(pnode->Name) - 1);
						strncpy(pnode->Title, ppl->Title ? ppl->Title : "N/A", sizeof(pnode->Title) - 1);
						strncpy(pnode->Description, ppl->Description ? ppl->Description : "N/A", sizeof(pnode->Description) - 1);
						strncpy(pnode->Author, ppl->Author ? ppl->Author : "N/A", sizeof(pnode->Author) - 1);
						strcpy(pnode->Path, path);
						pnode->Node.ln_Name = pnode->Name;

						AddTail(&PrefsPluginList, (APTR) pnode);
					}
			}
		ClosePlugin(ppl);
	}
return(pnode);
}



static void PrjToFront(APTR prj)
{
APTR win;
if(win = TR_ObtainWindow(prj)) {
		WindowToFront(win);
		ActivateWindow(win);
		TR_ReleaseWindow(win);
	}
}

static void pr_OpenAudioPrj(struct PrayerPrefs *prefs)
{

if(prj_Audio)
	PrjToFront(prj_Audio);
else 
	prj_Audio = TR_OpenProjectTags(Application, 	
		WindowTitle("Audio Settings"),
		QuickHelpOn(TRUE),
		TRWI_PubScreenName, Screen,
		WindowPosition(TRWP_CENTERDISPLAY),
		WindowDimensions(&audio_dim),
		WindowFlags(TRWF_NOSIZEGADGET),
		WindowBackfillReq,
		HorizGroupA,
			Space,
			VertGroupA,
				Space,
				GroupBox,
				ObjectBackfillB,
				HorizGroupA,
					Space,
					VertGroupA,
						Space,

						NamedFrameBox("MPEG Layer 3"),
						LineArray,
							SpaceS,
							BeginLine,
								SpaceS, TextNR("Mode"), Space, CycleGadget(&cy_mode, prefs->Layer3ForceMono, ID_LAYER3_MODE), SpaceS,
								EndLine,
							SpaceS,
							BeginLine,
								SpaceS, TextNR("Frequency"), Space, CycleGadget(&cy_freq, prefs->Layer3FreqDiv / 2, ID_LAYER3_FREQ), SpaceS,
								EndLine,
							SpaceS,
							BeginLine,
								SpaceS, TextNR("Quality"), Space, CycleGadget(&cy_qual, prefs->Layer3Qual, ID_LAYER3_QUAL),SpaceS,
								EndLine,
							Space,
							EndArray,
						Space,
						NamedFrameBox("MPEG Layer 1 & 2"),
						LineArray,
							SpaceS,
							BeginLine,
								SpaceS, TextNR("Mode"), Space, CycleGadget(&cy_mode, prefs->Layer12ForceMono, ID_LAYER12_MODE), SpaceS,
								EndLine,
							SpaceS,
							BeginLine,
								SpaceS, TextNR("Frequency"), Space, CycleGadget(&cy_freq, prefs->Layer12FreqDiv / 2, ID_LAYER12_FREQ), SpaceS,
								EndLine,
							SpaceS,
							BeginLine,
								SpaceS, TextNR("Quality"), Space, CycleGadget(&cy_qual, prefs->Layer12Qual, ID_LAYER12_QUAL), SpaceS,
								EndLine,
							Space,
							EndArray,
						Space,
						NamedFrameBox("Audio Driver"),
						LineArray,
							Space,
							BeginLine,
								Space, TextNR("Task Priority"), Space,
												GroupBox,
												VertGroupA, SpaceS,
												HorizGroupAC, SpaceS,
													SliderGadget(-10, 10, prefs->PlayerPri, ID_PLAYER_PRI),
													Space,
													Integer(prefs->PlayerPri), TRAT_ID, ID_PLAYER_PRI_INT, TRAT_MinWidth, 3,
													SpaceS,
													EndGroup,
												SpaceS,
												EndGroup,
								Space,
								EndLine,
							Space,
							BeginLine,
								Space, HorizGroupAC,
													TextNR("AHI"), Space,
													CheckBoxV((prefs->AudioDriverId == 3) ? TRUE : FALSE, ID_CHECKBOX_AHI),
													EndGroup,
								Space,
								Button("Select Mode", ID_OUTPUT_DRIVER_CONFIGURE),
								Space,
								EndLine,
							Space,
							EndArray,
						Space,
						EndGroup,
					Space,
					EndGroup,
				Space,
				EndGroup,
			Space,
			EndGroup,
		EndProject);

}

static void pr_OpenVisualPrj(struct PrayerPrefs *prefs)
{
int i = 0;

if(prj_Visual)
	PrjToFront(prj_Visual);
else {

		if(prefs->RemapFlag) {
				if(prefs->RemapPrecision < 0)
					i = 1;
				else if(prefs->RemapPrecision == 0)
					i = 2;
				else
					i = 3;
			}

	prj_Visual = TR_OpenProjectTags(Application, 	
		WindowTitle("Visual Settings"),
		QuickHelpOn(TRUE),
		WindowDimensions(&visual_dim),
		WindowPosition(TRWP_CENTERDISPLAY),
		TRWI_PubScreenName, Screen,
		WindowFlags(TRWF_NOSIZEGADGET),
		WindowBackfillReq,
		HorizGroupA,
			Space,
			VertGroupA,
				Space,
				GroupBox,
				ObjectBackfillB,
				HorizGroupA,
					Space,
					VertGroupA,
						Space,


						LineArray,
							SpaceS,
							BeginLine, SpaceS,
								TextNR("Robe"), Space, HorizGroup, ClippedTextBox(prefs->Robe, ID_ROBE_TITLE),
								GetFileButton(ID_ROBE_SELECT), EndGroup,
								SpaceS, EndLine,
							SpaceS,
							BeginLine, SpaceS,
								TextNR("Screen"), Space, HorizGroup, ClippedTextBox(prefs->Screen ? prefs->Screen : SCREEN_DEFAULT, ID_PUBSCREEN_TITLE),
								GetEntryButton(ID_PUBSCREEN_SELECT), EndGroup,
								SpaceS, EndLine,
							SpaceS,
							BeginLine, SpaceS,
								TextNR("Precision"), Space,
								CycleGadget(&cy_remap, i, ID_REMAP),
								Space,
								TextNR("RTG"),
								Space,
								CheckBoxV(prefs->RTG, ID_RTG), TRAT_Flags, TRBU_YRESIZE,
								Space, EndLine,
							SpaceS,
							BeginLine, SpaceS,
								TextNR("Icon"), Space,
								CycleGadget(&cy_icon, (prefs->Flags & PRF_ALWAYS_ADD_ICON) ? 1 : 0, ID_ICON),
								Space, EndLine,
							SpaceS,
							BeginLine, SpaceS,
								TextNR("Refresh"), Space,
								SliderGadget(5, 25, prefs->RefreshRate, ID_REFRESH_RATE),
								Space,
								TextN(fps_str(prefs->RefreshRate)), TRAT_ID, ID_REFRESH_RATE_TX, TRAT_MinWidth, 6,
								Space, EndLine,
							Space,
							EndArray,

						Space,
						EndGroup,
					Space,
					EndGroup,
				Space,
				EndGroup,
			Space,
			EndGroup,
		EndProject);
	}
}

static void pr_OpenPluginPrj(struct PrayerPrefs *prefs)
{
struct PrayerPlugin *ppl;
char *pauth = "N/A", *pdesc = "N/A";

if(prj_Plugins)
	PrjToFront(prj_Plugins);
else {
	if(ppl = OpenPlugin(prefs->CapturePlugin)) {
			if(ppl->Author)
				pauth = ppl->Author;
			if(ppl->Description)
				pdesc = ppl->Description;
		}

	prj_Plugins = TR_OpenProjectTags(Application, 	
		WindowTitle("Plugin Settings"),
		QuickHelpOn(TRUE),
		WindowDimensions(&plugin_dim),
		WindowPosition(TRWP_CENTERDISPLAY),
		TRWI_PubScreenName, Screen,
		WindowBackfillReq,
		HorizGroupA,
			Space,
			VertGroupA,
				Space,
				GroupBox,
				ObjectBackfillB,
				VertGroupA,
					Space,
					HorizGroupA,
						Space,
						VertGroupA,
							NamedSeparator("Decoder Plugins"),
							SpaceS,
							ListSS(&PrefsPluginList, ID_PLUGIN_LIST, 1, 0xffff),
							Space,
							VertGroupA,
								LineArray,
									BeginLine,
										TextNR("Description"),Space, ClippedTextBox("", ID_PLUGIN_DESCR),
										EndLine,
									SpaceS,
									BeginLine,
										TextNR("Author"),Space, ClippedTextBox("", ID_PLUGIN_AUTHOR),
										EndLine,
									EndArray,
								SpaceS,
								HorizGroupE,
									Button("Add Plugin", ID_PLUGIN_ADD),
									Button("Remove Plugin", ID_PLUGIN_REM),
									EndGroup,
								Space,
								NamedSeparator("Output Capture Plugin"),
								Space,

								LineArray,
									BeginLine,
										TextNR("Plugin"), Space, ClippedTextBox(prefs->CapturePlugin, ID_CAPTURE_PLUGIN), GetFileButton(ID_CAPTURE_PLUGIN_PICK),
										EndLine,
									SpaceS,
									BeginLine,
										TextNR("Description"),Space, ClippedTextBox(pdesc, ID_CAPTURE_PLUGIN_DESC),
										EndLine,
									SpaceS,
									BeginLine,
										TextNR("Author"),Space, ClippedTextBox(pauth, ID_CAPTURE_PLUGIN_AUTHOR),
										EndLine,
									SpaceS,
									BeginLine,
										TextNR("Save Path"), Space, StringGadget(prefs->CapturePath, ID_CAPTURE_PATH), GetDrawerButton(ID_CAPTURE_PATH_PICK),
										EndLine,

									EndArray,

								EndGroup,
							EndGroup,
						Space,
						EndGroup,
					Space,
					EndGroup,
				Space,
				EndGroup,
			Space,
			EndGroup,
		EndProject);

		if(ppl)
			ClosePlugin(ppl);
	}
}

static void pr_OpenHotkeysPrj(struct PrayerPrefs *prefs)
{
if(prj_Hotkeys)
	PrjToFront(prj_Hotkeys);
else 
	prj_Hotkeys = TR_OpenProjectTags(Application, 	
		WindowTitle("Hotkeys Settings"),
		QuickHelpOn(TRUE),
		TRWI_PubScreenName, Screen,
		WindowDimensions(&hotkeys_dim),
		WindowPosition(TRWP_CENTERDISPLAY),
		WindowBackfillReq,
		HorizGroupA,
			Space,
			VertGroupA,
				Space,
				GroupBox,
				ObjectBackfillB,
				HorizGroupA,
					Space,
					VertGroupAC,
						Space,
						NamedSeparator("Functions"),
						ListSS(&PrefsHotkeyList, ID_HOTKEY_LIST, 0, ~0), TRAT_MinWidth, 25, TRAT_MinHeight, 7,
						SpaceS,
						HorizGroupAC,
							StringGadget(hotkey_bf, ID_STRING_HOTKEY),
							SpaceS,
							CheckBox(ID_CHECKBOX_GLOBAL), TextN("Global"),
							Space,
							CheckBox(ID_CHECKBOX_REPEAT), TextN("Repeat"),
							EndGroup,
						Space,
						EndGroup,
					Space,
					EndGroup,
				Space,
				EndGroup,
			Space,
			EndGroup,
		EndProject);
}

static char *fps_str(UWORD fps)
{
static char ab[20];
SPrintf(ab, "%ld fps", fps);
return(ab);
}
