/*
 * =-=-=-=-=-=-=-=-=-=-=-  NOTICE  =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * The following is the license that applies to this copy of the Prayer2
 * software.  For a license to use the Prayer2 software under conditions
 * other than those described here, please contact MindNever Software at
 * following addresses:
 *
 * MindNever Software <info@mindnever.org>
 * MindNever Software +381 63 550-161 or +381 64 220-7115
 * 
 * =-=-=-=-=-=-=-=-=-=-  COPYRIGHT  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Prayer2 is:
 *   Copyright (c) 1999-2001 by Tamara Cetinjski <tamara@mindnever.org>
 *   Copyright (c) 2001-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *   Copyright (c) 1999-2002 by MindNever Software Group <info@mindnever.org>
 *
 * Tamara Cetinjski is
 *   Copyright (c) 1999-2002 by Vladimir Zidar <mr_W@mindnever.org>
 *
 *
 * =-=-=-=-=-=-=-=-=-=-  LICENSE  -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Redistributions in any form must be accompanied by information on
 *    how to obtain complete source code for the Prayer2 software and any
 *    accompanying software that uses the Prayer2 software.  The source code
 *    must either be included in the distribution or be available for no
 *    more than the cost of distribution plus a nominal fee, and must be
 *    freely redistributable under reasonable conditions.  For an
 *    executable file, complete source code means the source code for all
 *    modules it contains.  It does not include source code for modules or
 *    files that typically accompany the major components of the operating
 *    system on which the executable file runs.
 *
 * =-=-=-=-=-=-=-=-=-=-  NO WARRANTY =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= 
 *
 * THIS SOFTWARE IS PROVIDED BY MINDNEVER SOFTWARE ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, OR
 * NON-INFRINGEMENT, ARE DISCLAIMED.  IN NO EVENT SHALL MINDNEVER SOFTWARE
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, HAIR, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
#include	<exec/types.h>
#include	"pif.h"

#ifndef MAKE_ID
#define	MAKE_ID(a,b,c,d) (((a) << 24) | ((b) << 16) | ((c) << 8) | (d))
#endif


#define	ID_PRAYER_ROBEINFO	MAKE_ID('P','R','R','I')
#define	ID_PRAYER_WINDOW	MAKE_ID('P','R','W','N')
#define	ID_PRAYER_BUTTON	MAKE_ID('P','R','B','T')
#define	ID_PRAYER_SLIDER	MAKE_ID('P','R','S','L')
#define	ID_PRAYER_LEVEL		MAKE_ID('P','R','L','V')
#define	ID_PRAYER_DISPLAY	MAKE_ID('P','R','D','S')
#define	ID_PRAYER_LED			MAKE_ID('P','R','L','D')
#define	ID_PRAYER_FONT		MAKE_ID('P','R','F','N')
#define	ID_PRAYER_IMAGE		MAKE_ID('P','R','I','M')
#define	ID_PRAYER_LISTER	MAKE_ID('P','R','L','I')
#define	ID_PRAYER_TEXT		MAKE_ID('P','R','T','X')
#define	ID_PRAYER_MENU		MAKE_ID('P','R','M','N')
#define ID_PRAYER_GROUP		MAKE_ID('P','R','G','R')

struct	RobeChunk	{
		ULONG	ChunkType;
		ULONG	ChunkSize;
	};

struct RobeInfo	{
		ULONG	ChunkType;
		ULONG	ChunkSize;
		UWORD	RobeVersion;
		UWORD	NumWindows;
		UWORD	NumFonts;
		UWORD	NumImages;
		UWORD	NumTexts;
		UWORD	NumMenus; // V2
	};

struct RobeWindow {
	ULONG ChunkType; // PRWN
	ULONG ChunkSize; //
	UWORD	ImageID;
	UWORD	WindowID;
	UWORD Flags;	
	UWORD	ActiveImageID;
	UWORD	InActiveImageID;
	UWORD	StateImageX;
	UWORD	StateImageY;
	UWORD	TranspMaskID; //V3
	UBYTE	Title[32];
	UWORD	DefaultGroup;
};

struct	RobeGroup {
	ULONG	ChunkType;
	ULONG	ChunkSize;
	UWORD	GroupID;
	UWORD	Left;
	UWORD	Top;
	UWORD	Width;
	UWORD	Height;
	UBYTE Color[3];
	UWORD	ImageID;
};


struct	RobeMenu {
	ULONG	ChunkType;
	ULONG	ChunkSize;
	UWORD	Function;
	UWORD	Argument;
	UBYTE	Title[0];
};
	
struct	RobeText {
	ULONG	ChunkType; // PRTX
	ULONG	ChunkSize;
	UWORD	TextID;
	ULONG	RawLen;
	UBYTE	CompressedData[0];
};

struct	RobeButton	{
	ULONG ChunkType; // PRBT
	ULONG ChunkSize;
	UWORD ImageOffID;
	UWORD ImageOnID;
	UWORD ImageAltID;
	UWORD Left;
	UWORD Top;
	UWORD Function;
	UWORD Argument;
	UWORD	Width,Height;
	UWORD Function_RMB;
	UWORD Argument_RMB;
	UWORD	GroupID;
};


struct RobeSlider {
	ULONG ChunkType; // PRSL
	ULONG ChunkSize;
	UWORD	Function;
	UWORD Flags;
	UWORD ImageOffID;
	UWORD ImageOnID;
	UWORD ImageAltID;
	UWORD Left;
	UWORD Top;
	UWORD Width;
	UWORD Height;
	UWORD	Argument;
	UWORD	GroupID;
};

#define	PRF_WINDOW_BORDERLESS	(1L << 0)
#define	PRF_WINDOW_AUTOOPEN	(1L << 1)
#define	PRF_WINDOW_CENTERED	(1L << 2)
#define	PRF_WINDOW_HIDDEN		(1L << 3)
#define	PRF_WINDOW_BACKDROP	(1L << 4)

#define	PRF_SLIDER_VERTICAL	(1L << 0)
#define	PRF_SLIDER_TRANSPARENT	(1L << 1)

struct RobeLevel {
	ULONG ChunkType; // PRLV
	ULONG ChunkSize;
	UWORD Flags;
	UWORD Left;
	UWORD Top;
	UWORD NumLeds;
	UWORD LedHigh;
	UWORD LedWidth;
	UWORD LedHeight;
	UWORD LedSpace;
	UBYTE ColorOff[3];
	UBYTE ColorOnLow[3];
	UBYTE ColorOnHigh[3];
	UBYTE Type;
	UWORD	GroupID;
	};

#define	PRF_LEVEL_VERTICAL	(1L << 0)

struct	RobeDisplay {
	ULONG ChunkType; // PRDS
	ULONG ChunkSize;
	UWORD Flags;
	UWORD FontID;
	UWORD Left;
	UWORD Top;
	UWORD Width;
	UWORD Height;
	UBYTE BgColor[3];
	UBYTE FgColor[3];
	UWORD Type;
	UWORD	GroupID;
};

#define	PRF_DISPLAY_CENTERED	(1L << 0)
#define	PRF_DISPLAY_TRANSPARENT	(1L << 1)
#define PRF_DISPLAY_SCROLLING (1L << 2)

struct RobeLed {
	ULONG ChunkType; // PRLD
	ULONG ChunkSize;
	UWORD Left;
	UWORD Top;
	UWORD Width;
	UWORD Height;
	UBYTE ColorOff[3];
	UBYTE ColorOn[3];
	ULONG Pattern;
	UWORD	GroupID;
	};

struct RobeLister {
	ULONG ChunkType; // PRLI
	ULONG ChunkSize;
	UWORD	Left;
	UWORD	Top;
	UWORD	Width;
	UWORD	Height;
	UBYTE	BgColor[3];
	UBYTE	FgColor[3];
	UBYTE	BgSelect[3];
	UBYTE	FgSelect[3];
	UBYTE	BgActive[3];
	UBYTE	FgActive[3];
	UWORD	FontID;
	UWORD	Function;
	UWORD	Argument;
	UWORD	GroupID;
	};

struct RobeFont {
	ULONG ChunkType; // PRFN
	ULONG ChunkSize;
	UWORD	FontID;
	UWORD YSize;
	UBYTE Style;
	UBYTE Flags;
	UWORD XSize;
	UWORD Baseline;
	UWORD BoldSmear;
	UBYTE LoChar;
	UBYTE HiChar;
	UWORD Modulo;
	ULONG	RawLen;
	UBYTE	HaveCharSpace;
	UBYTE	HaveCharKern;
	UBYTE	CompressedData[0]; // Z compressed
	/* and actual font data, Modulo * YSize bytes */
	/* and HiChar-LoChar LONG-s for CharLoc */
	/* and HiChar-LoChar WORD-s for CharSpace */
	/* and HiChar-LoChar WORD-s for CharKern */
	};

struct RobeImage {
	ULONG ChunkType; // PRIM
	ULONG ChunkSize;
	UWORD	ImageID;
	struct	PIFHdr	Image;
	/* ... data ... */
	};
